/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "OpenThreadSimulator.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#if defined(HAVE_POSIX_TIMERS)
// clock_gettime
#include <time.h>
#endif

#if defined(_WIN32) || defined(WIN32)
#include <windef.h>
// timeGetTime
#include <mmsystem.h>
#else

// gettimeofday
#include <sys/time.h>
#endif

using namespace OpenThreads;

IMPLEMENT_SIMULATOR_FACTORY(100, OpenThreadSimulator);

struct TimerThread : public Thread {
    
    // For the timer thread to be able to wait for some desired time
    OpenThreads::Mutex mutex;
    OpenThreads::Condition condition;
    
    // Need to be initialized. See below
    double delay;

    OpenThreadSimulator* simulator;
    
    TimerThread(OpenThreadSimulator* simu) {
        simulator = simu;
    }
    
    virtual void run() {
        while (true) {
            mutex.lock();
            // delay too small for our precision!!!
            // increase delay as the intent is to release the CPU now
            // otherwise, the simulator will end up repeatedly calling this function so long as
            // it is in advance, then process only a few events, then again... which end up
            // with very bad performances and 100% CPU usage.
            // It is much better to relinquish the CPU now, be waked up later on, and let the
            // simulator catch up that ms
            if (delay<1e-3) delay=1e-3;
            condition.wait(&mutex, (long unsigned int)(delay * 1e3));
            // detection for timer cancellation. Use negative delay as flag
            if (delay>0) simulator->waitTimerCallBack();
            // next loop locks and waits for main thread to unlock
        }
    }

};


struct StartThread : public Thread {

    OpenThreadSimulator* simulator;
    
    StartThread(OpenThreadSimulator* simu) {
        simulator = simu;
    }

    virtual void run() {
        simulator->eventLoop();
    };
    
};


OpenThreadSimulator::OpenThreadSimulator()
{
    Thread::Init(); // allow threading if it wasn't done before
    waitSemaphoreCounter = 0;
    timerThread = new TimerThread(this);
    // initially block the thread
    timerThread->mutex.lock();
    timerThread->start();
    //timerThread->setStackSize(256); // no stack is used by the run method, but we never know
    startThread = new StartThread(this);
}

OpenThreadSimulator::~OpenThreadSimulator()
{
    timerThread->cancel();
    startThread->cancel();
    delete timerThread;
    delete startThread;
}

void OpenThreadSimulator::eventMutexLock( )
{
    eventMutex.lock();
}

void OpenThreadSimulator::eventMutexUnlock( )
{
    eventMutex.unlock();
}

bool OpenThreadSimulator::stepControlMutexTryLock( )
{
    return !stepControlMutex.trylock();
}

void OpenThreadSimulator::stepControlMutexLock( )
{
    stepControlMutex.lock();
}

void OpenThreadSimulator::stepControlMutexUnlock( )
{
    stepControlMutex.unlock();
}

void OpenThreadSimulator::waitSemaphorePost( )
{
    waitSemaphoreMutex.lock();
    ++waitSemaphoreCounter;
    waitSemaphoreCondition.signal();
    waitSemaphoreMutex.unlock();
}

void OpenThreadSimulator::waitSemaphoreFlush( )
{
    waitSemaphoreMutex.lock();
    waitSemaphoreCounter = 0;
    waitSemaphoreMutex.unlock();
}

void OpenThreadSimulator::waitSemaphoreWait( )
{
    waitSemaphoreMutex.lock();
    while (waitSemaphoreCounter<=0) {
        waitSemaphoreCondition.wait(&waitSemaphoreMutex);
    }
    --waitSemaphoreCounter;
    waitSemaphoreMutex.unlock();
}


void OpenThreadSimulator::startWaitTimer( double delay )
{
    if (!timerThread->isRunning()) {
        return; // should not happen
    }
    if (delay<=0) {                        // should not happen too!
        waitTimerCallBack();
        return; 
    }
    timerThread->delay = delay; // >0 now
    // unlock run loop and trigger condition wait
    timerThread->mutex.unlock();
}

void OpenThreadSimulator::cancelWaitTimer( )
{
    if (timerThread->mutex.trylock()) {
        //cerr << "trylock failed!" << endl;
        return; // don't cancel before the timer is waiting! Might happen
    }
    // flag to indicate the stop
    timerThread->delay = -1;
    // stop the wait for the condition
    timerThread->condition.signal();
    timerThread->mutex.unlock();
}

void OpenThreadSimulator::start( bool threaded )
{
    if (!threaded) {
        eventLoop();
        return;
    }
    
    if (startThread->isRunning()) {
        return; // should not happen
    }
    startThread->start();
}

double OpenThreadSimulator::getSystemTime( )
{

// POSIX_TIMERS => nanosec precision
// "Old" unix   => microsec precision, but wrap each day: TODO!
// Windows      => millisec precision

// The configure flag is positionned only if the function is available
#if defined(HAVE_POSIX_TIMERS)

    // Hey, high resolution clock is available. Nice!
    struct timespec ts;
    if (clock_gettime(CLOCK_MONOTONIC, &ts)==0) // 0 is success...
        return ts.tv_sec + ts.tv_nsec * 1e-9;
    // error using high-res clock => use standard clock
#endif

#if defined(_WIN32) || defined(WIN32)
    return timeGetTime() * 1e-3;

#else
    // Fallback for HAVE_POSIX_TIMERS. Too bad, let's use another POSIX call
    struct timeval tv;
    gettimeofday(&tv, 0); // don't care about timezone
    return tv.tv_sec + tv.tv_usec * 1e-6;
#endif

}

