/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "simulator.h"

#ifdef SIMULATION_STATISTICS
#include <iostream>
#endif

using namespace std;
    
/// Maximum number of events the simulator can hold at the same time
const int Simulator::MAX_EVENTS = 1000000;
/// Maximum number of stop events the simulator can handle
const int Simulator::MAX_STOP_EVENTS = 100;

Simulator::GenericDispatcher::GenericDispatcher( )
{
}


Simulator::GenericDispatcher::~ GenericDispatcher( )
{
}

bool Simulator::GenericDispatcher::dispatch( Simulator * simu ) const
{
    // Ugly hack! See comments in stop method!
    simu->stopEvent();
}

#if !defined(USE_SYSTEM_MEMORY_ALLOCATORS)

void* Simulator::GenericDispatcher::memory = 0;
void* Simulator::GenericDispatcher::freeMem = 0;

struct MemMarker {
    int size;
    void* nextFreeMem;
};

void Simulator::GenericDispatcher::allocateMemory( )
{
    // already allocated
    if (memory) return;
    // use max events as a hint. keep room for 3 long arguments, and take half
    // more to cover most cases
    int size = (int)(MAX_EVENTS * (sizeof(GenericDispatcher) + 3*sizeof(double)) * 1.5);
    memory = ::operator new(size);
    freeMem = memory;
    MemMarker* mm = static_cast<MemMarker*>(freeMem);
    mm->size = size;
    mm->nextFreeMem = 0;
}

void * Simulator::GenericDispatcher::operator new( size_t size )
{
    if (!freeMem) throw bad_alloc();
    void* best_fit = freeMem;
    void* best_previous = 0;
    int best_size = static_cast<MemMarker*>(best_fit)->size;
    size += 8; // space for real_size + alignment
    // find best hole to avoid fragmentation
    
    void* previous = freeMem;
    for (void* fit=static_cast<MemMarker*>(freeMem)->nextFreeMem; fit; fit = static_cast<MemMarker*>(fit)->nextFreeMem) {
        // wow, found a hole with exactly the right size
        if (static_cast<MemMarker*>(fit)->size == size) {
            best_fit = fit;
            best_previous = previous;
            break;
        }
        // change if bigger
        if (static_cast<MemMarker*>(fit)->size > best_size) {
            best_fit = fit;
            best_previous = previous;
        }
        // next
        previous = fit;
    }

    MemMarker* mm = static_cast<MemMarker*>(best_fit);
    if (mm->size < size) throw bad_alloc();

    // can't hold a mem marker in the remaining space => fill the gap for real_size
    if ((mm->size - size) < sizeof(MemMarker)) {
        size = mm->size;
        // there is no remaining space, so previous next points to this next
        if (best_previous) static_cast<MemMarker*>(best_previous)->nextFreeMem = mm->nextFreeMem;
        else freeMem = mm->nextFreeMem; // no previous => advance free Mem base
    } else {
        // manage remaining hole
        void* remainingSpace = static_cast<char*>(best_fit) + size;
        MemMarker* remm = static_cast<MemMarker*>(remainingSpace);
        remm->nextFreeMem = mm->nextFreeMem;
        remm->size = mm->size - size;
        // make the previous point to the remaining space
        if (best_previous) static_cast<MemMarker*>(best_previous)->nextFreeMem = remainingSpace;
        else freeMem = remainingSpace; // no previous => advance free Mem base
    }

    // mark real size and return free space
    *static_cast<int*>(best_fit) = size;
    return static_cast<char*>(best_fit)+8;
    
}

void Simulator::GenericDispatcher::operator delete( void *ptr, size_t size )
{
    // get the real base pointer for the deletion
    void* base = static_cast<char*>(ptr)-8;
    // get the real size
    int real_size = *static_cast<int*>(base);
    //MemMarker
    MemMarker* mmbase = static_cast<MemMarker*>(base);
    void* after = static_cast<char*>(base) + real_size;

    // simple case where there was no free mem => no merge, and get back the liberated area
    if (!freeMem) {
        freeMem = base;
        mmbase->size = real_size;
        mmbase->nextFreeMem = 0;
        return;
    }

    // look for potential merge with an existing free mem area
    
    // freeMem pointer points to continuous area directly after deleted area
    // -> merge
    MemMarker* mmfm = static_cast<MemMarker*>(freeMem);
    void* holeNext;
    bool merged = false;
    
    if (freeMem == after) {
        mmbase->size = mmfm->size + real_size;
        mmbase->nextFreeMem = mmfm->nextFreeMem;
        freeMem = base; // one-side merge done
        merged = true;
    }

    // freeMem pointer points to continuous area directly before deleted area
    else if (static_cast<char*>(freeMem)+mmfm->size==base) {
        mmfm->size += real_size; // englobe newly liberated area
        // one-side merge done
        merged = true;
    }

    else for (void* hole=freeMem; hole; hole = holeNext) {
        
        // compute only once in loop
        holeNext = static_cast<MemMarker*>(hole)->nextFreeMem;
        MemMarker* mmhn = static_cast<MemMarker*>(holeNext);
        
        // next pointer points to continuous area directly after deleted area
        // -> merge
        if (holeNext && (holeNext==after)) {
            mmbase->size = mmhn->size + real_size;
            mmbase->nextFreeMem = mmhn->nextFreeMem;
            static_cast<MemMarker*>(hole)->nextFreeMem = base;
            merged = true;
            break; // one-side merge done
        }

        // next pointer points to continuous area directly before deleted area
        if (holeNext && (static_cast<char*>(holeNext)+mmhn->size==base)) {
            mmhn->size += real_size; // englobe newly liberated area
            merged = true;
            break; // one-side merge done
        }
    }

    // No merge done => need to insert new hole at right place, keep order
    if (!merged) {
        mmbase = static_cast<MemMarker*>(base);
        // insert at first place
        if (base<freeMem) {
            mmbase->nextFreeMem = freeMem;
            mmbase->size = real_size;
            freeMem = base;
            return; // done
        }
        // insert at right place, keep order
        for (void* hole=freeMem; hole; hole = holeNext) {
            // compute only once in loop
            holeNext = static_cast<MemMarker*>(hole)->nextFreeMem;
            // insert if at the right place, or at the end
            if ((base<holeNext) || (holeNext==0)) {
                mmbase->nextFreeMem = holeNext;
                mmbase->size = real_size;
                static_cast<MemMarker*>(hole)->nextFreeMem = base;
                return; // done
            }
        }
        // should not reach this point
    }

    // so, there was a one-side merge
    // check if two chunks are not contiguous, by chance
    for (void* hole=freeMem; hole; hole = holeNext) {
        // compute only once in loop
        MemMarker* mmh = static_cast<MemMarker*>(hole);
        holeNext = mmh->nextFreeMem;

        // match => merge
        if (static_cast<char*>(hole)+mmh->size==holeNext) {
            MemMarker* mmhn = static_cast<MemMarker*>(holeNext);
            mmh->nextFreeMem = mmhn->nextFreeMem;
            mmh->size += mmhn->size;
            return; // done
        }
    }
    
}

#endif

#ifdef SIMULATION_STATISTICS
static long noWaitEvents = 0;
static long waitEvents = 0;
#endif

vector<Simulator::AbstractFactory*> Simulator::AbstractFactory::implementations;

Simulator::AbstractFactory::AbstractFactory(int priority, const char* ID) {
    this->priority = priority;
    this->ID = ID;
    implementations.push_back(this);
}

typedef const Simulator::GenericDispatcher* ConstDispatcherPointer;

Simulator::Simulator()
{
    timeReference = 0.0;
    time = 0.0;
    systemTimeReference = 0.0;
    timeFactor = 1.0;
    timeFactorInverse = 1.0;
    isRunningFlag = false;
    events = new Event[MAX_EVENTS];
    nextEvent = firstEvent = events;
    stopEvents = new ConstDispatcherPointer[MAX_STOP_EVENTS];
    numberOfStopEvents = 0;
    currentDispatcherInExecution = 0;
#if !defined(USE_SYSTEM_MEMORY_ALLOCATORS)
    GenericDispatcher::allocateMemory();

#ifdef SIMULATION_STATISTICS
    cout << "Using our own allocators" << endl;
#endif

#else

#ifdef SIMULATION_STATISTICS
    cout << "Using standard allocators" << endl;
#endif

#endif
}


Simulator::~Simulator()
{
}

Simulator * Simulator::newInstance( const char * ID )
{
    int prio = -1;
    AbstractFactory* af = 0;
#ifdef SIMULATION_STATISTICS
    const char *realID = 0;
#endif
    for (vector<AbstractFactory*>::iterator it = AbstractFactory::implementations.begin(); it != AbstractFactory::implementations.end(); ++it) {
#ifdef SIMULATION_STATISTICS
        cout << "Found simulator implementation: " << (*it)->ID << endl;
#endif
        // Honor explicit requests for an ID
        if ((ID!=0) && ((*it)->ID!=0) && !strcmp((*it)->ID, ID)) {
            af = *it;
#ifdef SIMULATION_STATISTICS
            realID = ID;
#endif
            break;
        }

        // look for highest priority
        if ((*it)->priority > prio) {
            prio = (*it)->priority;
            af = *it;
#ifdef SIMULATION_STATISTICS
            realID = (*it)->ID;
#endif
        }
    }

#ifdef SIMULATION_STATISTICS
    if (realID==0) cout << "No simulator implementation found!" << endl;
    else cout << "Using simulator implementation: " << realID << endl;
#endif

    // return highest priority implementation instance
    if (af!=0) return af->factory();
    return 0; // No implementation found!!!
}




void Simulator::flush( )
{
    eventMutexLock();
    nextEvent==firstEvent;
    // reset to the array start so as to avoid
    // a cyclic array whenever possible
    nextEvent = firstEvent = events;
    eventMutexUnlock();
}

const Simulator::GenericDispatcher* Simulator::post(const GenericDispatcher * dispatcher )
{
    eventMutexLock();
    internalPostInsideLock( time, dispatcher );
    eventMutexUnlock();
    return dispatcher;
}

const Simulator::GenericDispatcher* Simulator::postIn( double delay, const GenericDispatcher * dispatcher )
{
    eventMutexLock();
    internalPostInsideLock( time + delay, dispatcher );
    eventMutexUnlock();
    return dispatcher;
}

const Simulator::GenericDispatcher* Simulator::postAt( double absoluteTime, const GenericDispatcher * dispatcher )
{
    eventMutexLock();
    internalPostInsideLock( absoluteTime, dispatcher );
    eventMutexUnlock();
    return dispatcher;
}


void Simulator::internalPostInsideLock( double absoluteTime, const GenericDispatcher * dispatcher )
{

    // In this case, event queue is empty
    // position is OK because out of bounds check is done when event is consumed
    if (nextEvent==firstEvent) {
        nextEvent->dispatcher = dispatcher;
        nextEvent->executionDate = absoluteTime;
        // might be out of the array bounds, check done when adding next event
        nextEvent++;
        waitSemaphorePost();
        return;
    }
    
    // This case happens precisely... but we know the queue is not empty
    if (nextEvent==events+MAX_EVENTS) {
        // Array full, danger!
        // Actually, there would be space for one more element in one case, but at the risk
        // of confusion between first and last element
        // => would effectively treat the list as empty (modulo arithmetic)
        if (firstEvent <= events+1) {
            throw "Can't post : Event array is full!";
        }
        // if the new event is the last one, cycle the array and we're done.
        // otherwise, we'll have to shift at least one elements
        if ((nextEvent-1)->executionDate <= absoluteTime) {
            // use the first array cell
            events->dispatcher = dispatcher;
            events->executionDate = absoluteTime;
            // Position a correct next event
            nextEvent = events + 1;
            waitSemaphorePost();
            return;
        }
        // So, shift this one to the begining of the array, and process array from the end
        *(events) = *(nextEvent-1);
        // Pointer where to store the new event
        Event* event = nextEvent-1;
        // Position a correct next event
        nextEvent = events + 1;
        // loop to find correct position
        while ((event>firstEvent) && ((event-1)->executionDate > absoluteTime)) {
            *(event) = *(event-1);
            event--;
        }
        // In either case, (first element or date match), the element should be inserted
        event->dispatcher = dispatcher;
        event->executionDate = absoluteTime;
        waitSemaphorePost();
        return;
    }

    // Pointer where to store the new event
    Event* event = nextEvent;
    
    // Now, the queue is not empty, and nextEvent is a valid array position
    // Need to search where to add the new event to maintain queue order
    // This is a cyclic queue, so start with the first part of the array if this is the case
    if (firstEvent > nextEvent) {

        // Array full, danger! See comments above
        if (firstEvent == nextEvent+1) {
            throw "Can't post : Event array is full!";
        }

        // Search and shift at the same time
        while ((event>events) && ((event-1)->executionDate > absoluteTime)) {
            *(event) = *(event-1);
            event--;
        }
        
        // Luck : the event should be inserted in the first portion of the array
        // => no need to shift the array end
        if ((event>events) || ((event+MAX_EVENTS-1)->executionDate <= absoluteTime)) {
            event->dispatcher = dispatcher;
            event->executionDate = absoluteTime;
            nextEvent++;
            waitSemaphorePost();
            return;
        }

        // condition about last element in the aray not met => shift it as first element
        *events = *(events+MAX_EVENTS-1);
        event = events+MAX_EVENTS-1; // this element was handled as special case

        // now continue the search & shift in the last part of the array
        // code common with the nicest case below now that last cell is handled correctly
        // nextElement increment is OK even if in the first part of the array
    }

    // Last but not least : nicest case : firstEvent < nextEvent and nextEvent is valid
    // shift all elements while searching for the correct position
    // Pointer where to store the new event
    while ((event>firstEvent) && ((event-1)->executionDate > absoluteTime)) {
        *(event) = *(event-1);
        event--;
    }
    // In either case, (first element or date match), the element should be inserted
    event->dispatcher = dispatcher;
    event->executionDate = absoluteTime;
    waitSemaphorePost();
    nextEvent++; // if out of the array, check will be done next time
}


// TODO: cleanup this mess and really cancel the event!
struct NullDispatch : public Simulator::GenericDispatcher {
    virtual bool dispatch(Simulator* simu) const;
};

bool NullDispatch::dispatch(Simulator* simu) const {
    return false; // no deletion
}

static NullDispatch nullDispatcher;

void Simulator::cancel( const GenericDispatcher * dispatcher )
{
    eventMutexLock();
    if (currentDispatcherInExecution==dispatcher) {
        // As we control event consumption, the only case to get the same event there
        // is if it has already been removed from the queue but not yet dispatched (see step code)
        // In this case, a cyclic event may re-post itself, which is not desireable since we want
        // to cancel it. Treat it as special case here, and let the event destroy itself
        // after its current execution (after all, in simulation time, the event execution date
        // was reached before this cancel, so it has the right to execute one last time).
        // A cyclic event may thus cancel itself too!
        const GenericCyclicDispatcher* gcd = dynamic_cast<const GenericCyclicDispatcher*>(dispatcher);
        if (gcd!=0) gcd->period = -1; // period is mutable!
        // In any case, the dispatcher will now return true and be destroyed in the step method
        eventMutexUnlock();
        return;
    }
    
    // In this case, event queue is empty
    // nothing to cancel
    if (nextEvent==firstEvent) {
        eventMutexUnlock();
        return;
    }

    // pointer to start search
    Event* event = firstEvent;
    
    // Now, the queue is not empty
    // Need to search for the event
    // This is a cyclic queue, so start with the first part of the array if this is the case
    if (firstEvent > nextEvent) {
        Event* lastEvent = events+MAX_EVENTS;
        for (; event < lastEvent; ++event) {
            if (event->dispatcher==dispatcher) {
                event->dispatcher=&nullDispatcher;
                delete dispatcher;
                eventMutexUnlock();
                return;
            }
        }
        event = events;

        // now continue the search & shift in the last part of the array
        // code common with the nicest case below now that last cell is handled correctly
        // nextElement increment is OK even if in the first part of the array
    }

    // Last but not least : nicest case : firstEvent < nextEvent
    for (; event < nextEvent; ++event) {
        if (event->dispatcher==dispatcher) {
            event->dispatcher=&nullDispatcher;
            delete dispatcher;
            eventMutexUnlock();
            return;
        }
    }
    eventMutexUnlock();
}




bool Simulator::step( )
{
    // Only one object can control the simulation at the same time
    // No concurrent steps!
    if (!stepControlMutexTryLock()) return false;
    eventMutexLock();
    const GenericDispatcher * dispatcher = internalStepInsideLock( );
    currentDispatcherInExecution = dispatcher;
    eventMutexUnlock();

    if (dispatcher!=0) {
        // process the event outside the lock
        if (dispatcher->dispatch(this)) {
            eventMutexLock();
            currentDispatcherInExecution = 0;
            // Event consumed => cleanup
            delete dispatcher;
            eventMutexUnlock();
        }
        stepControlMutexUnlock();
        return true;
    }
    stepControlMutexUnlock();
    return false;
}


bool Simulator::step( double absoluteTime )
{
    // Only one object can control the simulation at the same time
    // No concurrent steps!
    if (!stepControlMutexTryLock()) return false;
    // So, let's note when we are for later correspondance
    eventMutexLock();
    timeReference = time;
    systemTimeReference = getSystemTime();
    // monitor stop events    
    isRunningFlag = true;
    eventMutexUnlock();
    

    // Call the real stuff now that we control the simulation
    bool ret = internalStepInsideControl(absoluteTime);

    // Stop in any case, this is not a start() function
    eventMutexLock();
    isRunningFlag = false;
    eventMutexUnlock();

    stepControlMutexUnlock();

    return ret;
}

bool Simulator::internalStepInsideControl( double absoluteTime )
{
    // Just process events so long as the event queue is not empty,
    // and the next event is in the range to process
    // This is basically a while loop with locking taken into account
    // while ((nextEvent!=firstEvent) && (nextEvent->executionDate <= absoluteTime))  step();
    // current time is thus advanced to the given time

    const GenericDispatcher * dispatcher = 0;
    // Will be set to true if an event exists, remain to false otherwise
    bool ret = false;
    while (true) {

        eventMutexLock();
        if ((nextEvent!=firstEvent) && (firstEvent->executionDate <= absoluteTime)) {
            dispatcher = internalStepInsideLock( );
            currentDispatcherInExecution = dispatcher;
            ret = true;
        }
        // Tricky time advance because this must happen in the locked section,
        // otherwise there may be pre-emption by another post/step...
        // Note that this second if condition is not related to
        // the first if, because the step may have changed it
        // (ex: repost and nextEvent no longer equal to firstEvent)
/*        if ((nextEvent==firstEvent) || (firstEvent->executionDate > absoluteTime)) {
            // Always advance to what what requested in the call at the end
            time = absoluteTime;
        }
*/
        eventMutexUnlock();
        
        // exit loop when there is no more event
        if (dispatcher==0) break;
        // process the event outside the lock so it can post itself again if wished
        if (dispatcher->dispatch(this)) {
            // destroy event unless stated otherwise (cyclic...)
            eventMutexLock();
            currentDispatcherInExecution = 0;
            // Event consumed => cleanup
            delete dispatcher;
            eventMutexUnlock();
        }
        // Reset loop exit condition in case if test condition is not satisfied
        dispatcher=0;

        // exit loop when the event was a stop event
        eventMutexLock();
        if (!isRunningFlag) {
            eventMutexUnlock();
            break;
        }
        eventMutexUnlock();
    }
    // Always advance to what what requested in the call at the end
    time = absoluteTime;
    return ret;
}

const Simulator::GenericDispatcher * Simulator::internalStepInsideLock( )
{
    // nothing to do when the event queue is empty
    if (nextEvent==firstEvent) return 0;

    // Process the first event in the queue
    const GenericDispatcher * ret = 0;
    
    // Advance current time to the execution date, if this is not a priority event
    if (firstEvent->executionDate>=0) time = firstEvent->executionDate;
    //else cout << "priority event!" << endl;
    
    // return dispatcher to process the event
    ret = firstEvent->dispatcher;

    // Next event is first event...
    firstEvent++;
    // ...except when there is no next event !
    if (nextEvent==firstEvent) {
        // in this case, reset to the array start so as to avoid
        // a cyclic array whenever possible
        nextEvent = firstEvent = events;
        return ret;
    }
    
    // Sanity check
    if (firstEvent==events+MAX_EVENTS) firstEvent = events;
    return ret;
}

double Simulator::getTime( ) const
{
    return time;
}


void Simulator::setTimeFactor( double factor )
{
    // rate change is from now on, it is not applied on previous values
    // => translate the time reference so that now is the new reference
    eventMutexLock();
    // no event may happen while doing the translation
    systemTimeReference = getSystemTime();
    timeReference = time;
    if (factor>0) {
        timeFactor = factor;
        timeFactorInverse = 1.0 / factor;
    }
    else {
        timeFactor = 0.0; // thus no
        timeFactorInverse = -1.0; // indicator
    }
    eventMutexUnlock();
}

double Simulator::getTimeFactor( ) const
{
    return timeFactor;
}


void Simulator::resetTime( )
{
    eventMutexLock();
    // translate events
    // cyclic array
    if (nextEvent<firstEvent) {
        Event* event ;
        for (event = events; event!=nextEvent; ++event) event->executionDate -= time;
        Event* max_event = events + MAX_EVENTS;
        for (event = firstEvent; event!=max_event; ++event) event->executionDate -= time;
    } else {
        // also treat the case of empty queue
        for (Event* event = firstEvent; event!=nextEvent; ++event) event->executionDate -= time;
    }
    time = 0.0;
    timeReference = 0.0;
    systemTimeReference = getSystemTime();
    eventMutexUnlock();
}


void Simulator::eventLoop( )
{
    // Only one object can control the simulation at the same time
    // No concurrent steps!
    // Wait till all pending steps are done
    stepControlMutexLock();

    // So, let's note when we are for later correspondance
    eventMutexLock();
    timeReference = time;
    systemTimeReference = getSystemTime();
    isRunningFlag = true;
    eventMutexUnlock();
    
        
    const GenericDispatcher * dispatcher = 0;
    bool hasPendingEvent = false;

    // stop on isRunningFlag false
    while (true) {
        
        double currentDate;
        
        // process all events until system time is reached
        // or an explicit stop event is reached
        // more events may be posted between 2 loops, but no event may be consumed
        // since we have the control mutex

        // exit loop when the event is a stop event, or when there is no event
        while (true) {
            eventMutexLock();
            currentDate = (getSystemTime() - systemTimeReference) * timeFactor + timeReference;
            if ((nextEvent!=firstEvent) && (firstEvent->executionDate <= currentDate)) {
                dispatcher = internalStepInsideLock( );
                currentDispatcherInExecution = dispatcher;
                eventMutexUnlock();
            }
            // exit loop when there is no more event before current date
            else {
                eventMutexUnlock();
                break;
            }
            
            // exit loop when there is no more event.
            // dispatcher was set, or else statement has broken loop
            if (dispatcher==0) break;
            // process the event outside the lock so it can post itself again if wished
            if (dispatcher->dispatch(this)) {
                eventMutexLock();
                currentDispatcherInExecution = 0;
                // Event consumed => cleanup
                delete dispatcher;
                eventMutexUnlock();
            }

#ifdef SIMULATION_STATISTICS
            noWaitEvents++;
#endif
            // exit loop on stop event condition
            eventMutexLock();
            if (!isRunningFlag) {
                eventMutexUnlock();
                break;
            }
            eventMutexUnlock();
        }

        // returning from this function should terminate the thread it is run in, if any
        eventMutexLock();
        
        // exit loop on stop event condition
        if (!isRunningFlag) {
            eventMutexUnlock();
            break;
        }
    
        // initialized inside test
        double nextDate;
        bool hasNextDate = false;

        // need to be put in protected section for consistency between the test
        // and the affectation
        if (nextEvent!=firstEvent) {
            nextDate = firstEvent->executionDate;
            currentDate = (getSystemTime() - systemTimeReference) * timeFactor + timeReference;
            if (nextDate<=currentDate) {
                // An event was posted in the short time between last test and the  
                // one in this protected section, or the next event was too close
                // => process it without waiting
                eventMutexUnlock();
                continue;
            }
            if (timeFactorInverse != -1.0) hasNextDate = true; // flag for later use
        }


        // The queue is empty in protected section => at this point, no event can be
        // posted and the queue will remain empty
        
        // Flush semaphore inside protected section => guaranty the semaphore has 0 
        // count since only post events increase the semaphore
        waitSemaphoreFlush();

        
        // At this point, we have empty queue and a 0 count semaphore.
        // => Only posting a new event can unlock the semaphore.

        // An helper function is provided to be called back by the timer.
        // @see waitTimerCallBack
        // But start the timer only when there is a next event some time in the future,
        // otherwise, the semaphore will just block till an event is posted
        if (hasNextDate) {
            // update current date to the last moment possible, to have a chance not to wait for nothing
            currentDate = (getSystemTime() - systemTimeReference) * timeFactor + timeReference;
            if (nextDate>currentDate) startWaitTimer((nextDate - currentDate) * timeFactorInverse);
            else {
                eventMutexUnlock();
                continue;
            }
        }
        
        eventMutexUnlock();

        // Wait outside the protected section, so posts are enabled during the wait.
        // Of course, if en event was posted in the short time between the end of the
        // protected section and this point, the semaphore is non zero and no wait occurs
        waitSemaphoreWait();
#ifdef SIMULATION_STATISTICS
        waitEvents++;
#endif

        // Whatever the case, the timer should be stopped now if it is still present
        // It doesn't matter if the synchronization is not exact and it has just expired
        // and increased the semaphore after the semaphore wait, because it will be
        // flushed anyway. On the other hand, the timer must be deleted before next
        // semaphore flush so an old timer does not mess up with the new delay
        cancelWaitTimer();

        // if the wait was cancelled by posting a stop event, then the main loop
        // will terminate soon
    }

    // allow another object to do manual steps, or start again
    stepControlMutexUnlock();

    // execute stop events outside the protected section => one of them may
    // trigger the loop again, blocking the others
    while (numberOfStopEvents>0) {
        if (stopEvents[--numberOfStopEvents]->dispatch(this))
            // Event consumed => cleanup
            delete stopEvents[numberOfStopEvents];
    }
}

void Simulator::waitTimerCallBack( )
{
    // this is enough, precaution was taken in the event loop
    // in case this happens while also posting an event
    waitSemaphorePost();
}


void Simulator::stop( bool waitForStop )
{
    // don't stop if already stopped!
    eventMutexLock();
    if (!isRunningFlag) {
        eventMutexUnlock();
        // Honor request to wait for event loop to finish.
        if (waitForStop) {
            // The lock will wait for the event processing loop to finish
            stepControlMutexLock();
            // Release the lock now we have it!
            stepControlMutexUnlock();
        }
#ifdef SIMULATION_STATISTICS
cout << "waitEvents:" << waitEvents << ", noWaitEvents:" << noWaitEvents << ", diff=" << noWaitEvents - waitEvents << ", ratio=" << (double)(noWaitEvents - waitEvents) / (noWaitEvents==0 ? 1.0 : (double)noWaitEvents) << endl;
#endif
        return;
    }
    eventMutexUnlock();
    
    // Post event at really the beginning of the queue
    // For this, use negative time "priority event"

    // either in the event catchup section of the event loop, or in
    // the wait section, a post will unlock the event loop
    // stop events are also recognized by the step to absolute time function
    postAt(-1.0, this, &Simulator::stopEvent);

    // problem: there is a bug somewhere (gcc, ld?)
    // and the GenericDispatcher virtual is called instead of the specialized
    // version. This does not happen when posting from outside this class!!!
    // => hack the generic class to do what we want...

    if (waitForStop) {
        // The lock will wait for the event processing loop to finish
        stepControlMutexLock();
        // Release the lock now we have it!
        stepControlMutexUnlock();
    }

#ifdef SIMULATION_STATISTICS
cout << "waitEvents:" << waitEvents << ", noWaitEvents:" << noWaitEvents << ", diff=" << noWaitEvents - waitEvents << ", ratio=" << (double)(noWaitEvents - waitEvents) / (noWaitEvents==0 ? 1.0 : (double)noWaitEvents) << endl;
#endif
}

void Simulator::stopEvent( )
{
    eventMutexLock();
    isRunningFlag = false;
    eventMutexUnlock();
}

bool Simulator::isRunning( )
{
    return isRunningFlag;
}


const Simulator::GenericDispatcher* Simulator::postAtStop( const GenericDispatcher * dispatcher )
{
    if (numberOfStopEvents>=MAX_STOP_EVENTS) throw "Can't post at Stop: Event array is full!";
    stopEvents[numberOfStopEvents++] = dispatcher;
    return dispatcher;
}


// ================ METHODS DESIGNED TO BE OVERRIDEN ===================


void Simulator::start( bool threaded )
{
    // Please override the start function to create a thread in your thread system
    // The default implementation blocks
    eventLoop();
}


double Simulator::getSystemTime( )
{
    // this will force events to be processed without waiting,
    // thus the wait semaphore and timer functions won't be used
    // in this default implementation
    return 0.0;
}


void Simulator::eventMutexLock( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::eventMutexUnlock( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::waitSemaphorePost( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::waitSemaphoreFlush( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::waitSemaphoreWait( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::startWaitTimer( double delay )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::cancelWaitTimer( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::stepControlMutexLock( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

void Simulator::stepControlMutexUnlock( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

bool Simulator::stepControlMutexTryLock( )
{
    // Please override this function for a particular system you use (SDL, Qt, posix, ...)
}

