/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef SIMULATOR_H
#define SIMULATOR_H

#include <memory>
#include <vector>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/**
Manager of the simulation, this class is responsible for handling simulation events

simulatedTime
 |factor
systemTime

simut = (syst - syst_ref) * factor + simut_ref

@author Nicolas Brodu
*/
class Simulator {
public:

    /// Maximum number of events the simulator can hold at the same time
    static const int MAX_EVENTS;
    /// Maximum number of stop events the simulator can handle
    static const int MAX_STOP_EVENTS;

// ===================== Dispatcher definitions ===========================
#include "simulator_dispatcher_templates.h"


// ======================= Public interface ===============================
public:

    /** Gets a new simulator instance for your particular system.
        Depending on the current system environment, more
        than one simulator implementation may be available.

        You can choose a specific implementation if you know its ID,
        or let the implementor decide which is the best implementation.
        
        This is the default, and silent fallback when the ID does not exist.
        If you really wish to check on a specific ID, you may consider dynamic
        casting the result to the type for this ID. You may also simply
        call the constructor of that implementation!

        @param ID A simulator implementation ID, or 0 to choose the default
     */
    static Simulator* newInstance(const char* ID = 0);
    
    /** Cancels the selected event */
    virtual void cancel(const GenericDispatcher* dispatcher);
    
    /** Flush all posted events without executing them
     */
    virtual void flush();
    
    /** Post an event at current time. Use this function together with a specialized
        dispatcher to post methods with more than 3 arguments.
        Note: Template methods are provided, included postCyclic methods.
        Note2: You can use this method with a cyclic dispatcher too.
        @param dispatcher Callback to the function or object method to call
     */
    virtual const GenericDispatcher* post(const GenericDispatcher* dispatcher);

    /** Post an event at current time.
        Any C "void method()"
        can be called this way.
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* post(void (*method)()) {
        return post(new StaticDispatcher(method));
    }

    /** Post a cyclic event at current time. 
        Any C "void method()"
        can be called this way.
        @param cycle  how often the method should be called
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postCyclic(double cycle, void (*method)()) {
        return post(new StaticCyclicDispatcher(cycle, method));
    }

    
    /** Post an event at current time + a certain delay in simulated time
        Use this function together with a specialized dispatcher to post
        methods with more than 3 arguments, or to pass static methods
        and global functions
        Note: Template methods are provided, included postCyclic methods.
        Note2: You can use this method with a cyclic dispatcher too.
        @param delay after current time, should be >0.
        @param dispatcher Callback to the function or object method to call
        @throw Exception when delay <0
     */
    virtual const GenericDispatcher* postIn(double delay, const GenericDispatcher* dispatcher);
    
    /** Post an event at current time + a certain delay in simulated time
        Any C "void method()"
        can be called this way.
        @param delay after current time, should be >0.
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postIn(double delay, void (*method)()) {
        return postIn(delay, new StaticDispatcher(method));
    }
    
    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C "void method()" can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postCyclicIn(double delay, double cycle, void (*method)()) {
        return postIn(delay, new StaticCyclicDispatcher(cycle, method));
    }

    
    /** Post an event at an absolute simulated time
        Use this function together with a specialized dispatcher to post
        methods with more than 3 arguments, or to pass static methods
        and global functions
        Note: Template methods are provided, included postCyclic methods.
        Note2: You can use this method with a cyclic dispatcher too.
        @param absoluteTime time at which to execute the event
        @param dispatcher Callback to the function or object method to call
        @throw Exception when the time is already passed
     */
    virtual const GenericDispatcher* postAt(double absoluteTime, const GenericDispatcher* dispatcher);

    /** Post an event at an absolute simulated time. Any C "void method()"
        can be called this way.
        @param absoluteTime time at which to execute the event
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postAt(double absoluteTime, void (*method)()) {
        return postAt(absoluteTime, new StaticDispatcher(method));
    }

    /** Post a cyclic event at an absolute simulated time. Any C "void method()"
        can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, void (*method)()) {
        return postAt(absoluteTime, new StaticCyclicDispatcher(cycle, method));
    }

    
    
    /** Process next event in the queue, if possible
     @return true if there was an event and it could be processed, false otherwise
     */
    virtual bool step();
    
    /** Process next events in the queue, if possible, till the given simulated time is reached
     @param absoluteTime The simulated time to reach
     @return true if there was at least an event and it could be processed, false otherwise
     */
    virtual bool step(double absoluteTime);

    
    /**
      Correspondance between once unit of time in system time and one unit in
      simulated time.
      Default is a factor of 1.0, meaning the simulation will try to run in
      soft real-time when start()ed. Higher factors mean faster than real-time,
      lower factors mean slower than real-time.
      

      <br>
      This function can be called at any moment, it will affect the next event.
      See also the @see start function. Setting a factor of 0 means "process events as
      fast as possible"<br>
      See also the @see getSystemTime function. Returning values for units a different
      than a second is another way to modify the range of the simulated time.<br>
      See also @see getSimulatedTime, @see getTimeReference and
      @see getSimulatedTimeReference, for more details how a change in time factor
      affects the correspondance between simulated time and internal time
      @param factor The relation between one unit(second) of system time and the
      time parameter, more precisely 1_unit_internal = 1_unit_system * timeFactor.
      If set to 0, then events are processed as fast as possible.
      Default is 1e-6 for microseconds.
      */
    virtual void setTimeFactor(double factor);

    /** @see setTimeFactor */
    virtual double getTimeFactor() const;

    /**
      Gets the current simulated time since the simulation started or the last reset.
      The simulated time may not be the same at all as the system time, even if a
      factor of 1.0 is set. Indeed, the simulation may have
      stopped and restarted (operator takes a coffee break), or
      the system may not be able to process the events fast enough to
      keep up with real-time. In any case, both share the same time unit, usually
      seconds but this maybe changed by overloading the getSystemTime function.
      @Note This function can be used any time, even
      when events are pending. Be aware that it
      is not thread-synchronized, so multiple runs of the
      same program may not return the same values. This was done on purpose
      for many reasons, good ones including not blocking
      the step and post functions, and also because most of
      the time thread-synchronization would be completely useless.
      Indeed, there is no way to guaranty that an event won't
      be processed during 2 consecutive unprotected instructions
      in your program. Rather than protecting only the getTime function,
      you would need to put it together with the other instructions
      inside a mutex protected block, which cleary cannot be done
      automatically for you. If this is what you really, really want, 
      then think twice about it and see if posting an event at the
      right date (possibly with a 0 delay) doesn't help. If it is really
      not the case, overload this class to make the eventMutexLock/Unlock
      functions accessible, and you're on your own.
     */
    virtual double getTime() const;
    
    /** Resets the current simulated time parameter to 0.
        Pending events are translated so they keep the same delay
        between current date and activation date.<br>
        This is especially useful before a simulation start() after a stop(),
        when no events are pending, to start a completely new simulation.
     */
    virtual void resetTime();
    
    /** Stop processing events automatically. You can use the step functions to
        advance in time.
        @param waitForStop if the function should wait for the last event to be processed
        The event posted at stop will be process after the function returns, though, in case
        one of them restarts the loop.
        @see comments of the start function about implementations.
     */
    virtual void stop(bool waitForStop = false);


    
    /** Post an event to be executed when the simulation stops.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        Use this function together with a specialized dispatcher
        to post methods with more than 3 arguments, or to pass
        static methods and global functions
        @param dispatcher Callback to the function or object method to call
     */
    virtual const GenericDispatcher* postAtStop(const GenericDispatcher* dispatcher);
    
    /** Post an event to be executed when the simulation stops.
        Any C "void method()" can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param method any C "void method()" with no argument
     */
    inline const GenericDispatcher* postAtStop(void (*method)()) {
        return postAtStop(new StaticDispatcher(method));
    }
    
// ================ PUBLIC METHODS DESIGNED TO BE OVERRIDEN ===================

    /**
      Start processing events at a rate determined by the time factor.
      This function returns immediately by default, and process the events in another thread.
      You may change this by setting the argument to false
      Of course, there is a limit at how fast you can process events, that is how
      small you can set the time factor for a fixed amount of events, depending
      on your system performances. Below this value, you'll probably see a large drift
      between simulated time and system time. In this case, you can gain an extra
      bit of performance by setting the factor to 0, which means "process the
      events as fast as possible".
      @note The default implementation does actually NOT start a thread, but blocks.
      Threads are system dependant. A helper protected function eventLoop() is
      provided so you can easily override this class for your particular thread
      system, if this is not done already.
      @param threaded whether a new thread should be started to process the event loop. Default is true
     */
    virtual void start(bool threaded = true);

    /// Returns true if, and only if, the simulator is running
    virtual bool isRunning();


    // C++ artifact
    virtual ~Simulator();


    /// See struct Factory
    struct AbstractFactory {
        int priority;
        const char *ID;
        AbstractFactory(int priority, const char* ID);
        virtual Simulator* factory() = 0;
        virtual ~AbstractFactory() {};
        static std::vector<AbstractFactory*> implementations;
    };
    
    /** Subclasses should provide an ID and a priority to compare this
        implementation with the default ones
        
        Just declare a global object with the class as template argument,
        this is enough (it will be initialized before execution).

        The highest priority implementation is chosen as default by getInstance.
        See also the Posix and OSG implementations (resp priority 100 and 50).
     */
    template <class T> struct Factory : public AbstractFactory {
        Simulator* factory() { return new T; }
        Factory(int priority, const char* ID) : AbstractFactory(priority,ID) {}
    };

#define IMPLEMENT_SIMULATOR_FACTORY(priority,name) \
Simulator::Factory<name> name ## _factory(priority, #name);
        
// ============== PROTECTED SECTION =============================    
    
protected:
    // the references are necessary to keep track of factor change, for example
    double time, timeReference, systemTimeReference;
    // correspondance between internal and simulated time
    double timeFactor, timeFactorInverse;

    
    /** this flag is used by stop events to stop the event loop
       it should never be set to true outside the event loop
       but can be set to false to stop it
    */
    bool isRunningFlag;

    // Defined in the class Namespace, no clash
    struct Event {
        double executionDate;
        const GenericDispatcher* dispatcher;
    };

    // all events, implemented as a cyclic array for efficiency (no dynamic memory allocation)
    // Better shift a few array elements when inserting an event between others, than allocating
    // memory at run-time.
    Event *events;
    Event * firstEvent, *nextEvent;

    // stop events
    const GenericDispatcher** stopEvents;
    int numberOfStopEvents;

    // manage special case where an event is being executed and cancels itself
    // in this case, special care should be taken to avoid double destruction
    // this variable is only accessed in event mutex protected sections
    const GenericDispatcher* currentDispatcherInExecution;
    
    /** This common portion is called by all post functions
        It is necessary separated so as to put reading the current time in the lock
        for the 2 simple post functions.
     */
    virtual void internalPostInsideLock(double absoluteTime, const GenericDispatcher* dispatcher);

    /** Common portion between the 2 step functions
        This code is designed to be executed inside a lock, and it is very important that
        the returned dispatcher is called oustide the lock, because the called function
        may re-post itself for cyclic activation and will thus need a lock during post
        @return a dispatcher to call for the current step, outside the lock, or null if there is no current step.
     */
    virtual const GenericDispatcher* internalStepInsideLock();

    /** Common portion between the step to absolute date and the event loop
        This code is designed to be executed in a section protected from concurrent
        simulation control 
     */
    virtual bool internalStepInsideControl(double absoluteTime);
    
// ======= METHODS DESIGNED TO BE CALLED BACK BY OVERRIDEN METHODS =====
    /**
      To be run in a thread, should be used by the start function
     */
    virtual void eventLoop();

    /** This stop event is posted by the stop method to cleanly
    exit from the event loop*/
    virtual void stopEvent();
    
    /**
      Must be called when the wait timer expires
     */
    virtual void waitTimerCallBack();


    /// protected constructor: User API is the public newInstance, or a public constructor
    /// of a subtype
    Simulator();

// ================ METHODS DESIGNED TO BE OVERRIDEN ===================
public:
    /** Override this method to return the current system time,
        expressed in system time units (usually seconds).
        @note This function is <b>only</b> used for computing
        differences, so the time origin doesn't matter.
        This allows you to use system clocks using any time reference
        mechanism, like the POSIX monotonic high-resolution clocks.
        @see getTime()
     */
    virtual double getSystemTime();

protected:
    /**
       Override this method to lock the mutex that will be used for event related functions.
       The function should block as expected until the mutex is available
       No locking by default, see CPP.
    */
    virtual void eventMutexLock();
    
    /**
       Override this method to unlock the mutex that will be used for event related functions
       No locking by default, see CPP.
    */
    virtual void eventMutexUnlock();

    /**
       Override this method to try locking a mutex that will be used for controlling
       the simulation. Only one thread can call step, or the event loop at the same time.
       The function should return true on success, or return immediately false otherwise
       No locking by default, see CPP.
    */
    virtual bool stepControlMutexTryLock();

    /**
       Override this method to locking the mutex that will be used for controlling
       the simulation. Only one thread can call step, or the event loop at the same time.
       The function should block as expected until the mutex is available
       No locking by default, see CPP.
    */
    virtual void stepControlMutexLock();
    
    /**
       Override this method to unlock the mutex that will be used for controlling
       the simulation.
    */
    virtual void stepControlMutexUnlock();
    
    /**
       Override this method to increment a semaphore that will be used
       for waiting for a new event in the event loop
    */
    virtual void waitSemaphorePost();
    
    /**
       Override this method to flush a semaphore that will be used
       for waiting for a new event in the event loop.
       By flushing, we mean that the semaphore count should be zero after
       this call. This call is done in a protected section such that no post
       will be done to the semaphore while calling this function. If your
       system doesn't propose better, you might thus try creating a new
       semaphore with 0 count.
    */
    virtual void waitSemaphoreFlush();
    
    /**
       Override this method to implement the "standard wait till semaphore
       count >0" call on a semaphore that will be used for waiting
       for a new event in the event loop
    */
    virtual void waitSemaphoreWait();

    /**
        Override this method to start some kind of timer, that will call the
        timer callback provided after the given delay. The delay is expressed 
        in system time units, so implementations don't need to use the time factor.
        If your system does not provide timers, an implementation may be to wait
        in another thread for either some amount of time or some unblocking
        condition : @see cancelWaitTimer().
        @param delay the time to wait before calling the timer callback,
        expressed in system time units as used by the getSystemTime method.
     */
    virtual void startWaitTimer(double delay);
    
    /**
        Override this method to immediately stop the timer started by the
        startWaitTimer method. If you system does not provide such an operation,
        a clean implementation may be waiting for either a timout or a condition
        in a separate thread, and this cancel function triggers the condition.
        Then, by testing which of the timer or condition stopped the waiting, you
        can either call the timer callback or return immediately.
        @see startWaitTimer
    */
    virtual void cancelWaitTimer();

    
    
};



#endif

