/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**

This file is intended to be included directly inside the simulator class.
It has been separated for the sake of clarity of the simulator interface.

TODO: Generate this file automatically (with perl, etc.) to ease maintainance
*/


public:

// make it our friend so it can access MAX_EVENTS as a hint.
friend class GenericDispatcher;

/**
    This base class of object method wrappers is put in the public namespace of the
    Simulator so it can be used with the generic post methods.
    Most probably you won't need it for normal operation.
*/
class GenericDispatcher {
friend class Simulator; // yep, simulator pre-allocates memory before run-time.
public:
    // avoid templates from creating a weak symbol for this
    GenericDispatcher();

    /**
        Dispatch the event. A reference to the managing simulator is provided
        return true if the event is consumed and should be deleted, false
        if is is re-used
     */
    virtual bool dispatch(Simulator* simu) const;

    // necessary for operator delete to work correctly
    virtual ~GenericDispatcher();
    
#if !defined(USE_SYSTEM_MEMORY_ALLOCATORS)
    // Use a better new and delete to avoid real-time dynamic memory allocation
    // Since events are stored till used at a date
    // it is possibly to pre-allocate a big buffer, store the events
    // one by one, and deallocate them when used.
    // In this system, there is memory fragmentation when an event for date t+d (with d>0)
    // is stored before an event for date t. When the event at date t is deallocated, there is
    // a hole. Fortunately, the memory is automatically defragmented when t+d is reached and
    // the event at that date is released. 
    void* operator new(size_t);
    void operator delete(void*, size_t);
    //void* operator new[](size_t);
    //void operator delete[](void*, size_t);
protected:

    static void allocateMemory();
    // big memory, allocated once and re-used for fast object creation
    // at run-time
    static void* memory;
    static void* freeMem;
#endif
};

struct GenericCyclicDispatcher : public GenericDispatcher {
    mutable double period;
    GenericCyclicDispatcher(double p) : period(p) {}
};

/* TODO: Self cancelling objects maintain a list of all events using them and
when they are destroyed cancel all corresponding simulation events.
Implement this effect using specialized templates for which T is of type
SelfCancellingObject
struct SelfCancellingObject {
    
    virtual ~SelfCancellingObject();
};

TODO2: rather than the above, introduce a cancelAllEventsFor(void* object)
and provide a checkUsing(void*) function in generic dispatcher, overloaded
by the specific dispatchers. Then, the simulator can ask all generic
dispatchers for the posted events if they use the given object and remove
them if this is the case
*/


// Necessary for C++ contravariance rule for pointers to members
template<class T> struct Dispatcher : public GenericDispatcher {
    typedef void (T::* TFunc)(); // pointer to member type
    T* pointer;
    TFunc function;
    inline Dispatcher(T* t, TFunc f) : pointer(t), function(f) {}
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)();
        return true;
    }

};

template<class T> struct CyclicDispatcher : public GenericCyclicDispatcher {
    typedef void (T::* TFunc)(); // pointer to member type
    T* pointer;
    TFunc function;
    inline CyclicDispatcher(double p, T* t, TFunc f) : GenericCyclicDispatcher(p), pointer(t), function(f) {}
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)();
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }

};

template<class T, class A> struct Dispatcher1 : public GenericDispatcher {
    typedef void (T::* TFunc)(A a); // pointer to member type with one argument
    T* pointer;
    TFunc function;
    A arg;
    inline Dispatcher1(T* t, TFunc f, A a) : pointer(t), function(f), arg(a) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg);
        return true;
    }
};

template<class T, class A> struct CyclicDispatcher1 : public GenericCyclicDispatcher {
    typedef void (T::* TFunc)(A a); // pointer to member type with one argument
    T* pointer;
    TFunc function;
    A arg;
    inline CyclicDispatcher1(double p, T* t, TFunc f, A a) : GenericCyclicDispatcher(p), pointer(t), function(f), arg(a) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};

template<class T, class A1, class A2> struct Dispatcher2 : public GenericDispatcher {
    typedef void (T::* TFunc)(A1,A2); // pointer to member type with two arguments
    T* pointer;
    TFunc function;
    A1 arg1;
    A2 arg2;
    inline Dispatcher2(T* t, TFunc f, A1 a1, A2 a2) : pointer(t), function(f), arg1(a1), arg2(a2) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg1,arg2);
        return true;
    }
};

template<class T, class A1, class A2> struct CyclicDispatcher2 : public GenericCyclicDispatcher {
    typedef void (T::* TFunc)(A1,A2); // pointer to member type with two arguments
    T* pointer;
    TFunc function;
    A1 arg1;
    A2 arg2;
    inline CyclicDispatcher2(double p, T* t, TFunc f, A1 a1, A2 a2) : GenericCyclicDispatcher(p), pointer(t), function(f), arg1(a1), arg2(a2) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg1,arg2);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};

template<class T, class A1, class A2, class A3> struct Dispatcher3 : public GenericDispatcher {
    typedef void (T::* TFunc)(A1,A2,A3); // pointer to member type with three arguments
    T* pointer;
    TFunc function;
    A1 arg1;
    A2 arg2;
    A3 arg3;
    inline Dispatcher3(T* t, TFunc f, A1 a1, A2 a2, A3 a3) : pointer(t), function(f), arg1(a1), arg2(a2), arg3(a3) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg1,arg2,arg3);
        return true;
    }
};

template<class T, class A1, class A2, class A3> struct CyclicDispatcher3 : public GenericCyclicDispatcher {
    typedef void (T::* TFunc)(A1,A2,A3); // pointer to member type with three arguments
    T* pointer;
    TFunc function;
    A1 arg1;
    A2 arg2;
    A3 arg3;
    inline CyclicDispatcher3(double p, T* t, TFunc f, A1 a1, A2 a2, A3 a3) : GenericCyclicDispatcher(p), pointer(t), function(f), arg1(a1), arg2(a2), arg3(a3) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (pointer->*function)(arg1,arg2,arg3);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};

// plain C callbacks
struct StaticDispatcher : public GenericDispatcher {
    typedef void (*Func)(); // pointer to static func
    Func function;
    inline StaticDispatcher(Func f) : function(f) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)();
        return true;
    }

};

struct StaticCyclicDispatcher : public GenericCyclicDispatcher {
    typedef void (*Func)(); // pointer to member type
    Func function;
    inline StaticCyclicDispatcher(double p, Func f) : GenericCyclicDispatcher(p), function(f) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)();
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }

};

template<class A> struct StaticDispatcher1 : public GenericDispatcher {
    typedef void (*Func)(A a); // pointer to C function with one argument
    Func function;
    A arg;
    inline StaticDispatcher1(Func f, A a) : function(f), arg(a) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg);
        return true;
    }
};

template<class A> struct StaticCyclicDispatcher1 : public GenericCyclicDispatcher {
    typedef void (*Func)(A a); // pointer to member type with one argument
    Func function;
    A arg;
    inline StaticCyclicDispatcher1(double p, Func f, A a) : GenericCyclicDispatcher(p), function(f), arg(a) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};

template<class A1, class A2> struct StaticDispatcher2 : public GenericDispatcher {
    typedef void (*Func)(A1,A2); // pointer to C function with two arguments
    Func function;
    A1 arg1;
    A2 arg2;
    inline StaticDispatcher2(Func f, A1 a1, A2 a2) : function(f), arg1(a1), arg2(a2) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg1,arg2);
        return true;
    }
};

template<class A1, class A2> struct StaticCyclicDispatcher2 : public GenericCyclicDispatcher {
    typedef void (*Func)(A1,A2); // pointer to C function with two arguments
    Func function;
    A1 arg1;
    A2 arg2;
    inline StaticCyclicDispatcher2(double p, Func f, A1 a1, A2 a2) : GenericCyclicDispatcher(p), function(f), arg1(a1), arg2(a2) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg1,arg2);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};

template<class A1, class A2, class A3> struct StaticDispatcher3 : public GenericDispatcher {
    typedef void (*Func)(A1,A2,A3); // pointer to C function with three arguments
    Func function;
    A1 arg1;
    A2 arg2;
    A3 arg3;
    inline StaticDispatcher3(Func f, A1 a1, A2 a2, A3 a3) : function(f), arg1(a1), arg2(a2), arg3(a3) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg1,arg2,arg3);
        return true;
    }
};

template<class A1, class A2, class A3> struct StaticCyclicDispatcher3 : public GenericCyclicDispatcher {
    typedef void (*Func)(A1,A2,A3); // pointer to C function with three arguments
    Func function;
    A1 arg1;
    A2 arg2;
    A3 arg3;
    inline StaticCyclicDispatcher3(double p, Func f, A1 a1, A2 a2, A3 a3) : GenericCyclicDispatcher(p), function(f), arg1(a1), arg2(a2), arg3(a3) {}
    // Inline, just for cases where the compiler can determine this is a local call
    virtual bool dispatch(Simulator* simu) const {
        (*function)(arg1,arg2,arg3);
        // Use postAt for perfect cycle, post recomputes current time
        if (period>=0) { // period is mutable, cancel detector
            simu->postAt(simu->getTime() + period, this);
            return false;
        } else return true; // deletion
    }
};



    /** Post an event at current time.
        Any C "void method(..)"
        with one argument can be called this way.
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* post(void (*method)(A), A argument) {
        return post(new StaticDispatcher1<A>(method, argument));
    }

    /** Post an event at current time.
        Any C "void method(..)"
        with two arguments can be called this way.
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* post(void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return post(new StaticDispatcher2<A1,A2>(method, argument1, argument2));
    }

    /** Post an event at current time.
        Any C "void method(..)"
        with three arguments can be called this way.
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* post(void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return post(new StaticDispatcher3<A1,A2,A3>(method, argument1, argument2, argument3));
    }
    
    /** Post an event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* post(T* object, void (T::* method)()) {
        return post(new Dispatcher<T>(object, method));
    }
    
    /** Post an event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* post(T* object, void (T::* method)(A), A argument) {
        return post(new Dispatcher1<T,A>(object, method, argument));
    }
    
    /** Post an event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* post(T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return post(new Dispatcher2<T,A1,A2>(object, method, argument1, argument2));
    }
    
    /** Post an event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* post(T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return post(new Dispatcher3<T,A1,A2,A3>(object, method, argument1, argument2, argument3));
    }
    
    
    /** Post a cyclic event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postCyclic(double cycle, T* object, void (T::* method)()) {
        return post(new CyclicDispatcher<T>(cycle, object, method));
    }
    
    /** Post a cyclic event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postCyclic(double cycle, T* object, void (T::* method)(A), A argument) {
        return post(new CyclicDispatcher1<T,A>(cycle, object, method, argument));
    }
    
    /** Post a cyclic event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postCyclic(double cycle, T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return post(new CyclicDispatcher2<T,A1,A2>(cycle, object, method, argument1, argument2));
    }
    
    /** Post a cyclic event at current time. Any C++ method on any
        arbitrary object can be called this way.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postCyclic(double cycle, T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return post(new CyclicDispatcher3<T,A1,A2,A3>(cycle, object, method, argument1, argument2, argument3));
    }

    /** Post a cyclic event at current time. 
        Any C "void method(..)"
        with one argument can be called this way.
        @param cycle  how often the method should be called
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postCyclic(double cycle, void (*method)(A), A argument) {
        return post(new StaticCyclicDispatcher1<A>(cycle, method, argument));
    }

    /** Post a cyclic event at current time. 
        Any C "void method(..)"
        with two arguments can be called this way.
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postCyclic(double cycle, void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return post(new StaticCyclicDispatcher2<A1,A2>(cycle, method, argument1, argument2));
    }

    /** Post a cyclic event at current time.
        Any C "void method(..)"
        with three arguments can be called this way.
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postCyclic(double cycle, void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return post(new StaticCyclicDispatcher3<A1,A2,A3>(cycle, method, argument1, argument2, argument3));
    }

    /** Post an event at current time + a certain delay in simulated time
        Any C "void method(..)"
        with one argument can be called this way.
        @param delay after current time, should be >0.
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postIn(double delay, void (*method)(A), A argument) {
        return postIn(delay, new StaticDispatcher1<A>(method, argument));
    }

    /** Post an event at current time + a certain delay in simulated time
        Any C "void method(..)"
        with two arguments can be called this way.
        @param delay after current time, should be >0.
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postIn(double delay, void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return postIn(delay, new StaticDispatcher2<A1,A2>(method, argument1, argument2));
    }

    /** Post an event at current time + a certain delay in simulated time
        Any C "void method(..)"
        with three arguments can be called this way.
        @param delay after current time, should be >0.
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postIn(double delay, void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postIn(delay, new StaticDispatcher3<A1,A2,A3>(method, argument1, argument2, argument3));
    }
    
        
    /** Post an event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postIn(double delay, T* object, void (T::* method)()) {
        return postIn(delay, new Dispatcher<T>(object, method));
    }
    
    /** Post an event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postIn(double delay, T* object, void (T::* method)(A), A argument) {
        return postIn(delay, new Dispatcher1<T,A>(object, method, argument));
    }
    
    /** Post an event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postIn(double delay, T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return postIn(delay, new Dispatcher2<T,A1,A2>(object, method, argument1, argument2));
    }
    
    /** Post an event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postIn(double delay, T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postIn(delay, new Dispatcher3<T,A1,A2,A3>(object, method, argument1, argument2, argument3));
    }


    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, T* object, void (T::* method)()) {
        return postIn(delay, new CyclicDispatcher<T>(cycle, object, method));
    }
    
    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, T* object, void (T::* method)(A), A argument) {
        return postIn(delay, new CyclicDispatcher1<T,A>(cycle, object, method, argument));
    }
    
    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return postIn(delay, new CyclicDispatcher2<T,A1,A2>(cycle, object, method, argument1, argument2));
    }
    
    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C++ method on any arbitrary object can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postIn(delay, new CyclicDispatcher3<T,A1,A2,A3>(cycle, object, method, argument1, argument2, argument3));
    }


    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C "void method(..)" with one argument can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, void (*method)(A), A argument) {
        return postIn(delay, new StaticCyclicDispatcher1<A>(cycle, method, argument));
    }

    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C "void method(..)" with two arguments can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return postIn(delay, new StaticCyclicDispatcher2<A1,A2>(cycle, method, argument1, argument2));
    }

    /** Post a cyclic event at current time + a certain delay in simulated time
        Any C "void method(..)" with three arguments can be called this way.
        @param delay after current time, should be >0.
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postCyclicIn(double delay, double cycle, void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postIn(delay, new StaticCyclicDispatcher3<A1,A2,A3>(cycle, method, argument1, argument2, argument3));
    }
    

    /** Post an event at an absolute simulated time. Any C "void method(..)"
        with one argument can be called this way.
        @param absoluteTime time at which to execute the event
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postAt(double absoluteTime, void (*method)(A), A argument) {
        return postAt(absoluteTime, new StaticDispatcher1<A>(method, argument));
    }

    /** Post an event at an absolute simulated time. Any C "void method(..)"
        with two arguments can be called this way.
        @param absoluteTime time at which to execute the event
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postAt(double absoluteTime, void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return postAt(absoluteTime, new StaticDispatcher2<A1,A2>(method, argument1, argument2));
    }

    /** Post an event at an absolute simulated time. Any C "void method(..)"
        with three arguments can be called this way.
        @param absoluteTime time at which to execute the event
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postAt(double absoluteTime, void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAt(absoluteTime, new StaticDispatcher3<A1,A2,A3>(method, argument1, argument2, argument3));
    }

            
    /** Post an event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postAt(double absoluteTime, T* object, void (T::* method)()) {
        return postAt(absoluteTime, new Dispatcher<T>(object, method));
    }
    
    /** Post an event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postAt(double absoluteTime, T* object, void (T::* method)(A), A argument) {
        return postAt(absoluteTime, new Dispatcher1<T,A>(object, method, argument));
    }
    
    /** Post an event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postAt(double absoluteTime, T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return postAt(absoluteTime, new Dispatcher2<T,A1,A2>(object, method, argument1, argument2));
    }
    
    /** Post an event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postAt(double absoluteTime, T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAt(absoluteTime, new Dispatcher3<T,A1,A2,A3>(object, method, argument1, argument2, argument3));
    }

    /** Post a cyclic event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, T* object, void (T::* method)()) {
        return postAt(absoluteTime, new CyclicDispatcher<T>(cycle, object, method));
    }
    
    /** Post a cyclic event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, T* object, void (T::* method)(A), A argument) {
        return postAt(absoluteTime, new CyclicDispatcher1<T,A>(cycle, object, method, argument));
    }
    
    /** Post a cyclic event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return postAt(absoluteTime, new CyclicDispatcher2<T,A1,A2>(cycle, object, method, argument1, argument2));
    }
    
    /** Post a cyclic event at an absolute simulated time. Any C++ method on any
        arbitrary object can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAt(absoluteTime, new CyclicDispatcher3<T,A1,A2,A3>(cycle, object, method, argument1, argument2, argument3));
    }


    /** Post a cyclic event at an absolute simulated time. Any C "void method(..)"
        with one argument can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, void (*method)(A), A argument) {
        return postAt(absoluteTime, new StaticCyclicDispatcher1<A>(cycle, method, argument));
    }

    /** Post a cyclic event at an absolute simulated time. Any C "void method(..)"
        with two arguments can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return postAt(absoluteTime, new StaticCyclicDispatcher2<A1,A2>(cycle, method, argument1, argument2));
    }

    /** Post a cyclic event at an absolute simulated time. Any C "void method(..)"
        with three arguments can be called this way.
        @param absoluteTime time at which to execute the event
        @param cycle  how often the method should be called
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postCyclicAt(double absoluteTime, double cycle, void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAt(absoluteTime, new StaticCyclicDispatcher3<A1,A2,A3>(cycle, method, argument1, argument2, argument3));
    }

    
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline const GenericDispatcher* postAtStop(T* object, void (T::* method)()) {
        return postAtStop(new Dispatcher<T>(object, method));
    }
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline const GenericDispatcher* postAtStop(T* object, void (T::* method)(A), A argument) {
        return postAtStop(new Dispatcher1<T,A>(object, method, argument));
    }
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline const GenericDispatcher* postAtStop(T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        return postAtStop(new Dispatcher2<T,A1,A2>(object, method, argument1, argument2));
    }
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline const GenericDispatcher* postAtStop(T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAtStop(new Dispatcher3<T,A1,A2,A3>(object, method, argument1, argument2, argument3));
    }



    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param method any C "void method(..)" with 1 argument
        @param argument the argument to pass to the method
     */
    template<class A> inline const GenericDispatcher* postAtStop(void (*method)(A), A argument) {
        return postAtStop(new StaticDispatcher1<A>(method, argument));
    }
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline const GenericDispatcher* postAtStop(void (*method)(A1,A2), A1 argument1, A2 argument2) {
        return postAtStop(new StaticDispatcher2<A1,A2>(method, argument1, argument2));
    }
    
    /** Post an event to be executed when the simulation stops.
        Any C++ method on any arbitrary object can be called this way.
        These events have no execution date, but are called when the simulation stops
        in the order they were posted by postAtStop. The events are then consumed.
        No more then MAX_STOP_EVENTS can be posted this way.
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline const GenericDispatcher* postAtStop(void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        return postAtStop(new StaticDispatcher3<A1,A2,A3>(method, argument1, argument2, argument3));
    }

