/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <simulator.h>

using namespace std;

struct Bench {
    Simulator *simu;
    int c;
    
    Bench(Simulator *s) {simu=s; c = 0;}
    void nop();
    void repost();
    void counter();
    void cyclicCounter();
};

void Bench::nop( )
{
    // do nothing, just measure the time it takes to execute the post operation
}

void Bench::repost( )
{
    simu->post(this, &Bench::repost);
}


void Bench::counter( )
{
    if (c++<10000000) simu->post(this, &Bench::counter);
}

void Bench::cyclicCounter( )
{
    if (c++>=10000000) simu->stop();
}

bool hasStopped = false;
void stopCallback() {
    hasStopped = true;
}

int main(int argc, char** argv) {

    char* ID = 0;
    if (argc>1) ID = argv[1];
    
    Simulator *simu = Simulator::newInstance(ID);

    if (simu==0) {
        cout << "No simulator implementation available" << endl;
        return -1;
    }
    
    Bench bench(simu);
    
    cout << "Test 1: measuring 10000000 explicit posts/steps" << endl;
    double time = simu->getSystemTime();
    
    for (int i=0; i<10000000; ++i) {
        simu->post(&bench, &Bench::nop);
        if (!simu->step()) {
            cout << "Test 1 failed at i=" << i << endl;
            break;
        }
    }

    cout << "Test 1 result: " << simu->getSystemTime() - time << "s" << endl;


    cout << "Test 2: measuring 10000000 steps with automatic repost" << endl;
    time = simu->getSystemTime();
    
    simu->post(&bench, &Bench::repost);
    for (int i=0; i<10000000; ++i) {
        if (!simu->step()) {
            cout << "Test 2 failed at i=" << i << endl;
            break;
        }
    }
    
    cout << "Test 2 result: " << simu->getSystemTime() - time << "s" << endl;
    simu->flush(); // only way to get rid of the auto-repost!
    
    cout << "Test 3: measuring 10000000 steps with internal counter, automatic repost and step to date" << endl;
    time = simu->getSystemTime();
    
    simu->post(&bench, &Bench::counter);
    simu->step(10000.0); // time do not advance that much, so 10000.0 reached when there is no more event
    
    cout << "Test 3 result: " << simu->getSystemTime() - time << "s" << endl;
    
    cout << "Test 4: measuring 10000000 steps with internal counter, 0-cyclic event and start/stop" << endl;
    bench.c = 0;
    time = simu->getSystemTime();
    
    simu->postCyclic(0.0, &bench, &Bench::cyclicCounter);
    simu->postAtStop(&stopCallback);
    simu->start(false); // no thread, use the main program
    
    cout << "Test 4 result: " << simu->getSystemTime() - time << "s" << endl;
}

