/***************************************************************************
 *   Copyright (C) 2005 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <simulator.h>

/// Very simple diffusion reaction in 3D using finite differences.
/// Problem is the numerical stability condition. One solution would be to
/// use the Crank-Nicholson scheme, as described in (for example):
///    http://farside.ph.utexas.edu/teaching/329/lectures/node94.html
/// but then it requires either double FFT + 1 tridiagonal matrix inversion
/// or 3 matrix inversion in a 3-step algorithm. Plus here the world is
/// cyclic, the the matrices aren't tridiagonal at all...
///
/// So, what's here is the basic version, see (for example)
///    http://www.me.ucsb.edu/~moehlis/APC591/tutorials/tutorial5/tutorial5.html
///
/// And when time step is too large or grid space too small, there may be problems...
/// ... please ensure that dt is less than 0.5 * Sum_xzy (dx)^2 / D_xyz, with
///  D_xyz diffusion coefficients (Von Neumann stability analysis)

using namespace std;

    
Simulator *simu;

class Diffusion {
protected:
    double lastIntegrationStepTime;
    
    /// Diffusion coefficients constants with grid size
    double Dxc, Dyc, Dzc;
    
    /// diffusion coefficients for one given time step (see step function)
    double Cx, Cy, Cz, Csum2;

    /** Re-used from locality DB: the origin is the super-brick corner minimum coordinates */
    double originx, originy, originz;

    /** Re-used from locality DB: length of the edges of the super-brick */
    double sizex, sizey, sizez;

    /** Re-used from locality DB: number of sub-brick divisions in each direction */
    int divx, divy, divz;

    /// double buffering technique to avoid expensive array copy
    double *volume1;
    double *volume2;
    double **resVolume;
public:
    
    /// data structure for xyz bricks, actually a pointer to a triple array.
    double **volume;


    Diffusion( double originx, double originy, double originz, double sizex, double sizey, double sizez, int divx, int divy, int divz, double Dx, double Dy, double Dz )
    {
        this->divx = divx;
        this->divy = divy;
        this->divz = divz;
        this->originx = originx;
        this->originy = originy;
        this->originz = originz;
        this->sizex = sizex;
        this->sizey = sizey;
        this->sizez = sizez;
        // pre-compute x,y,z diffusion coefficients with grid size
        Dxc = sizex / divx;
        Dxc = Dx / (Dxc * Dxc);
        Dyc = sizey / divy;
        Dyc = Dy / (Dyc * Dyc);
        Dzc = sizez / divz;
        Dzc = Dz / (Dzc * Dzc);
        // initialize double buffer
        int vsize = divx * divy * divz;
        volume1 = new double[vsize];
        volume2 = new double[vsize];
        volume = &volume1;
        resVolume = &volume2;
        int offset = 0;
        for (int i=0; i<vsize; ++i) {
            volume1[i] = 0.0;
            volume2[i] = 0.0;
        }
        lastIntegrationStepTime = -1;
    }

    /// consider world is cyclic => no boundary condition!!!
    double& cyclicVolume(int x, int y, int z) {
        if (x<0) x += divx;
        else if (x>=divx) x -= divx;
        if (y<0) y += divy;
        else if (y>=divy) y -= divy;
        if (z<0) z += divz;
        else if (z>=divz) z -= divz;
        // return reference, array may be modified
        return (*volume)[z * (divx * divy) + y * divx + x];
    }
        
/*    // integrate a cell by finite difference with its neighbors
    double integrate(double xyz, double xm1yz, double xp1yz, double xym1z, double xyp1z, double xyzm1, double xyzp1) {
        double res = xyz;
        res -= res * Csum2;
        res += Cx * ( xm1yz + xp1yz );
        res += Cy * ( xym1z + xyp1z );
        res += Cz * ( xyzm1 + xyzp1 );
        return ret;
    }
*/

    /// integration step
    void step() {
    
        // First time is init
        if (lastIntegrationStepTime<0) {
            lastIntegrationStepTime = simu->getTime();
            return;
        }
    
        // time does not advance while processing event
        const double now = simu->getTime();
        double elapsedTime = now - lastIntegrationStepTime;
        lastIntegrationStepTime = now;
        
        // Compute finite difference
        // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant

        Cx = Dxc * elapsedTime;
        Cy = Dyc * elapsedTime;
        Cz = Dzc * elapsedTime;
        Csum2 = (Cx + Cy + Cz)*2.0;
        
        cout << "Cx=" << Cx << endl;
        cout << "Cx+Cy+Cz=" << Cx+Cy+Cz << endl;
        if (Csum2>=1.0) cout << "Error! Cx+Cy+Cz out of stability range: " << Csum2*0.5 << ">=0.5" << endl;
        else if (Csum2>=0.9) cout << "Warning! Cx+Cy+Cz close to instability: " << Csum2*0.5 << " close to 0.5" << endl;
        
        // put result of computation in other buffer
        double res, thisCell;
        int thisIndex = 0;
        for (int z=0; z<divz; ++z) for (int y=0; y<divy; ++y) for (int x=0; x<divx; ++x) {
            res = (*volume)[thisIndex];
            thisCell = res * 2.0;
            res += Cx * ( cyclicVolume(x-1, y, z) - thisCell + cyclicVolume(x+1, y, z) );
            res += Cy * ( cyclicVolume(x, y-1, z) - thisCell + cyclicVolume(x, y+1, z) );
            res += Cz * ( cyclicVolume(x, y, z-1) - thisCell + cyclicVolume(x, y, z+1) );
            (*resVolume)[thisIndex] = res;
            ++thisIndex;
        }

        // buffer switch
        double **tmpVolume = volume;
        volume = resVolume;
        resVolume = tmpVolume;
        
    }

    // rudimentary text-mode display
    void display(int z) {
        cout << "X-Y plane at t = " << lastIntegrationStepTime << endl;
        int w = cout.width();
        for (int y=0; y<divy; ++y) {
            for (int x=0; x<divx; ++x) {
                cout.width(10);
                cout.precision(5);
                cout << (*volume)[z * (divx * divy) + y * divx + x];
            }
            cout << endl;
        }
        cout.width(w);
    }
    
};

int main(int argc, char** argv) {

    char* ID = 0;
    if (argc>1) ID = argv[1];
    
    simu = Simulator::newInstance(ID);

    if (simu==0) {
        cout << "No simulator implementation available" << endl;
        return -1;
    }

    Diffusion diffusion(-5.0, -5.0, -5.0, 5.0, 5.0, 5.0, 10, 10, 10, 1.0, 1.0, 1.0);

    // give some value in the center to begin with, and let it diffuse
    (*diffusion.volume)[5 * 10*10 + 5 * 10 + 5] = 10.0;

    for (int i=0; i<100; ++i) {
        simu->postAt(i/26.0, &diffusion, &Diffusion::step);
        // display at z=0
        simu->postAt((i+0.5)/26.0, &diffusion, &Diffusion::display, 5);
    }
    
    simu->step(10.0);

    cout << endl << "What should see above is the diffusion result in the Z=0 plane. Initially there is just a 10.0 quantity of matter in the center, and it is diffused progressively through the whole space. In the limit, uniform concentration of 0.01 in the 1000 cells should be reached (conservation of matter)." << endl;
}

