/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <math.h>

#include "Agent.h"
#include "TranslateTransform.h"
#include "Environment.h"
#include "Obstacle.h"
using namespace Environment;
using namespace std;

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Utility.h"
using namespace Utility;


// Class Static variables
ShortestAngleAttitudeCorrection Agent::shortestAngleAttitudeCorrection;
list<Agent*> Agent::allAgents;

Agent::Agent()
{
    speed = 0.0;
    // default attitude is identity transform, world and local forward directions match
    forward.set(0,1,0);
    steeringForce.set(0.0,0.0,0.0);
    steeringForceLen = 0;
    attitudeCorrection = &shortestAngleAttitudeCorrection;

    newPosition = getPosition();
    newAttitude = getAttitude();
    newForward = forward;
    newSpeed = 0.0;
    
    // Ad-hoc hack!
    forwardRadius = 0.01;
    
    lastIntegrationStepTime = -1; // init at first step call
    lastAIStepTime = -1;          // init at first step call
    lastSteeringComputationTime = -1;
    lastWanderTime = -1;

    steppingStatus = 0;
    
    allAgents.push_back(this);
}

Agent::~Agent() {
   allAgents.remove(this);
}    

void Agent::destroy()
{
    // see integration step. Since called purely from simulator thread, no need for mutex lock
    if (steppingStatus!=0) {
        steppingStatus = -1;
        return;
    }

    // call parent delay, if any
    AgentCore::destroy();
}


void Agent::integrationStep( )
{
    // First time is init
    if (lastIntegrationStepTime<0) {
        lastIntegrationStepTime = simu->getTime();
        return;
    }

    // time does not advance while processing event
    const double now = simu->getTime();
    double elapsedTime = now - lastIntegrationStepTime;
    lastIntegrationStepTime = now;
   
    // acceleration takes mass & energy storage into account
    double invMassTime;
    if (energy!=Infinity) invMassTime = elapsedTime / (mass + energy * getEnergyFootprintRatio());
    else invMassTime = elapsedTime / mass;
    // (1/ + 4*) instead of (3/ + 3*)
    // don't declare it for one temporary use
    //osg::Vec3d integAcceleration = steeringForce * invMassTime;


    // change in speed vector by agent contributions only (no gravity)
    // use newForward & newSpeed temporarily
    newForward = steeringForce * invMassTime;
    newSpeed = newForward.length2();
    // no steering force, no clamping
    if (newSpeed>0) {
        if (energy!=Infinity) {
            if (energy>0) {
                // Impose energy restriction on steering.
                // => maximum speed change the agent can produce by itself. When speed=0 the agent will still fall down (see below)
                // speed * elapsedTime => distance => D*force = energy consumption
                // => (remaining energy / force) / elapsed time gives max limit for velocity
                double steeringMomentum = steeringForceLen * elapsedTime;
                double speedLimit = energy / steeringMomentum;
                double speedLimitSq = speedLimit*speedLimit;
                if (newSpeed > speedLimitSq) {
                    // clamp to speed limit
                    newForward *= sqrt(speedLimitSq / newSpeed); // Gain one mul at expense of slight loss of precision
                    newSpeed = speedLimit;
                } else newSpeed = sqrt(newSpeed);
                // new forward in correct length now
                // newSpeed is real contribution value, not squared.
                // reduce energy by the amount necessary for this contribution
                energy -= newSpeed * steeringMomentum;
                if (energy<0) energy=0;
            }
            else newForward.set(0,0,0); // no contribution from steering force this step
        }
    }
    
    // Add this contribution to the old speed to get new speed.
    // Hack for rotational momentum effect, not physically based!
    newForward += forward * (speed + forwardRadius);
    newSpeed = newForward.normalize() - forwardRadius;
    // Clamp to max speed the agent can reach by itself (it may go faster with the help of gravity, though)
    if (newSpeed > getMaxSpeed()) {
        // This is what this agent can add by itself, at most
        // This agent may have gained momentum with gravity, and we can't clip it by force of will
        if (speed>getMaxSpeed()) {
            newSpeed = speed; // keep the extra we had
        }
        else newSpeed = getMaxSpeed();
    }
    // It's very important to always keep a normalized forward, especially for attitude update
    else if (newSpeed<=0) {
        newSpeed = 0;
        newForward = forward;
    }

    // Throw in gravity
    if (gravity>0) {
        newForward *= newSpeed;
        newForward.z() -= gravity * invMassTime;
        newSpeed = newForward.normalize();
        // only possibility for newForward to be null is if forces exactly compensate gravity
        // In this case, speed=0 OK, but keep forward normalized & up
        if (newSpeed==0) newForward = forward;
    }
    
    // Throw in drag coefficient
    // dragCoefficient * speed^2 is force
    // dragCoefficient * speed^2 * invMassTime is speed
    // reduce speed to take drag into account
    if (dragCoefficient >0) newSpeed -= dragCoefficient * newSpeed * newSpeed * invMassTime;
    
    // Compute new attitude.
    attitudeCorrection->updateAttitude(newAttitude, attitude, newForward, forward);
    
    // compute new position in world space.
    newPosition = wrapPosition(position + newForward * newSpeed * elapsedTime);

    // position delete flag helper
    steppingStatus = 1;
    updateIntegration(elapsedTime, now);

    // get ready for next integration step
    forward = newForward;
    speed = newSpeed;

    setAttitude(newAttitude);
    setPosition(newPosition);

    // apply self-maintainance energy threshold. Child might have changed energy in updateIntegration;
    // so do this at the end of the function
    if (getSelfMaintainEnergyCost()<=0.0) return;
    
    energy -= getSelfMaintainEnergyCost() * elapsedTime;
    if (energy<=0) steppingStatus = -1;

    // destroy at the very end of the step, this object might be invalid afterwards
    // also take into account calls to destroy during updateIntegration
    if (steppingStatus < 0) {
        steppingStatus = 0;
        destroy();
        return;
    }
    steppingStatus = 0;
}

void Agent::aiStep( )
{
    // First time is init
    if (lastAIStepTime<0) {
        lastAIStepTime = simu->getTime();
        return;
    }
    
    const double now = simu->getTime();
    steppingStatus = 1;
    updateAI(now - lastAIStepTime, now);
    lastAIStepTime = now;

    clampToMaxForce( );

    // clamp to max energy
    if (energy>getMaxEnergy()) energy=getMaxEnergy();
    
    // destroy at the very end of the step, this object might be invalid afterwards
    // also take into account calls to destroy during updateIntegration
    if (steppingStatus < 0) {
        steppingStatus = 0;
        destroy();
        return;
    }
    steppingStatus = 0;
}

void Agent::clampToMaxForce( )
{
    // Temporary square of the value for comparison purpose
    steeringForceLen = steeringForce.length2();
    const double maxForceSq = getMaxForceSquared();
    
    if (steeringForceLen>maxForceSq) {
        // renormalize to max force. steeringForceLen!=0 at this point
        steeringForce *= sqrt(maxForceSq / steeringForceLen);
        steeringForceLen = getMaxForce();
    }
    else steeringForceLen = steeringForce.length();
}


void Agent::updateIntegration( double elapsedTime, double currentTime )
{
    // Do nothing by default
}

void Agent::updateAI( double elapsedTime, double currentTime )
{
    // Do nothing by default
}


/// Directly stolen from OpenSteer Library and adapted to this framework
double Agent::predictNearestApproachTime( Agent * agent )
{
    // imagine we are at the origin with no velocity,
    // compute the relative velocity of the other vehicle
    // NB043011: No wrapup needed for direction vectors, no position used
    const osg::Vec3d relVelocity = agent->forward * agent->speed - forward * speed;
    const double relSpeed = relVelocity.length();

    // for parallel paths, the vehicles will always be at the same distance,
    // so return 0 (aka "now") since "there is no time like the present"
    if (relSpeed == 0) return 0;
    
    // NB041130: speed up computations by replacing 4/ by 1/ & 4*
    const double invSpeed = 1.0 / relSpeed;

    // Now consider the path of the other vehicle in this relative
    // space, a line defined by the relative position and velocity.
    // The distance from the origin (our vehicle) to that line is
    // the nearest approach.

    // Take the unit tangent along the other vehicle's path
    const osg::Vec3d relTangent = relVelocity * invSpeed;

    // find distance from its path to origin (compute offset from
    // other to us, find length of projection onto path)
    // return (relTangent * (position - agent->position)) * invSpeed;

    const double* tgptr = relTangent.ptr();
    const double* p1ptr = position.ptr();
    const double* p2ptr = agent->position.ptr();
    
    if (Environment::cycleX==0.0) {
        if (Environment::cycleY==0.0) {
            return invSpeed * (tgptr[0] * (p1ptr[0] - p2ptr[0])
                             + tgptr[1] * (p1ptr[1] - p2ptr[1])
                             + tgptr[2] * (p1ptr[2] - p2ptr[2]));
        } else {
            return invSpeed * (tgptr[0] * (p1ptr[0] - p2ptr[0])
                             + tgptr[1] * remainder(p1ptr[1] - p2ptr[1], Environment::cycleY)
                             + tgptr[2] * (p1ptr[2] - p2ptr[2]));
        }
    } else {
        if (Environment::cycleY==0.0) {
            return invSpeed * (tgptr[0] * remainder(p1ptr[0] - p2ptr[0], Environment::cycleX)
                             + tgptr[1] * (p1ptr[1] - p2ptr[1])
                             + tgptr[2] * (p1ptr[2] - p2ptr[2]));
        } else {
            return invSpeed * (tgptr[0] * remainder(p1ptr[0] - p2ptr[0], Environment::cycleX)
                             + tgptr[1] * remainder(p1ptr[1] - p2ptr[1], Environment::cycleY)
                             + tgptr[2] * (p1ptr[2] - p2ptr[2]));
        }
    }
}


// overload parent to use cache
void Agent::findNeighbors(double radius) {
    if (lastSteeringComputationTime==simu->getTime())
        return;

    lastSteeringComputationTime=simu->getTime();
    AgentCore::findNeighbors(radius);
}


//=======================================================================

//                        Steering functions

//=======================================================================

osg::Vec3d Agent::steerForSeek( const osg::Vec3d & targetPosition )
{
    const osg::Vec3d offset = wrappedDiffVector(targetPosition, position);
    const double len2 = offset.length2();
    const double maxForceSq = getMaxForceSquared();
    if (len2>maxForceSq) {
        return offset / sqrt(maxForceSq / len2) - forward * ((simu->getTime() - lastAIStepTime) * speed);
    }
    return offset - forward * ((simu->getTime() - lastAIStepTime) * speed);
}

osg::Vec3d Agent::steerForFlee( const osg::Vec3d & targetPosition )
{
    const osg::Vec3d offset = wrappedDiffVector(position, targetPosition);
    const double len2 = offset.length2();
    const double maxForceSq = getMaxForceSquared();
    if (len2>maxForceSq) {
        return offset / sqrt(maxForceSq / len2) - forward * ((simu->getTime() - lastAIStepTime) * speed);
    }
    return offset - forward * ((simu->getTime() - lastAIStepTime) * speed);
}


osg::Vec3d Agent::steerToAvoid( Localized* object)
{

    // should not happen, but...
    if (object==this) return osg::Vec3d(0,0,0);

    osg::Vec3d deltaPosition = object->getPosition();
    double safeRadius = this->radius;

    // Is it a moving target? => use estimated movement
    Agent* agent = dynamic_cast<Agent*>(object);
    if (agent!=0) {
        deltaPosition += agent->forward * (agent->speed * predictNearestApproachTime(agent));
        safeRadius += agent->radius;
    } else {
        // Is it an obstacle with radius?
        Obstacle* obs = dynamic_cast<Obstacle*>(object);
        if (obs!=0) safeRadius += obs->getRadius();
    }
    
    // End delta pos vector
    deltaPosition -= position;

    // estimate collision risk based on triangle relation: do we pass close enough?
    double dp2 = deltaPosition.length2();

    // oops, already on the place
    if (dp2==0) {
        // quaternion rotation of local-space X radius vector
        osg::Vec3d side(radius * attitude.w(), radius * attitude.z(), -radius * attitude.y());
        // choose right or left, at random
        osg::Vec3d ret = Utility::Random() < 0.5 ? side : -side;
        return ret;
    }

    // use triangle relation A^2 = B^2 + C^2 - 2.B.C.cos(B,C) to get estimated distance
    // we'll pass from target if we continue in this direction
    // here B = deltaPosition and C = distance * forward vector
    double distance = sqrt(dp2);
    double dpf = deltaPosition * forward;
    double passageSq = 2.0 * (dp2 - distance * dpf);

    // now, if passage in less than radius, risk! Otherwise, estimate no risk
    if (passageSq > safeRadius * safeRadius) {
        return osg::Vec3d(0,0,0);
    }

    // vector from object position to nearest point on our forward direction
    // steering in this direction will accentuate the difference with object position
    osg::Vec3d tangential = (forward * dpf) - deltaPosition;

    // oops, exactly straight forward
    if (tangential.normalize()==0) {
        // quaternion rotation of local-space X radius vector
        osg::Vec3d side(radius * attitude.w(), radius * attitude.z(), -radius * attitude.y());
        // choose right or left, at random
        osg::Vec3d ret = Utility::Random() < 0.5 ? side : -side;
    }

    return tangential;
}

osg::Vec3d Agent::steerForWander( double changeRate )
{
    // Make this decision only once per step, otherwise it's not consistent
    if (simu->getTime()==lastWanderTime) return wanderCache;
    lastWanderTime = simu->getTime();

    if (changeRate<=0.0) {wanderCache = steeringForce; return wanderCache;}
    if (changeRate>1.0) {changeRate=1.0;}
    
    const double rescale = changeRate * 2.0;
    const double offset = - changeRate;
    
    // we already have previous steering force!
    // In all 3 directions, new force is centered on old force
    // with uniform distribution in (-changerate, +changerate) 
    // proportionally to the old force length.
    // Exception: when no steering force was set, use forward
    osg::Vec3d previousForce = steeringForce;
    if (previousForce.length2()==0.0) previousForce = forward;
    
    const double maxForce = getMaxForce();
    wanderCache = osg::Vec3d(
        previousForce.x() + maxForce * (offset + rescale * Utility::Random()),
        previousForce.y() + maxForce * (offset + rescale * Utility::Random()),
        previousForce.z() + maxForce * (offset + rescale * Utility::Random())
    );
    return wanderCache;
}

osg::Vec3d Agent::steerForEvasion( Agent * menace )
{
// NB: try this simpler & less hacky version
const double elapsedTime = simu->getTime() - lastAIStepTime;
return steerForFlee(menace->getPosition() + menace->forward * (menace->speed * elapsedTime));
    
    osg::Vec3d offset = wrappedDiffVector(menace->getPosition(), position);

    return steerForFlee(menace->getPosition() + menace->forward * offset.normalize());
}

osg::Vec3d Agent::steerForPursuit( Agent * quarry )
{
// NB: try this simpler & less hacky version
const double elapsedTime = simu->getTime() - lastAIStepTime;
return steerForSeek(quarry->getPosition() + quarry->forward * (quarry->speed * elapsedTime));


    // offset from this to quarry, that distance, unit vector toward quarry
    osg::Vec3d offset = wrappedDiffVector(quarry->getPosition(), position);
    const double distance = offset.normalize();

    // how parallel are the paths of "this" and the quarry
    // (1 means parallel, 0 is pependicular, -1 is anti-parallel)
    const double parallelness = forward * quarry->forward;

    // how "forward" is the direction to the quarry
    // (1 means dead ahead, 0 is directly to the side, -1 is straight back)
    const double forwardness = forward * offset;

    const int f = forwardness < -0.707f ? -1 : (forwardness > 0.707f ? 1 : 0);
    const int p = parallelness < -0.707f ? -1 : (parallelness > 0.707f ? 1 : 0);

    double timeFactor = 0.0; // to be filled in below

    // Break the pursuit into nine cases, the cross product of the
    // quarry being [ahead, aside, or behind] us and heading
    // [parallel, perpendicular, or anti-parallel] to us.
    if (forwardness>0.707) {
        if (parallelness > 0.707)        timeFactor = 4.0;  // ahead, parallel
        else if (parallelness < -0.707)  timeFactor = 0.85; // ahead, anti-parallel
        else                             timeFactor = 1.8;  // ahead, perpendicular
    } else if (forwardness<-0.707) {
        if (parallelness > 0.707)        timeFactor = 0.5;  // behind, parallel
        else if (parallelness < -0.707)  timeFactor = 2.0;  // behind, anti-parallel
        else                             timeFactor = 2.0;  // behind, perpendicular
    } else {
        if (parallelness > 0.707)        timeFactor = 1.0;  // aside, parallel
        else if (parallelness < -0.707)  timeFactor = 4.0;  // aside, anti-parallel
        else                             timeFactor = 0.8;  // aside, perpendicular
    }

    // estimated time until intercept of quarry
    const double et = distance * timeFactor / speed;

    // estimated position of quarry at intercept
    return steerForSeek (quarry->getPosition() + quarry->forward * (quarry->speed * et));
}

osg::Vec3d Agent::steerForTargetSpeed( double targetSpeed )
{
    const double elapsedTime = simu->getTime() - lastAIStepTime;
    
    // acceleration takes mass & energy storage into account
    double massTime;
    if (energy!=Infinity) massTime = (mass + energy * getEnergyFootprintRatio()) / elapsedTime;
    else massTime = mass / elapsedTime;

    return forward * ((targetSpeed - speed) * massTime);
}

