/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CROGAI_AGENT_H
#define CROGAI_AGENT_H

#include "AgentCore.h"
#include "ShortestAngleAttitudeCorrection.h"

/**

    Simple vehicule model reusing ideas from Craig Reynolds GDC'99 presentation
    - http://www.red3d.com/cwr/steer/gdc99/
    - http://opensteer.sourceforge.net/
    
    The main differences between this model and the OpenSteer library are:
    - The AI rate, integration rate, and frame rate are separated. This allows batch
    computing overnight of large simulations, non-real-time rendering (slower or
    even faster if your system can do it) and also to apply AI at low frequency.
    - Energy storage is taken into account. When there is no energy, the car stops and the prey halts...
    - Attitude correction is user-defined by helper classes.
    
    Integration update => at least once per frame in graphics mode, may be more often
    - Automatic
    - Sould ideally be as fast as possible
    - Should at least update position & attitude
    - In a scenario using energy, this is the granularity for energy consumption
    - Is the mathematical integration step, which is especially significant for speed & position. A higher frequency
      gives a better approximation: Consequently, changing the frequency also changes these values, which
      may be surprising at first when running the same simulation with the same parameters. You've been warned!
    
    AI update => From time to time
    - Can take more time than integration step.
    - Atomic with respect to integration update (if AI takes more than one frame, values
      are applied to the next frame).
    - Allows to simulate more agents by spreading time-consuming agents AI updates in a time interval: Don't
      do all AI updates in one go, then wait next cycle, this would be a waste of resources!
      Ex: do half of them at t, half of them at t+cycle/2. (Or split in 1/3, 1/4...). You may also spread
      them randomly on the interval, using a random seed to keep the determinism, or uniformly with a
      predefined agent order.

    Attitude correction:
    - When mapping the old forward direction to the new forward, an infinity of rotations can do the job.
    We must choose one. By default this is the shortest angle rotation bewteen both vectors, thanks
    to quaternions. This should be equivalent to the default in the OpenSteer library, as the quaternion
    we use combines both rotations that are done there. See the ShortestAngleAttitudeCorrection class. 
    - Another policy is provided to force the agent to keep an "up" vector close to a target, which
    actually defines an horizontal plane. See the KeepUpAttitudeCorrection class.
    - No policy was made for banking. Anyone please?

    Physics:
    - It is based on a point-mass model, as in the OpenSteer library. However, I've added in energy consumption,
      a drag effect, an energy footprint ratio, and a hack to simulate rotational momentum.
    - In a scenario with energy, the energy footprint conversion ratio simulates the body mass necessary to hold
      this energy. necessary_mass = energy * ratio. See also AgentCore.h
      . Ex: prey/predator => this is the energy stored in stomach => more predator mass.
      . Ex2: car race => this is the weight of carburant... or nothing for solar cars
      . Set this to 0 to simulate an infinite amount of energy can be stored without mass increase (solar cars).
    - There is no rotational momentum. An ad-hoc hack is used to simulate the fact more massive objects take
      longer to turn on themselves given the same force. It is not physically based, but adds some realism at
      very low cost!
    - Don't take the physics in this class too seriously, it is certainly not realistic enough for an industrial
      material property simulation. However, as very often in games, we're not interested in perfect representation
      of reality (whatever this means, an equation is also only a representation!). What's interesting is a 
      realistic enough simulation at a reasonable computational cost, which is this implementation intent.
      
    
    @author Nicolas Brodu
*/

class Agent : public AgentCore {
public:
    Agent();
        
/////////////////////////////////////////////////////////////////////////////
/// AI Attributes
/////////////////////////////////////////////////////////////////////////////
    
    /** Speed of the agent, along its forward direction, expressed in OSG space unit / simulator time unit.

        You should not normally need to change this value directly, but rather set a steering force.
        Of course, if your particular environment has slow-down zones (water?) then maybe...
    */
    double speed;

    /** Current forward vector in world coordinates: initialized to Y in local space to begin with,
        so orient your models accordingly at load time (or override this in your constructor)
        
        You should not normally need to change this value directly, but rather set a steering force.
        Of course, if your particular environment has special zones (water swirl?) then maybe...
    */
    osg::Vec3d forward;
    
    /** Current steering force set by AI, in world coordinates.
        Apply this force to the current agent at each integration update,
        so long as it can (so song as energy remains), till next AI update.

        One very important goal of the updateAI overload is to update this variable.
        Note: It is clamped to maxForce at the end of the AI step, after updateAI
        is called.
    */
    osg::Vec3d steeringForce;
        
    /** Attitude correction policy for this agent.
        
        By default, all agents use a ShortestAngle attitude correction, which has the nice advantage
        of being stateless so it can be shared among agents.
        This field is <b>not</b> deleted in the destructor (the shared ShortestAngle is handled separately).
        So if you change the default policy, it's your responsability to manage this field.
    */
    AttitudeCorrection *attitudeCorrection;


/////////////////////////////////////////////////////////////////////////////
///
///    AI related methods
///
///    Note: Use these methods in the updateAI routine to get force vectors to apply
///    on this agent. The same force will remain in place till the next AI update,
///    so for ex steerForSeek needs to be called again since the agent has moved
///    (it is not called each time in the integration step).
///
///    Method names are voluntarily copied from the OpenSteer library.
///    Differences in implementation are expressely given for each method.
///
/////////////////////////////////////////////////////////////////////////////

    /** Return a force to apply to steer toward the given point.
        Roughly equivalent to the experimental version of the OpenSteer Library
    */
    virtual osg::Vec3d steerForSeek(const osg::Vec3d& targetPosition);

    /** Return a force to apply to steer to the opposite direction of the given point
        Roughly equivalent to the experimental version of the OpenSteer Library
    */
    virtual osg::Vec3d steerForFlee(const osg::Vec3d& targetPosition);

    /** Return a force to apply to steer to avoid collision with a given
        object. Then:
        - No collision in view: return 0 vector
        - Collision in view: returns a tangential vector (from current forward),
          in a direction maximizing separation with object.
        
        Notes:
        - This algorithm makes the distinction between Agents, obstacles,
          and anything else. Agents future position is estimated based on their
          visible current speed & direction. Agents & Obstacles radius add to
          our own to compute the minimum safe distance passage.
        - This is not at all the same algorithm as in the OpenSteer Library
     */
    virtual osg::Vec3d steerToAvoid( Localized* object );

    /** Not only do we seek to avoid the agent, but also try to escape it.
        This is exactly the same algorithm as in the OpenSteer Library
        without maximum prediction time, and adapted to this framework.

        We don't include a maximum for prediction time because that's one more
        parameter we don't have a clear default value for, and because anyway
        this function will usually be called for neighbors, with the intent to
        evade NOW!!!
    */
    virtual osg::Vec3d steerForEvasion(Agent* menace);
    
    /** The opposite of evasion, try to follow the other agent.
        This is exactly the same algorithm as in the OpenSteer Library
        without maximum prediction time, and adapted to this framework.
        
        We don't include a maximum for prediction time because that's one more
        parameter we don't have a clear default value for, and because anyway
        this function will usually be called for neighbors, with the intent to
        pursue them so long as they are visible.
    */
    virtual osg::Vec3d steerForPursuit(Agent* quarry);

    
    /** Returns a "random" direction, based on the previous direction.

        The idea is the same as in the OpenSteer library in the sense that
        there is no abrupt change in direction, allowing consistent turns.

        But this implementation does not constrain the new force on a sphere.
        Instead of the sphere radius, a new parameter is introduced to
        quantify the "smoothness" of the changes.

        Note: The standard Utility::Random() function is used, you may control the seed.
        
        @param changeRate how much change is allowed to the previous direction.
               A value of 0.0 indicates no change is allowed, useless.
               A value of 1.0 indicates the new force can potentially be in any
               direction (the old direction is still the prefered on average).
    */
    virtual osg::Vec3d steerForWander(double changeRate);

    /** Try to adjust speed to the desired speed. Do not change direction, useful for breaking.

        The implementation in the OpenSteer library uses the speed delta as a force.

        This implementation use numerical integration to predict the effect of the returned
        force, such that if applied the alone target speed would be reached at next AI step.
        It's not clamped to max force at this point.

        @param targetSpeed the desired speed to reach
    */
    virtual osg::Vec3d steerForTargetSpeed(double targetSpeed);
    
//////////////////////////////////////////////////////////////////////////
/// Publicly Schedulable Simulation methods
//////////////////////////////////////////////////////////////////////////

    /** Integration update: update position and attitude by taking into account
        forces set by AI. This is the method that should be scheduled,
        subclasses should override the updateStep method to modify the position
        and attitude _before_ the locality database is updated!

        Note: The very first call to this method actually sets the starting point
              in time for the first cycle. It returns immediately.
              Subsequent calls can then use the (simulation) time delta for integration.
     */
    virtual void integrationStep();

    /** Public AI step. This is the method that should be scheduled,
        subclasses should override the updateAI protected method which is given
        time information.
        
        Note: The very first call to this method actually sets the starting point
              in time for the first cycle. It returns immediately.
              Subsequent calls can then use the (simulation) time delta for integration.
     */
    virtual void aiStep();


/////////////////////////////////////////////////////////////////////////////
/// Administrative stuff
/////////////////////////////////////////////////////////////////////////////

    /** Destroys the agent. Really, don't use delete.... */
    virtual void destroy();

    /** A list containing all objects of this class
        Subset of Localized, useful for faster algorithms
    */
    static std::list<Agent*> allAgents;

protected:
    virtual ~Agent();
    

//////////////////////////////////////////////////////////////////////////
/// Simulation methods for subclasses
//////////////////////////////////////////////////////////////////////////

    /** 
        Subclasses should override the updateAI method to implement an AI.
        At least, a new steering force should be given.
        All steerForXXX methods are available.
     */
    virtual void updateAI(double elapsedTime, double currentTime);
    
    /** 
        Subclasses should override the updateStep method to modify the position
        and attitude _before_ the locality database is updated!
        They can also modify newPosition and newAttitude directly in this method:
        the update consistency lock in done only once in the integrationStep method,
        after this call
     */
    virtual void updateIntegration(double elapsedTime, double currentTime);
    
    /** new Position usable in the updateStep method, set by integrationStep.
        This will be the new position of this agent, so it can be modified in
        the updateStep method
    */
    osg::Vec3d newPosition;
    
    /** new Attitude usable in the updateStep method, set by integrationStep.
        This will be the new attitude of this agent, so it can be modified in
        the updateStep method
    */
    osg::Quat newAttitude;
    
    /** new Forward vector usable in the updateStep method, set by integrationStep.
        This will be the new forward vector of this agent, so it can be modified in
        the updateStep method
    */
    osg::Vec3d newForward;
    
    /** new Speed usable in the updateStep method, set by integrationStep.
        This will be the new speed of this agent, so it can be modified in
        the updateStep method
    */
    double newSpeed;





//////////////////////////////////////////////////////////////////////////
/// Other protected methods - internal utilities
//////////////////////////////////////////////////////////////////////////
    
    /// Directly stolen from OpenSteer Library and adapted to this framework
    virtual double predictNearestApproachTime(Agent* agent);

    
//////////////////////////////////////////////////////////////////////////
/// Protected fields
//////////////////////////////////////////////////////////////////////////
    
    /// Pre-compute force len at AI frequency, not integration step frequency...
    double steeringForceLen;

    /// Ad-hoc hack to avoid instantaneous rotation on place due to point-mass model.
    /// While this is no substitute for a proper rotational momentum, it still
    /// allows agents with more mass to turn more slowly on themselves, adding realism.
    double forwardRadius;
    
    /// step control variable: explicit name
    double lastIntegrationStepTime;
    
    /// step control variable: explicit name
    double lastAIStepTime;

    /// Shared class variable. The shortest angle policy is stateless, which makes this possible.
    static ShortestAngleAttitudeCorrection shortestAngleAttitudeCorrection;

    
//////////////////////////////////////////////////////////////////////////
/// Purely internal stuff.
//////////////////////////////////////////////////////////////////////////
private:
    // cache the results of neighbor computations: time of the cached results
    double lastSteeringComputationTime;
    std::multimap<const double, Localized*> cachedNeighborMap;
    double lastWanderTime;
    osg::Vec3d wanderCache;
    int steppingStatus;
    
public:
    virtual void clampToMaxForce();
    
    // overload parent to use cache
    virtual void findNeighbors(double radius);
};


#endif
