/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>
#include <sstream>
#include <math.h>

#include <osg/ShapeDrawable>
#include <osg/Geode>
#include <osg/Projection>
#include <osg/MatrixTransform>

#include "KeepUpAttitudeCorrection.h"
#include "CrogaiAgent.h"
#include "Environment.h"
#include "Utility.h"
using namespace Environment;
using namespace std;

bool CrogaiAgent::seeOnlyInFront = true;

std::map<int, int> CrogaiAgent::scores;
std::map<int, int> CrogaiAgent::populations;
double CrogaiAgent::defaultSearchRadius = 30.0;
bool CrogaiAgent::displayScores = false;
bool CrogaiAgent::displayPopulationNumbers = true;
bool CrogaiAgent::displayPopulationCounter = true;

Logger* CrogaiAgent::logPopulations = 0;
Logger* CrogaiAgent::logScores = 0;

/*
bool CrogaiAgent::displayGenerations = true;
std::map<int, int> CrogaiAgent::generations;
Logger* CrogaiAgent::logGenerations = 0;
*/

// statistics
std::map<int, std::vector<double> > CrogaiAgent::lifeTimeMean;
std::map<int, std::vector<double> > CrogaiAgent::lifeTimeVariance;
std::map<int, std::vector<double> > CrogaiAgent::maxEnergyReachedMean;
std::map<int, std::vector<double> > CrogaiAgent::maxEnergyReachedVariance;
std::map<int, std::vector<double> > CrogaiAgent::offspringQuantityMean;
std::map<int, std::vector<double> > CrogaiAgent::offspringQuantityVariance;
std::map<int, std::vector<int> > CrogaiAgent::generationsPopulationNumber;
std::map<int, std::vector<double> > CrogaiAgent::generationsPopulationFirstTime;
Logger* CrogaiAgent::logStatistics = 0;

// Horribly ugly hack for immediate needs
typedef void (*PopulationUpdateCallback)( int delta, int specie );
static vector<PopulationUpdateCallback> populationUpdateCallbacks;

osg::Vec4 CrogaiAgent::getDefaultColorForSpecie( int specie )
{
    switch(specie) {
        case 1: return osg::Vec4(0,0,1,1); // blue
        case 2: return osg::Vec4(1,0,0,1); // red
        case 3: return osg::Vec4(0,1,0,1); // green
        case 4: return osg::Vec4(1,0,1,1); // magenta
        case 5: return osg::Vec4(0,1,1,1); // cyan
        case 6: return osg::Vec4(1,1,0,1); // yellow
        case 7: return osg::Vec4(1,1,1,1); // white
        case 8: return osg::Vec4(0.05,0.05,0.05,1); // very dark, just in case there would be a black background
    }
    // compute a strange color for anything else
    return osg::Vec4(
        sin((double)specie)*0.5+0.5,
        tanh(sin((double)specie)*cos((double)specie)*10.0)*0.5+0.5,
        cos((double)specie)*0.5+0.5,
        1.0
    );
}

// Quick hack to compensate lack of per-character color in osgText::Text
// Proper solution would be to overload osgText
std::string CrogaiAgent::getDefaultColorNameForSpecie( int specie )
{
    switch(specie) {
        case 1: return "blue";
        case 2: return "red";
        case 3: return "green";
        case 4: return "magenta";
        case 5: return "cyan";
        case 6: return "yellow";
        case 7: return "white";
        case 8: return "dark";
    }
    ostringstream os;
    os << "color" << specie;
    return os.str();
}



CrogaiAgent::CrogaiAgent(int specie) {
    
    this->specie = specie;
    generation = 0;
    offspringQuantity = 0;
    aiEvent = 0;
    integrationEvent = 0;
    searchRadius = defaultSearchRadius;
    birthDate = (simu!=0) ? simu->getTime() : 0.0;
    maxEnergyReached = 0;

    ++populations[specie];

    if (displayPopulationCounter) {
        ostringstream os;
        os << "Population: " << allAgents.size() - 1; // account for removal of this one too
        Utility::setStatus(os.str(), Utility::TOP_LEFT);
    }

    if (displayPopulationNumbers) {
        ostringstream os;
        for (map<int, int>::iterator it = populations.begin(); it != populations.end(); ++it) {
            if (it != populations.begin()) os << ", ";
            os << getDefaultColorNameForSpecie(it->first) << ":" << it->second;
        }
        Utility::setStatus(os.str(), Utility::TOP_RIGHT);
    }

    if (logPopulations) {
        Logger::iterator lit = logPopulations->begin();
        *lit++ = simu->getTime();
        for (map<int, int>::iterator it = populations.begin(); it != populations.end(); ++it) {
            *lit++ = it->second;
        }
        logPopulations->record();
    }

    if (!populationUpdateCallbacks.empty()) {
        for (vector<PopulationUpdateCallback>::iterator it = populationUpdateCallbacks.begin();
        it != populationUpdateCallbacks.end(); ++it) {
            PopulationUpdateCallback callback = *it;
            callback(1, specie);
        }
    }
    
    //message = 0.0; // what do we have to say anyway?

    // new avatar, new radius => *.5 * sqrt( 1*1 + 2*2 + 0.5*0.5 )
    radius = 1.14564392374;

    // Choose random direction in X/Y plane
    osg::Vec3d rforward;
    do {rforward.set(Utility::Random()*2.0 - 1.0, Utility::Random()*2.0 - 1.0, 0.0);}
    while (rforward.normalize() == 0.0);
    // Attitude quaternion must be updated since we change direction
    attitudeCorrection->updateAttitude(attitude, attitude, rforward, forward);
    forward = rforward;

}


CrogaiAgent::~ CrogaiAgent( )
{
    if (aiEvent!=0) {
        simu->cancel(aiEvent);
        aiEvent = 0;
    }
    if (integrationEvent!=0) {
        simu->cancel(integrationEvent);
        integrationEvent = 0;
    }
    // Notifies all students we're out of the game
    for (list<CrogaiAgent*>::iterator it = students.begin(); it != students.end(); ++it) {
        (*it)->removeTeacher(this);
    }

    --populations[specie];
    
    if (displayPopulationCounter) {
        ostringstream os;
        os << "Population: " << allAgents.size() - 1; // account for removal of this one too
        Utility::setStatus(os.str(), Utility::TOP_LEFT);
    }

    if (displayPopulationNumbers) {
        ostringstream os;
        for (map<int, int>::iterator it = populations.begin(); it != populations.end(); ++it) {
            if (it != populations.begin()) os << ", ";
            os << getDefaultColorNameForSpecie(it->first) << ":" << it->second;
        }
        Utility::setStatus(os.str(), Utility::TOP_RIGHT);
    }
    
    if (logPopulations) {
        Logger::iterator lit = logPopulations->begin();
        *lit++ = simu->getTime();
        for (map<int, int>::iterator it = populations.begin(); it != populations.end(); ++it) {
            *lit++ = it->second;
        }
        logPopulations->record();
    }

    // Update these statistics that can only be known after agent death
    if (logStatistics) updateStatistics();

    if (!populationUpdateCallbacks.empty()) {
        for (vector<PopulationUpdateCallback>::iterator it = populationUpdateCallbacks.begin();
        it != populationUpdateCallbacks.end(); ++it) {
            PopulationUpdateCallback callback = *it;
            callback(-1, specie);
        }
    }
}


void CrogaiAgent::addStudent(CrogaiAgent* student) {
    students.push_back(student);
}
    
void CrogaiAgent::removeStudent(CrogaiAgent* student) {
    students.remove(student);
}

void CrogaiAgent::removeTeacher( CrogaiAgent * teacher )
{
    // do nothing, we're not a student
}

void CrogaiAgent::learnAI( const double elapsedTime, const double currentTime, CrogaiAgent * teacher )
{
    // do nothing by default, we're not a student
}


osg::Box* CrogaiAgent::crogaiAgentShape = new osg::Box(osg::Vec3(0,0,0),1,2,0.5);

osg::Node *CrogaiAgent::buildAvatar()
{
    // Share the body of these agents across all members of the class, thanks to ref counting
    osg::ShapeDrawable *body = new osg::ShapeDrawable(crogaiAgentShape);

    // Allow the color to be changed at run time
    //body->setUseDisplayList( false );
    body->setColor(getDefaultColorForSpecie(specie));

    // OSG used geodes as generic containers. Bridge from osg::Shape to osg::Node
    osg::Geode* geode = new osg::Geode();
    geode->addDrawable(body);
    return geode;
}


void CrogaiAgent::aiStep()
{
    // First time is init. See parent aiStep
    if (lastAIStepTime<0) {
        lastAIStepTime = simu->getTime();
        return;
    }
    
    const double currentTime = simu->getTime();
    double elapsedTime = currentTime - lastAIStepTime;
    
    // Calls parent AI step => updates lastAIStepTime so store it before
    TerrainAgent::aiStep();

    // No students, nothing else to do
    if (students.empty()) return;
    
    // Notifies all students. They may notify their own students if needed!
    for (list<CrogaiAgent*>::iterator it = students.begin(); it != students.end(); ++it) {
        (*it)->learnAI(elapsedTime, currentTime, this);
    }

}


// Handle collision detection
void CrogaiAgent::updateIntegration(const double elapsedTime, const double currentTime)
{
    TerrainAgent::updateIntegration(elapsedTime, currentTime);
    
    // Ad-hoc hack to make collisions effective & more efficient
    double radius22 = 2.0 * radius * radius;
    
    double len;
    CrogaiAgent* winner = 0;
    CrogaiAgent* looser = 0;
    CrogaiAgent* draw = 0;
    CrogaiAgent* predator = 0;
    CrogaiAgent* prey = 0;
    osg::Vec3d collisionVector(0,0,0);

    int collisionType = 0;
    
    // collision detection. Use current neighbor list
    for (multimap<const double,Localized*>::iterator itn = neighbors->begin(); itn != neighbors->end(); ++itn) {
        if (itn->second == this) continue; // should not happen, but...
        
        len = wrappedDist2(itn->second->getPosition(), newPosition);
        if (len <= radius22) {
        
            collisionVector = wrappedDiffVector(itn->second->getPosition(), newPosition);
            collisionType = 1; // bump off by default. indicator that !=0 too
            
            // Bounce from unknown things from outer space!!!
            CrogaiAgent* crogai = dynamic_cast<CrogaiAgent*>(itn->second);
            if (crogai==0) break;

            // handle fights between preys & predators
            if (isPredator(crogai)) {
                // special case for both predator & prey
                if (isPrey(crogai)) {

                    // we're caught from behind in see only in front scenario => loose
                    if (seeOnlyInFront && (collisionVector * forward < 0)) {
                        winner = crogai;
                        looser = this;
                        break;
                    }
                    
                    // hehe, we are behind in see only in front scenario , we win!
                    if (seeOnlyInFront && (crogai->forward * forward >0)) {
                        winner = this;
                        looser = crogai;
                        break;
                    }

                    // Energy scenario: use this to determine combat issue
                    if ((energy!=Infinity) && (crogai->energy!=Infinity)) {
                        // first, get how much energy each one is willing to spend
                        double fightEnergy = getEnergyForFight(); // save it before, it may be affected by next line
                        energy -= crogai->getEnergyForFight();
                        crogai->energy -= fightEnergy;

                        if (energy<=0) {
                            energy = 0;
                            if (crogai->energy<=0) {
                                crogai->energy = 0;
                                looser = this;
                                draw = crogai;
                                break;
                            }
                            winner = crogai;
                            looser = this;
                            break;
                        }
                        
                        if (crogai->energy<=0) {
                            crogai->energy = 0;
                            winner = this;
                            looser = crogai;
                            break;
                        }
                        // there may just be no winner and looser
                        prey = this; predator = crogai;
                        collisionType = 2;
                        break;
                    }
                    
                    // Not an energy scenario, no way to get a winner
                    break;
                }

                // Not special case: encounter with a predator... Ooops!!!

                // we're caught from behind in see only in front scenario => loose immediately
                if (seeOnlyInFront && (collisionVector * forward < 0)) {
                    winner = crogai;
                    looser = this;
                    break;
                }

                // otherwise, share common code for fight
                prey = this;
                predator = crogai;
                
            }
            else if (isPrey(crogai)) {

                // In see only in front scenario:
                // foolish prey bumping from behind??? good news! => fight it
                // Note: this case shouldn't happen as it doesn not appear in the neighbors list at this point
                // => check if we're behind it => we win immediately
                if (seeOnlyInFront && (collisionVector * forward >= 0) && (crogai->forward * forward >0)) {
                    winner = this;
                    looser = crogai;
                    break;
                }

                // general case: fight
                prey = crogai;
                predator = this;
            }

            // So, there is a fight
            if (prey!=0) {
                // Energy scenario: use this to determine combat issue
                if ((prey->energy!=Infinity) && (predator->energy!=Infinity)) {
    
                    double fightEnergy = predator->getEnergyForFight();
                
                    // predator is willing to spend more energy than prey have => death
                    if (fightEnergy > prey->energy) {
                        predator->energy -= prey->energy; // predator looses fight energy
                        prey->energy = 0;
                        winner = predator;
                        looser = prey;
                        break;
                    }
                    
                    // otherwise, prey just escapes as fast as it can!
                    predator->energy -= fightEnergy; // predator looses fight energy
                    prey->energy -= fightEnergy;
                    collisionType = 2;
                }
                
                // Not an energy scenario, no way to get a winner
                break;
            }

            // at this point, no fight => collision with a passer-by
            // In any case, one collision is enough
            break;
        }
    }

    if (collisionType==0) return;
    
/*
Actions:
A: change dir
B: slow down
C: fast escape
D: remove from other neighbors to avoid multiple encounters at same time step.

Collision types:
1: passer-by, obstacle: A, B
2: draw fight, but no-one dies: A, C, D for prey; B, D, for predator.
3: winner-looser-draw: B for winner (if any)
*/
    
    if (winner) {
        scores[specie]++;
        if (displayScores) {
            ostringstream os;
            os << "Scores: ";
            for (map<int, int>::iterator it = scores.begin(); it != scores.end(); ++it) {
                if (it!=scores.begin()) os << ", ";
                os << it->first << ":" << it->second;
            }
            Utility::setStatus(os.str(), Utility::TOP_LEFT);
        }
        
        if (logScores) {
            Logger::iterator lit = logScores->begin();
            *lit++ = simu->getTime();
            for (map<int, int>::iterator it = scores.begin(); it != scores.end(); ++it) {
                *lit++ = it->second;
            }
            logScores->record();
        }

        winner->winnerAction(looser);
        // Action B for winner: Hack to slow down on collision
        if (winner==this) newSpeed *= 0.3;
        else winner->speed *= 0.3;
        collisionType = 3;
    }
    
    if (looser) {
        looser->looserAction(winner); // winner is 0 in case of draw
        looser->speed = 0;
        looser->destroy(); // destroy it so far as agents are concerned
        collisionType = 3;
    }
    
    if (draw) {
        draw->looserAction(0);
        draw->speed = 0;
        draw->destroy(); // destroy it so far as agents are concerned
        collisionType = 3;
    }

    // done!
    if (collisionType == 3) return;

    if (collisionType==1) {
        // Action A: change direction
        if (collisionVector.normalize()==0)
            collisionVector = newForward; // bounce back if exactly at the same place of colliding target
    
        // Attitude quaternion must be updated since we change direction
        attitudeCorrection->updateAttitude(newAttitude, newAttitude, -collisionVector, newForward);
    
        // Now update our field
        newForward = -collisionVector;

        // Action B: Hack to slow down on collision
        newSpeed *= 0.3;

        return;
    }

    // collision type is 2 at this point
    
    // Step 1: Handle prey
    
    if (prey==this) {
        // Action A: change direction
        // bounce back if exactly at the same place of colliding target
        if (collisionVector.normalize()==0)
            collisionVector = newForward;
        // Attitude quaternion must be updated since we change direction
        attitudeCorrection->updateAttitude(newAttitude, newAttitude, -collisionVector, newForward);
        // Now update our field
        newForward = -collisionVector;

        // Action C: escape!
        steeringForce = newForward * getMaxForce();
        
    } else {
        // Action A: change direction
        // bounce back if exactly at the same place of colliding target
        if (collisionVector.normalize()==0)
            collisionVector = -prey->forward;
        // Attitude quaternion must be updated since prey changes direction
        // In simulator steps, no need for mutex protect (setAttitude handles protection from OSG already)
        osg::Quat att = prey->getAttitude();
        prey->attitudeCorrection->updateAttitude(att, att, collisionVector, prey->forward);
        // Now update fields
        prey->forward = collisionVector;
        prey->setAttitude(att);

        // Action C: escape!
        prey->steeringForce = prey->forward * prey->getMaxForce();
    }
    // Action D: remove neighbor to avoid multiple collisions record
    for (multimap<const double,Localized*>::iterator itn = prey->neighbors->begin(); itn != prey->neighbors->end(); ++itn) {
        if (itn->second == predator) {
            prey->neighbors->erase(itn);
            break;
        }
    }
    
    // Step 2: Handle predator
    
    // Action B for: Hack to slow down
    if (predator==this) newSpeed *= 0.3;
    else predator->speed *= 0.3;
    
    // Action D: remove neighbor to avoid multiple collisions record
    for (multimap<const double,Localized*>::iterator itn = predator->neighbors->begin(); itn != predator->neighbors->end(); ++itn) {
        if (itn->second == prey) {
            predator->neighbors->erase(itn);
            break;
        }
    }

}

void CrogaiAgent::energyGainAction(double gain)
{
    energy += gain;
    
    // maintain stat
    if (maxEnergyReached<energy) maxEnergyReached = energy;
}

void CrogaiAgent::winnerAction( CrogaiAgent * looser )
{
    // Digestion is 100% perfect. All mass converted into energy!
    energyGainAction(looser->mass / getEnergyFootprintRatio());
}

void CrogaiAgent::looserAction( CrogaiAgent * winner )
{
    // do nothing by default
}

void CrogaiAgent::updateChild( CrogaiAgent * child )
{
    // update child generation
    child->generation = generation + 1;

    // compute stats, if requested
    if (logStatistics) {
        
        vector<int> *vi = &generationsPopulationNumber[specie];
        if (child->generation >= vi->size()) vi->resize(child->generation+1, 0);
        ++(*vi)[child->generation];

        vector<double> *vd = &generationsPopulationFirstTime[specie];
        if (child->generation >= vd->size()) vd->resize(child->generation+1, 0.0);
        if ((*vd)[child->generation] == 0.0) (*vd)[child->generation] = simu->getTime();

/*        
        if (child->generation > generations[specie]) {
            generations[specie] = child->generation;
            
            if (displayGenerations) {
                ostringstream os;
                os << "Generations: ";
                for (map<int, int>::iterator it = generations.begin(); it != generations.end(); ++it) {
                    if (it!=generations.begin()) os << ", ";
                    os << it->first << ":" << it->second;
                }
                Utility::setStatus(os.str());
            }
            
            if (logGenerations!=0) {
                Logger::iterator lit = logGenerations->begin();
                *lit++ = simu->getTime();
                for (map<int, int>::iterator it = generations.begin(); it != generations.end(); ++it) {
                    *lit++ = it->second;
                }
                logGenerations->record();
            }
        }
*/
            
        // Update number of offsprings this agent has spawned
        ++offspringQuantity;
    }
    
    double px = getPosition().x();
    double py = getPosition().y();

    // Place child somewhere around us.
    double angle = 6.2831853 * Utility::Random();
    double x = px + cos(angle) * getRadius() * 2.0;
    double y = py + sin(angle) * getRadius() * 2.0;

    // TODO: check for obstacles...
    // TODO: cyclicity all right?
    
    // Position Agent at floor surface
    child->setPosition(x,y, terrain->getFloorZ(x,y));
    child->setAttitude(getAttitude());
    child->forward = forward;
    // update up vector: same forward, attitude and current up as we have, but different target Up
    ((KeepUpAttitudeCorrection*)child->attitudeCorrection)->set(
        terrain->getFloorNormal(x,y),
        ((KeepUpAttitudeCorrection*)attitudeCorrection)->getUpVector()
    );
    // update attitude for new target
    ((KeepUpAttitudeCorrection*)child->attitudeCorrection)->matchTarget(child->attitude, child->forward);
    // force transform update now
    //child->updateTransform();

    // Give some speed to child so it's not immobile: take 10% of our kinetic energy
//    child->speed = (mass + energy * getEnergyFootprintRatio()) * speed * 0.1 / (child->mass + child->energy * child->getEnergyFootprintRatio());
//    speed *= 0.9;


    

}


void CrogaiAgent::findNeighbors( double radius )
{
    // call parent function
    TerrainAgent::findNeighbors(radius);

    // done if general case
    if (!seeOnlyInFront) return;
    
    // see only what's in front of us: eliminate agents behind
    // STL iterator should still be OK after erase as only the erased element is invalidated (order kept)
    for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ) {

        if (wrappedDiffVector(it->second->getPosition(), getPosition()) * forward < 0)
            neighbors->erase(it++);
        else
            ++it;
    }
}

bool CrogaiAgent::isPrey( CrogaiAgent * agent )
{
    return agent->specie != specie;
}

bool CrogaiAgent::isPredator( CrogaiAgent * agent )
{
    return agent->specie != specie;
}

double CrogaiAgent::getEnergyForFight( )
{
    return energy;
}

void CrogaiAgent::updateStatistics( )
{
    double age =  ( (simu!=0) ? simu->getTime() : 0.0 ) - birthDate;
    vector<double> *v;

    v = &lifeTimeMean[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += age;

    v = &lifeTimeVariance[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += age * age;

    v = &maxEnergyReachedMean[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += maxEnergyReached;

    v = &maxEnergyReachedVariance[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += maxEnergyReached * maxEnergyReached;

    v = &offspringQuantityMean[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += offspringQuantity;

    v = &offspringQuantityVariance[specie];
    if (generation >= v->size()) v->resize(generation+1, 0.0);
    (*v)[generation] += offspringQuantity * offspringQuantity;
}

static int logNumberOfSpecies = 0;

void CrogaiAgent::createStatisticsLogger( const std::string & id, int numSpecies )
{
    logNumberOfSpecies = numSpecies;
    vector<string> names;
    names.push_back("generations");
    for (int i=1; i<=numSpecies; ++i) {
        ostringstream os;
        os << "LifeTimeMeanSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "LifeTimeStandardDeviationSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "MaxEnergyReachedMeanSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "MaxEnergyReachedStandardDeviationSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "OffspringQuantityMeanSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "OffspringQuantityStandardDeviationSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "GenerationsPopulationNumberSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
        os.str("");
        os << "GenerationsPopulationFirstTimeSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
    }
    logStatistics = Utility::createLogger(id, names);
}

void CrogaiAgent::recordStatistics( )
{
    if (!logStatistics) return;

    // TODO. Maybe change structure
    
    // first step: get max gen
    int maxGeneration = 0;
    vector<double> *v;

    for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {

        v = &lifeTimeMean[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &lifeTimeVariance[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &maxEnergyReachedMean[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &maxEnergyReachedVariance[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &offspringQuantityMean[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &offspringQuantityVariance[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
    }

    // second step: make each vector match max size
    for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {
        lifeTimeMean[specie].resize(maxGeneration, 0.0);
        lifeTimeVariance[specie].resize(maxGeneration, 0.0);
        maxEnergyReachedMean[specie].resize(maxGeneration, 0.0);
        maxEnergyReachedVariance[specie].resize(maxGeneration, 0.0);
        offspringQuantityMean[specie].resize(maxGeneration, 0.0);
        offspringQuantityVariance[specie].resize(maxGeneration, 0.0);
        generationsPopulationNumber[specie].resize(maxGeneration, 0);
        generationsPopulationFirstTime[specie].resize(maxGeneration, 0.0);
    }

    // now we can safely log variables
    for (int generation = 0; generation<maxGeneration; ++generation) {
        int varIdx = 0;
        (*logStatistics)[varIdx++] = generation;
        for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {
            double n = (generationsPopulationNumber[specie])[generation];
            if (n==0.0) n=1.0;

            double ninv = 1.0 / n;

            double m = (lifeTimeMean[specie])[generation] * ninv;
            double v = (lifeTimeVariance[specie])[generation];
            (*logStatistics)[varIdx++] = m;
            (*logStatistics)[varIdx++] = sqrt(v * ninv - m * m);

            m = (maxEnergyReachedMean[specie])[generation] * ninv;
            v = (maxEnergyReachedVariance[specie])[generation];
            (*logStatistics)[varIdx++] = m;
            (*logStatistics)[varIdx++] = sqrt(v * ninv - m * m);
            
            m = (offspringQuantityMean[specie])[generation] * ninv;
            v = (offspringQuantityVariance[specie])[generation];
            (*logStatistics)[varIdx++] = m;
            (*logStatistics)[varIdx++] = sqrt(v * ninv - m * m);
            
            (*logStatistics)[varIdx++] = n;
            (*logStatistics)[varIdx++] = (generationsPopulationFirstTime[specie])[generation];
        }
        // record an entry
        logStatistics->record();
    }
    
}

void CrogaiAgent::createPopulationsLogger( const std::string & id, int numSpecies )
{
    vector<string> names;
    names.push_back("time");
    for (int i=1; i<=numSpecies; ++i) {
        ostringstream os;
        os << "PopulationSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
    }
    logPopulations = Utility::createLogger(id, names);
    
    // add stats for first generation
    for (int i=1; i<=numSpecies; ++i) {
        generationsPopulationNumber[i].resize(1);
        (generationsPopulationNumber[i])[0] = populations[i];
        generationsPopulationFirstTime[i].resize(1);
        (generationsPopulationFirstTime[i])[0] = 0.0;
    }
    
}

void CrogaiAgent::createScoresLogger( const std::string & id, int numSpecies )
{
    vector<string> names;
    names.push_back("time");
    for (int i=1; i<=numSpecies; ++i) {
        ostringstream os;
        os << "ScoreSpecie" << i;
        names.push_back(os.str()); // put name in vector and reset it
    }
    logScores = Utility::createLogger(id, names);

    // Initialize score map
    for (int i=1; i<=numSpecies; ++i) scores[i] = 0;
}

void CrogaiAgent::addPopulationUpdateCallback( void (* callback)( int delta, int specie ) )
{
    populationUpdateCallbacks.push_back(callback);
}


static int populationMaxSamples = 0;
static map<int, vector< pair<double, double> > > populationSamples;
static map<int, double> meanSampleTime;
static map<int, double> minSampleTime;
static map <int, osg::Geometry*> graphs;

static map<int, OpenThreads::Mutex*> populationSamplesAccess;
static OpenThreads::Mutex populationSamplesDirtyFlagAccess;
static bool populationSamplesDirtyFlag = false;
static osg::Projection* graphHUD = 0;
static osg::MatrixTransform* graphsTransform = 0;

// forward declaration
static void updatePopulationGraphs();

static void addSceneHelper(osg::Projection* p) {
    osg::Group* topmost = dynamic_cast<osg::Group*>(viewer->getTopMostSceneData());
    if (topmost==0) return;
    topmost->addChild(p);
    graphHUD = p;
    // initial update
    updatePopulationGraphs();
}

static void removeHUDHelper() {
    osg::Group* topmost = dynamic_cast<osg::Group*>(viewer->getTopMostSceneData());
    if (topmost==0) return;
    topmost->removeChild(graphHUD);
    graphHUD = 0;
}

void CrogaiAgent::displayPopulationGraph( bool status )
{
    // batch mode, or graphics not yet set up
    if (batch || (viewer==0)) return;

    if (!status) {
        if (viewer && graphHUD) viewer->post(&removeHUDHelper);
        return;
    }

    // Activate already activated hud? => nothing to do
    if (graphHUD) return;

    // Sample collection not activated? => no graph
    if (populationSamples.empty()) return;
    
    osg::Group* topmost = dynamic_cast<osg::Group*>(viewer->getTopMostSceneData());
    if (topmost==0) return;

    // Scale the graph here rather than in update
    osg::Geode *graphsNode = new osg::Geode();
    osg::StateSet* stateset3 = graphsNode->getOrCreateStateSet();
    stateset3->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
    stateset3->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
    stateset3->setMode(GL_LINE_SMOOTH,osg::StateAttribute::ON);
    stateset3->setMode(GL_BLEND,osg::StateAttribute::ON);
    stateset3->setRenderBinDetails(12,"RenderBin");
    //stateset3->setAttribute(osg::StateAttribute::ANTIALIAS,osg::StateAttribute::ON);
    graphsTransform = new osg::MatrixTransform;
    graphsTransform->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
    graphsTransform->setMatrix(osg::Matrix::identity());
    graphsTransform->addChild(graphsNode);

    osg::Projection* projection = new osg::Projection;
    projection->setMatrix(osg::Matrix::ortho2D(0,WindowWidth,0,WindowHeight));
    projection->addChild(graphsTransform);

    
    // Create the graphs
    for (map<int, double>::iterator it = minSampleTime.begin(); it!=minSampleTime.end(); ++it) {
        osg::Geometry* graph = new osg::Geometry;
        graph->setUseDisplayList(false);
        graph->addPrimitiveSet(new osg::DrawArrays(osg::PrimitiveSet::LINE_STRIP));
        osg::Vec3Array* coords = new osg::Vec3Array;
        coords->resize(populationMaxSamples);
        graph->setVertexArray(coords);
        osg::DrawArrays *da = dynamic_cast<osg::DrawArrays*>(graph->getPrimitiveSet(0));
        // Initialy draw no point
        if( da != 0 )
        {
            da->setFirst(0);
            da->setCount(0);
        }
        // Add this graph
        osg::Vec4Array* colors = new osg::Vec4Array;
        colors->push_back(CrogaiAgent::getDefaultColorForSpecie(it->first));
        graph->setColorArray(colors);
        graph->setColorBinding(osg::Geometry::BIND_OVERALL);
        graphsNode->addDrawable(graph);
        graphs[it->first] = graph;
    }

    viewer->post(&addSceneHelper, projection);
}

void CrogaiAgent::togglePopulationGraph( )
{
    displayPopulationGraph(!graphHUD);
}

static void updatePopulationGraphs() {

    // no lock necessary: graphHUD always modified from viewer thread
    if (!graphHUD) return;

    double maxTime = 0.0;
    double maxPop = 0.0;

    for (map<int, vector< pair<double, double> > >::iterator it = populationSamples.begin(); it != populationSamples.end(); ++it) {

        vector< pair<double, double> > & samples = it->second;
        osg::Geometry* graph = graphs[it->first];
        osg::Vec3Array *coords = dynamic_cast<osg::Vec3Array *>( graph->getVertexArray() );
        if (!coords) continue;
        osg::PrimitiveSet *ps = graph->getPrimitiveSet(0);
        osg::DrawArrays *da = dynamic_cast<osg::DrawArrays*>(ps);
        if (!da) continue;

        osg::Vec3Array::iterator cit = coords->begin();

        populationSamplesAccess[it->first]->lock();
        for (vector< pair<double, double> >::iterator sit = samples.begin(); sit!=samples.end(); ++sit) {
           cit->set(sit->first, sit->second, 0.0);
           if (sit->second > maxPop) maxPop = sit->second;
           ++cit;
        }
        if (samples.back().first>maxTime) maxTime = samples.back().first;
        populationSamplesAccess[it->first]->unlock();
        

        da->setFirst(0);
        da->setCount( samples.size() );
    }

    double xoffset = WindowWidth*0.03;
    double xscale = (maxTime!=0.0) ? (WindowWidth*0.94 / maxTime) : 0.0;
    double yoffset = WindowHeight*0.1;
    double yscale = (maxPop!=0.0) ? (WindowHeight*0.8 / maxPop) : 0.0;
    osg::Matrix mat;
    mat.makeScale(xscale, yscale, 1.0);
    mat.setTrans(xoffset, yoffset, 0.0);
    graphsTransform->setMatrix(mat);

    populationSamplesDirtyFlagAccess.lock();
    populationSamplesDirtyFlag = false;
    populationSamplesDirtyFlagAccess.unlock();

}

static void samplePopulationCallback(int delta, int specie) {

    populationSamplesAccess[specie]->lock();
    
    vector< pair<double, double> > & samples = populationSamples[specie];

    // Resample when list is full
    if (samples.size() == populationMaxSamples) {

        // Remove all samples for which delta < meanSampleTime
        // Then use the mean as new min!

        // use new vector, then swap
        vector< pair<double, double> > newSamples;
        newSamples.reserve(populationMaxSamples);

        // first finish mean computation
        double mean = meanSampleTime[specie] / (double)populationMaxSamples;

        // now iterate & remove samples
        vector< pair<double, double> >::iterator it = samples.begin();
        double lastTime = it->first;
        double newMean = 0.0;
        newSamples.push_back(*it);
        ++it;
        for (; it!=samples.end(); ++it) {

            // ignore too short-timed sample
            if (it->first - lastTime < mean) continue;

            // Keep this sample
            newSamples.push_back(*it);
            // Update stats
            newMean += it->first - lastTime;
            lastTime = it->first;
        }
        // swap vectors to validate new samples. Old samples will be discarded
        samples.swap(newSamples);
        // update mean and min!
        minSampleTime[specie] = mean;
        meanSampleTime[specie] = newMean;
    }

    double now = simu->getTime();
    double deltaTime = now - samples.back().first;

    // ignore if new sample is lower than min time, or update
    if (deltaTime > minSampleTime[specie]) {
        // so, insert new sample & keep stats
        populationSamples[specie].push_back(make_pair(now, (double)CrogaiAgent::populations[specie]));
        meanSampleTime[specie] += deltaTime;

        populationSamplesDirtyFlagAccess.lock();
        if ((!populationSamplesDirtyFlag) && (graphHUD!=0)) {
            viewer->post(&updatePopulationGraphs);
        }
        populationSamplesDirtyFlag = true;
        populationSamplesDirtyFlagAccess.unlock();
    }
    
    populationSamplesAccess[specie]->unlock();
}

void CrogaiAgent::samplePopulationHistory( int maxSamples, int nspecies )
{
    if (maxSamples==0) {
        for (vector<PopulationUpdateCallback>::iterator it = populationUpdateCallbacks.begin();
        it != populationUpdateCallbacks.end();) {
            PopulationUpdateCallback callback = *it;
            if (callback == &samplePopulationCallback) it = populationUpdateCallbacks.erase(it);
            else ++it;
        }
        return;
    }

    if (populationMaxSamples>0) {
        for (map<int, OpenThreads::Mutex*>::iterator it = populationSamplesAccess.begin();
        it != populationSamplesAccess.end(); ++it) delete it->second;
    }
    populationMaxSamples = maxSamples;
    addPopulationUpdateCallback(&samplePopulationCallback);

    double now = simu->getTime();

    for (int i=1; i<=nspecies; ++i) {
        populationSamples[i].clear();
        populationSamples[i].reserve(populationMaxSamples);
        // add at least one element to begin with, for delta time computations in the callback
        populationSamples[i].push_back(make_pair(now, (double)populations[i]));
        meanSampleTime[i] = 0.0;
        minSampleTime[i] = 0.0;
        populationSamplesAccess[i] = new OpenThreads::Mutex;
    }
}

void CrogaiAgent::recordPopulationSamples( const std::string & baseId )
{
    
    for (map<int, vector< pair<double, double> > >::iterator it = populationSamples.begin(); it != populationSamples.end(); ++it) {
        // create a logger
        vector<string> names;
        names.push_back("sampleTime");
        ostringstream os;
        os << "samplePopulationSpecie" << it->first;
        names.push_back(os.str());
        os.str("");
        os << baseId << "Specie" << it->first;
        Logger* log = Utility::createLogger(os.str(),names);

        // Record samples for this specie
        for (vector< pair<double, double> >::iterator sit = it->second.begin(); sit!=it->second.end(); ++sit) {
            (*log)[0] = sit->first;
            (*log)[1] = sit->second;
            log->record();
        }
        delete log;
    }
}

