/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef CROGAIAGENT_H
#define CROGAIAGENT_H

#include <map>
#include <list>
#include <osg/ShapeDrawable>
#include <osg/Geometry>

#include "simulator.h"

#include "TerrainAgent.h"

#include "Logger.h"

/** Base agent for the Crogai scenario:
- Adds a notion of specie, and associated avatar. The specie should be a value >0, 0 is reserved for obstacles.
- Handle a common integration step, with policy for agent collision (prey/predator).
- AI is left open for subclasses, the whole game is to make the best one!
- Any such agent can be a teacher for another agent. Provide the necessary hooks.
- Subclasses may create children agents when requested by the system, for artificial evolution...

@author Nicolas Brodu
*/
class CrogaiAgent : public TerrainAgent
{
public:

    /** Provide a default specie different from obstacles */
    CrogaiAgent(int specie = 1);

    /** Specie of this agent, to make the prey/predator game.
        This should be fixed, except in a "vampire" or conversion
        scenario where the specie could be altered at run-time...
        This could also be called team, or class, or type.
    */
    int specie;
    
    /** Generation of this agent. This is initially 0, and will be updated
        by incrementing the parent generation in the "updateChild" method
    */
    int generation;

    /** Number of offsprings this agent has spawned. This is initially 0, 
        and will be updated by incrementing the parent generation in
        the "updateChild" method
    */
    int offspringQuantity;

    /// Birth date of this agent, in simulator time units. Used to compute statistics
    double birthDate;

    /// The default color for the given specie
    static osg::Vec4 getDefaultColorForSpecie(int specie);
    static std::string getDefaultColorNameForSpecie( int specie );

    /** Score for each specie. Key is specie, value is score.
    */
    static std::map<int, int> scores;
    
    /// Whether to display score counters on top left. Default is false
    static bool displayScores;
    
    /// Whether to log score values. Default is null, so no log.
    static Logger* logScores;
    
    /// Creates the logScores with the given ID and max number of species
    /// utility routine
    static void createScoresLogger(const std::string& id, int numSpecies);
    
    /** Population for each specie. Key is specie, value is population count.
    */
    static std::map<int, int> populations;
    
    /// Whether to display population statistics on top right. Default is true
    static bool displayPopulationNumbers;
    
    /// Whether to display population counter on top left. Default is true
    static bool displayPopulationCounter;
    
    /// Whether to log population values. Default is null, so no log.
    static Logger* logPopulations;
    
    /// Creates the logPopulations with the given ID and max number of species
    /// utility routine
    static void createPopulationsLogger(const std::string& id, int numSpecies);
    
//    /** Record the maximum generation reached by each specie */
//    static std::map<int, int> generations;
    
//    /// Whether to display generation counters in the status bar. Default is true
//    static bool displayGenerations;
    
//    /// Whether to log generation values. Default is null, so no log.
//    static Logger* logGenerations;

    /// How long an agent lives on average
    static std::map<int, std::vector<double> > lifeTimeMean;
    static std::map<int, std::vector<double> > lifeTimeVariance;
    
    /// The maximum energy an agent has had. High count mean it has eaten a lot a preys
    static std::map<int, std::vector<double> > maxEnergyReachedMean;
    static std::map<int, std::vector<double> > maxEnergyReachedVariance;

    /// number of offsprings an agent produced during its life
    static std::map<int, std::vector<double> > offspringQuantityMean;
    static std::map<int, std::vector<double> > offspringQuantityVariance;

    /// Keep track of how many agents were produced in each specie, for each generation.
    /// Necessary to normalize the previous mean/variance
    static std::map<int, std::vector<int> > generationsPopulationNumber;
    
    /// Keep also track of when the first agent of a given generation was produced
    static std::map<int, std::vector<double> > generationsPopulationFirstTime;

    /// Creates a logger with the given ID and max number of species
    /// Use this to monitor the previous statistics, or by default they won't even be computed
    /// This routine also initializes the statistics for the generation 0
    /// Thus, you should first create all the initial agents, then trigger the log.
    static void createStatisticsLogger(const std::string& id, int numSpecies);

    /// Tells the logger to record all its statistics now. It won't be possible to change
    /// the values afterward, so this should normally be called at the end of your program
    static void recordStatistics();
    
    /// The statistics logger. Please use createStatisticsLogger to allocate it, unless
    /// you known what you're doing.
    /// These can't be updated in real time, and
    /// generations co-exist at the same time and their statistics may be
    /// updated in parallel (which would require to "go back" in the log file
    /// to update a value). => the logger is used only at the end of the program,
    /// when the last agent is destroyed, but it is also used as a flag indicator
    /// whether to bother computing the statistics at all!
    /// null by default
    static Logger* logStatistics;

    /// Called from destructor, to update these statistics that can only be known after agent death
    /// May be called explicitly in some special cases, though
    void updateStatistics();


    /// TODO: Revamp Simulator interface system and re-use dispatchers both here, in EventQueueViewer,
    /// and for simu->postXXX including simu->postAtStop
    /// Ugly hack for immediate needs!
    static void addPopulationUpdateCallback(void (*callback)(int delta, int specie));

    /** Maintain a history of the population, for display purposes.
        This uses an algorithm to automatically re-sample the population history
        when it becomes too large.
        @param maxSamples indicates how much point to keep, at most. Use 0 to disable sampling completely, and thus return to the default behaviour.
    */
    static void samplePopulationHistory(int maxSamples, int nspecies);

    static void recordPopulationSamples(const std::string& baseId);

    /// Whether to display the population graph, or not, when not in batch mode.
    /// May be called any time.
    static void displayPopulationGraph(bool status);
    static void togglePopulationGraph();
    
    /// Whether agents see only in front of them. True by default.
    static bool seeOnlyInFront;

    
    //double message; // message to the world

    /// Add a student to this teacher
    virtual void addStudent(CrogaiAgent* student);
    
    /// Remove a student to this teacher - It has finished training
    virtual void removeStudent(CrogaiAgent* student);
    
    /** Hook into AI stub to notifies all students. Students are notified
        with the learnAI routine, to allow them to make the difference.
        If overloading aiStep, Subclasses should call this before their own routine.
        Note: Subclasses should use updateAI as usual and not worry
    */
    virtual void aiStep();
    
    /// overload parent to use eliminate agents behind this one: it sees
    /// only what's in front, if this option was choosen (see 'seeOnlyInFront')
    /// TODO: integrate this directly in the locality query, so it's possible
    ///       to quickly eliminate bricks completely behind
    ///       => might expect an average of 50% perf increase!!!
    virtual void findNeighbors(double radius);

    
protected:

/////////////////////
/// Subclasses API
/////////////////////

    /** Override this to specify an action to do when an agent gains energy
        Default is to increase our own energy with this amount, and maintain max energy stats.
    */
    virtual void energyGainAction(double gain);
    
    /** Override this to specify an action to do when an agent wins over another one
        Default is to acquire energy equivalent to looser mass.
        Note: Looser destruction and score counting are already handled outside this function
        You should on the other hand use the parent function to handle energy
    */
    virtual void winnerAction(CrogaiAgent* looser);

    /** Override this to specify an action to do when an agent loses, before it is destroyed.
        Default is to do nothing. Use it to do an action each time before an agent is destroyed,
        and use winnerAction only for score counting and such other actions that may miss destroyed
        agents (see note 2).
        Note: Looser destruction and score counting are already handled outside this function.
        Note 2: There is a need for this function because there may be cases where a draw occurs,
                and there is no winner. Thus winnerAction is not called in these cases.
        @param winner The winner agent, or null if this is a draw
    */
    virtual void looserAction(CrogaiAgent* winner);

    /** Handle collision detection in addition to parent's terrain handling
        Subclasses should call this before their own routine, if any
    */
    virtual void updateIntegration(const double elapsedTime, const double currentTime);

    /** For learning subclasses to be notified when this agent leaves the game
    */
    virtual void removeTeacher(CrogaiAgent* teacher);
    
    /** Students registered on this teacher should have overloaded this.
        This is the function called during teacher updateAI to notify all
        students.
        It does nothing by default, and is never called if your object is
        not registered as a student.
    */
    virtual void learnAI(const double elapsedTime, const double currentTime, CrogaiAgent* teacher);


    /** Allow subclasses to identify preys based on their own criteria.
        Used during collision.
        By default, consider any agent of different specie as prey.
    */
    virtual bool isPrey(CrogaiAgent* agent);
    
    /** Allow subclasses to identify preys based on their own criteria.
        Used during collision.
        By default, consider any agent of different specie as predator.
    */
    virtual bool isPredator(CrogaiAgent* agent);

    /** Return the amount of energy this predator is willing to spend in a fight with the prey.
        By default, set no limit (return the current amount of energy) and thus fight to
        death even if the prey escapes (which is pretty desperate for a predator).
    */
    virtual double getEnergyForFight();
    
/////////////////////
/// Internal use
/////////////////////
    
    /** Subclasses may call this function when spawning to set base class parameters
        This function should copy all base parameters to the child, then do mutations.
        This base class function position the child somewhere aroung the parent, but does nothing else.
        @param child a newly created instance of a subclass for which base fields should be updated.
    */
    virtual void updateChild(CrogaiAgent* child);

    osg::Node * buildAvatar();

    // Access to this member is necessary to change the color at run-time
    //osg::ShapeDrawable *body;

    /// shared instance of box
    static osg::Box* crogaiAgentShape;

    std::list<CrogaiAgent*> students;
    
    virtual ~CrogaiAgent();
    
    /// Used for statistics purpose
    double maxEnergyReached;

public:    
    static double defaultSearchRadius;
    double searchRadius;
    // This member allows to cancel AI of dead preys
    const Simulator::GenericDispatcher* aiEvent;
    // This member allows to cancel integration of dead preys
    const Simulator::GenericDispatcher* integrationEvent;
};

#endif
