/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef EVENTQUEUEVIEWER_H
#define EVENTQUEUEVIEWER_H

#include <osgProducer/Viewer>
#include <OpenThreads/Mutex>
#include <vector>

/** Extend the OSG viewer to add a thread-safe asynchronous event queue.

    The goal is to post-pone OSG-critical events (like scene graph changes) to
    when they can be safely executed (the update method of the viewer).

    Any thread can post such events at any time, they will be executed in order
    they were received at the next scene graph update.
        
    Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
    template way of posting events was "inspired" from a more complete simulator framework.
    See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.
    
    @author Nicolas Brodu
*/
class EventQueueViewer : public osgProducer::Viewer
{
public:

    /** Parent constructor "overload". See parent doc. */
    EventQueueViewer();
        
    /** Parent constructor "overload". See parent doc. */
    EventQueueViewer(Producer::CameraConfig *cfg);

    /** Parent constructor "overload". See parent doc. */
    EventQueueViewer(const std::string& configFile);

    /** Parent constructor "overload". See parent doc. */
    EventQueueViewer(osg::ArgumentParser& arguments);

    /** Parent destructor "overload". See parent doc. */
    virtual ~EventQueueViewer();


    /** Base class of templatized object method wrappers
        This is put in the public namespace of the
        Viewer so it can be used with the generic post methods.
        Most probably you won't need it for normal operation, templates handle up to three arguments.
    */
    struct GenericDispatcher {
        // avoid templates from creating a weak symbol for this
        GenericDispatcher();
    
        /// Dispatch the event.
        virtual void dispatch() const;
    
        // necessary for operator delete to work correctly
        virtual ~GenericDispatcher();
    };
    
    // Necessary for C++ contravariance rule for pointers to members
    template<class T> struct Dispatcher : public GenericDispatcher {
        typedef void (T::* TFunc)(); // pointer to member type
        T* pointer;
        TFunc function;
        inline Dispatcher(T* t, TFunc f) : pointer(t), function(f) {}
        virtual void dispatch() const {
            (pointer->*function)();
        }
    
    };
    template<class T, class A> struct Dispatcher1 : public GenericDispatcher {
        typedef void (T::* TFunc)(A a); // pointer to member type with one argument
        T* pointer;
        TFunc function;
        A arg;
        inline Dispatcher1(T* t, TFunc f, A a) : pointer(t), function(f), arg(a) {}
        virtual void dispatch() const {
            (pointer->*function)(arg);
        }
    };
    template<class T, class A1, class A2> struct Dispatcher2 : public GenericDispatcher {
        typedef void (T::* TFunc)(A1,A2); // pointer to member type with two arguments
        T* pointer;
        TFunc function;
        A1 arg1;
        A2 arg2;
        inline Dispatcher2(T* t, TFunc f, A1 a1, A2 a2) : pointer(t), function(f), arg1(a1), arg2(a2) {}
        virtual void dispatch() const {
            (pointer->*function)(arg1,arg2);
        }
    };
    template<class T, class A1, class A2, class A3> struct Dispatcher3 : public GenericDispatcher {
        typedef void (T::* TFunc)(A1,A2,A3); // pointer to member type with three arguments
        T* pointer;
        TFunc function;
        A1 arg1;
        A2 arg2;
        A3 arg3;
        inline Dispatcher3(T* t, TFunc f, A1 a1, A2 a2, A3 a3) : pointer(t), function(f), arg1(a1), arg2(a2), arg3(a3) {}
        virtual void dispatch() const {
            (pointer->*function)(arg1,arg2,arg3);
        }
    };
    
    // plain C callbacks
    struct StaticDispatcher : public GenericDispatcher {
        typedef void (*Func)(); // pointer to static func
        Func function;
        inline StaticDispatcher(Func f) : function(f) {}
        virtual void dispatch() const {
            (*function)();
        }
    
    };
    template<class A> struct StaticDispatcher1 : public GenericDispatcher {
        typedef void (*Func)(A a); // pointer to C function with one argument
        Func function;
        A arg;
        inline StaticDispatcher1(Func f, A a) : function(f), arg(a) {}
        virtual void dispatch() const {
            (*function)(arg);
        }
    };
    template<class A1, class A2> struct StaticDispatcher2 : public GenericDispatcher {
        typedef void (*Func)(A1,A2); // pointer to C function with two arguments
        Func function;
        A1 arg1;
        A2 arg2;
        inline StaticDispatcher2(Func f, A1 a1, A2 a2) : function(f), arg1(a1), arg2(a2) {}
        virtual void dispatch() const {
            (*function)(arg1,arg2);
        }
    };
    template<class A1, class A2, class A3> struct StaticDispatcher3 : public GenericDispatcher {
        typedef void (*Func)(A1,A2,A3); // pointer to C function with three arguments
        Func function;
        A1 arg1;
        A2 arg2;
        A3 arg3;
        inline StaticDispatcher3(Func f, A1 a1, A2 a2, A3 a3) : function(f), arg1(a1), arg2(a2), arg3(a3) {}
        virtual void dispatch() const {
            (*function)(arg1,arg2,arg3);
        }
    };

    
    /** Post an event, scheduled to be executed at next frame update.

        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        Note2: In this implementation, the dispatcher MUST have been allocated with new. This is the case with all
               templatized versions of post, but you've been warned if you use your own dispatcher.
        
        See parent "update" method too.
    */
    virtual void post(const GenericDispatcher* dispatcher);
    
    /** Post an event, scheduled to be executed at next frame update.
        Any C "void method()" can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param method any C "void method()" with no argument
     */
    inline void post(void (*method)()) {
        post(new StaticDispatcher(method));
    }

    /** Post an event, scheduled to be executed at next frame update.
        Any C "void method(..)" with one argument can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param method any C "void method()" with one argument
        @param argument the argument to pass to the method
     */
    template<class A> inline void post(void (*method)(A), A argument) {
        post(new StaticDispatcher1<A>(method, argument));
    }

    /** Post an event, scheduled to be executed at next frame update.
        Any C "void method(..)" with two arguments can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param method any C "void method(..)" with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class A1, class A2> inline void post(void (*method)(A1,A2), A1 argument1, A2 argument2) {
        post(new StaticDispatcher2<A1,A2>(method, argument1, argument2));
    }

    /** Post an event, scheduled to be executed at next frame update.
        Any C "void method(..)" with three arguments can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param method any C "void method(..)" with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class A1, class A2, class A3> inline void post(void (*method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        post(new StaticDispatcher3<A1,A2,A3>(method, argument1, argument2, argument3));
    }
    
    /** Post an event, scheduled to be executed at next frame update.
        Any C++ method on any arbitrary object can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param object any C++ object
        @param method any C++ method on this object with no argument
     */
    template<class T> inline void post(T* object, void (T::* method)()) {
        post(new Dispatcher<T>(object, method));
    }
    
    /** Post an event, scheduled to be executed at next frame update.
        Any C++ method on any arbitrary object can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param object any C++ object
        @param function any C++ method on this object with one argument
        @param argument the argument to pass to the method
     */
    template<class T, class A> inline void post(T* object, void (T::* method)(A), A argument) {
        post(new Dispatcher1<T,A>(object, method, argument));
    }
    
    /** Post an event, scheduled to be executed at next frame update.
        Any C++ method on any arbitrary object can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param object any C++ object
        @param function any C++ method on this object with two arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
     */
    template<class T, class A1, class A2> inline void post(T* object, void (T::* method)(A1,A2), A1 argument1, A2 argument2) {
        post(new Dispatcher2<T,A1,A2>(object, method, argument1, argument2));
    }
    
    /** Post an event, scheduled to be executed at next frame update.
        Any C++ method on any arbitrary object can be called this way.
        
        The events are posted in a thread-safe way.
        Use this function to delay code at a place where is it safe for OSG: ex removing a scene graph branch. This code
        may not be safe in you own thread.

        Note: This is not a whole simulator, especially there is no notion of a deterministic simulated time. But the
        template way of posting events was "inspired" from a more complete simulator framework.
        See http://nicolas.brodu.free.fr/en/programmation/crogai/index.html for a project using this simulator.

        See parent "update" method too.
        
        @param object any C++ object
        @param function any C++ method on this object with three arguments
        @param argument1 the first argument to pass to the method
        @param argument2 the second argument to pass to the method
        @param argument3 the third argument to pass to the method
     */
    template<class T, class A1, class A2, class A3> inline void post(T* object, void (T::* method)(A1,A2,A3), A1 argument1, A2 argument2, A3 argument3) {
        post(new Dispatcher3<T,A1,A2,A3>(object, method, argument1, argument2, argument3));
    }

    
    /** Process queued events, then call parent code in globally locked section
        to prevent concurrent modifications of the scene graph.
    */
    virtual void update();

    /** Call parent code in globally locked section to prevent concurrent 
        modifications of the scene graph. */
    //virtual void frame();

    // sync does not need mutex protection, and should not as it blocks and we precisely
    // want the simulator to do work during this sync if it can (reclaim unused CPU)

    /** Globally locks or unlocks the viewer.
        Guarantee that during a protected section, the scenegraph may be safely modified
        Also makes things deterministic, unlike posting an event to the next update
        => use event posts to next update when non-determinism isn't a problem (like changing a color, etc...)
        => use global lock when the order is very important (like a simulator step adding a new agent)
    */
    //inline void lock() {globalAccessMutex.lock();}
    //inline void unlock() {globalAccessMutex.unlock();}
    
protected:
    std::vector<const GenericDispatcher*> events;
    OpenThreads::Mutex eventsAccessMutex;
    OpenThreads::Mutex globalAccessMutex;
};

#endif
