/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <sstream>

#include <osg/Geometry>
#include <osg/Geode>

#include <math.h>

#include "Utility.h"
#include "Obstacle.h"
#include "ExplicitAIAgent.h"
#include "GrassyTerrain.h"
using namespace Environment;
using namespace std;

double ExplicitAIAgent::sporeHalfLife = 120.0;
double ExplicitAIAgent::maxGrazingSpeed = 1.0;
double ExplicitAIAgent::sporeEmissionEnergyThreshold = 1000.0;
double ExplicitAIAgent::sporeEmissionEnergyCost = 10.0;
double ExplicitAIAgent::maxMass = 2.0;
double ExplicitAIAgent::minMass = 0.5;

// Statistics
std::map<int, std::vector<ExplicitAIAgent::Genome*> > ExplicitAIAgent::genomesMean;
std::map<int, std::vector<ExplicitAIAgent::Genome*> > ExplicitAIAgent::genomesVariance;
Logger* ExplicitAIAgent::logGenomes = 0;


ExplicitAIAgent::ExplicitAIAgent( int specie, const std::vector< int >& predatorSpecies, const std::vector< int >& preySpecies )
: CrogaiAgent(specie) {

    this->predatorSpecies = predatorSpecies;
    this->preySpecies = preySpecies;

    genome = new Genome(specie);
    
    double childMass = (maxMass - minMass) * genome->childMassRatio + minMass;
    if (childMass<minMass) childMass = minMass; // pathologic mutation case
    spawnCriticalMass = childMass + minMass;
}

ExplicitAIAgent::~ ExplicitAIAgent( )
{
    // genome may continue to live as spores in the cloud!
    genome->unref();
}

ExplicitAIAgent::Genome::Genome( int specie )
{
    this->specie = specie;
    setHalfLife(ExplicitAIAgent::sporeHalfLife);
    osg::Vec4 color = CrogaiAgent::getDefaultColorForSpecie(specie);
    r = color.x();
    g = color.y();
    b = color.z();
    // Give sensible, but random, weights for each agent
    hungryRatioThreshold = Utility::Random() * 0.3 + 0.4;
    hungryLastResortRatioThreshold = Utility::Random() * 0.05 + 0.03;
    hungryGrassWeight = - Utility::Random() * 0.5; // if no prey & no predator, make sure total is negative
    hungryPreyWeight = Utility::Random() * 0.5 + 0.5; // preys? => preference for hunting
    hungryPredatorWeight = Utility::Random() * 0.5 + 0.25;
    hungryWeightedThreshold = Utility::Random() * 0.2 - 0.1; // some variation
    grazeRate = Utility::Random() * 0.1 + 0.1;
    grazeNextAmountThreshold = Utility::Random() * 0.1 + 0.05;
    grazeForceCoef = Utility::Random() * 0.3 + 0.7;
    huntMateForwardCoef = Utility::Random() * 2.0 - 1.0;
    huntMateEvasionCoef = Utility::Random() * 0.4;
    huntWanderCoef = Utility::Random() * 0.5 + 0.5;
    huntWanderRate = Utility::Random() * 0.5 + 0.1;
    huntAvoidCoef = Utility::Random() * 2.0 + 3.0;
    huntPreyCoef = Utility::Random() + 1.0;
    huntPredatorCoef = Utility::Random() + 1.0;
    huntMateCoef = Utility::Random() * 0.5 + 0.5;
    seekQuietPreyCutoff = Utility::Random() * 0.2 + 0.1;
    seekQuietOtherEvasionCoef = Utility::Random() + 1.0;
    seekQuietWanderCoef = Utility::Random() * 0.5 + 0.5;
    seekQuietWanderRate = Utility::Random() * 0.5 + 0.1;
    seekQuietAvoidCoef = Utility::Random() * 2.0 + 3.0;
    seekQuietPreyCoef = Utility::Random() * 0.5 + 0.1;
    seekQuietPredatorCoef = Utility::Random() + 1.0;
    seekQuietOtherCoef = Utility::Random() * 0.5 + 0.2;
    quietWeightForObstacle = - Utility::Random() * 0.5; // obstacle seen as good !
    quietWeightForUnknown = Utility::Random() + 0.3;
    quietWeightForPrey = - Utility::Random();     // prey seen as good sign!
    quietWeightForPredator = Utility::Random() + 1.0;  // but predator very bad!
    quietWeightForMate = Utility::Random() - 0.5; // mate = ?
    quietThreshold = Utility::Random() * 2.0 + 3.0; // seek places where weighted sum is less than this
    energyForFightRatio = Utility::Random() * 0.5 + 0.25; // spend no more than this proportion of energy in a fight
    energyConvertionRatio = Utility::Random() * 0.5 + 0.25;
    childEnergyRatio = Utility::Random() * 0.5 + 0.25; // pass on this amount of energy to the child, keep the rest
    childMassRatio = Utility::Random() * 0.2; // child initial mass. 0 is minMass, 1 is (maxMass - minMass)
}

ExplicitAIAgent::Genome::Genome()
{
    specie = 0;
    r = 0;
    g = 0;
    b = 0;

    hungryRatioThreshold = 0.0;
    hungryLastResortRatioThreshold = 0.0;
    hungryGrassWeight = 0.0;
    hungryPreyWeight = 0.0;
    hungryPredatorWeight = 0.0;
    hungryWeightedThreshold = 0.0;
    grazeRate = 0.0;
    grazeNextAmountThreshold = 0.0;
    grazeForceCoef = 0.0;
    huntMateForwardCoef = 0.0;
    huntMateEvasionCoef = 0.0;
    huntWanderCoef = 0.0;
    huntWanderRate = 0.0;
    huntAvoidCoef = 0.0;
    huntPreyCoef = 0.0;
    huntPredatorCoef = 0.0;
    huntMateCoef = 0.0;
    seekQuietPreyCutoff = 0.0;
    seekQuietOtherEvasionCoef = 0.0;
    seekQuietWanderCoef = 0.0;
    seekQuietWanderRate = 0.0;
    seekQuietAvoidCoef = 0.0;
    seekQuietPreyCoef = 0.0;
    seekQuietPredatorCoef = 0.0;
    seekQuietOtherCoef = 0.0;
    quietWeightForObstacle = 0.0;
    quietWeightForUnknown = 0.0;
    quietWeightForPrey = 0.0;
    quietWeightForPredator = 0.0;
    quietWeightForMate = 0.0;
    quietThreshold = 0.0;
    energyForFightRatio = 0.0;
    energyConvertionRatio = 0.0;
    childEnergyRatio = 0.0;
    childMassRatio = 0.0;
}


void ExplicitAIAgent::graze( const double elapsedTime, const double currentTime )
{
    // First get down on earth, if necessary
    if (altitude>radius) {
        steeringForce = -floorNormal;
        return;
    }

    // Then slow down, if moving too fast to graze
    if (speed > maxGrazingSpeed) {
        // Steer for a residual movement, allowing to graze happilly
        steeringForce = steerForTargetSpeed(maxGrazingSpeed * 0.5);
        return;
    }

    GrassyTerrain* yummyGrassFloor = dynamic_cast<GrassyTerrain*>(terrain);

    // Graze floor!
    energyGainAction( yummyGrassFloor->graze(getPosition().x(), getPosition().y(), genome->grazeRate * getMaxEnergy() * elapsedTime));

    // Check what to do for next move
    osg::Vec3d nextPos = wrapPosition(getPosition() + forward * (speed * elapsedTime));
    double nextGraze = yummyGrassFloor->getNormalizedGrassDensity(nextPos.x(), nextPos.y());

    // Don't spend extra energy steering if not necessary and we'll have enough food!
    if (nextGraze > genome->grazeNextAmountThreshold) {
        steeringForce.set(0.0,0.0,0.0);
        return;
    }

    // Now, search for a better place
    double maxDist = getMaxSpeed() * elapsedTime;
    
    // quaternion rotation of local-space X vector
    osg::Vec3d side(attitude.w(), attitude.z(), -attitude.y());

    osg::Vec3d bestMove;
    double bestGraze = -1.0; // negative to ensure first random draw is best

    // Sample surrounding space to find a good place
    for (int i=0; i<5; ++i) {
        double fcoef = Utility::Random();
        if (!seeOnlyInFront) fcoef = fcoef * 2.0 - 1.0; // allow to search backward too if possible
        double scoef = Utility::Random() * 2.0 - 1.0; // search on either side
        osg::Vec3d deltaVec = (forward * fcoef) + (side * scoef);
        deltaVec.normalize(); // may be 0 in extreme case, but does not matter
        nextPos = wrapPosition(getPosition() + deltaVec * maxDist);
        nextGraze = yummyGrassFloor->getGrassDensity(nextPos.x(), nextPos.y());
        if (nextGraze > bestGraze) {
            bestGraze = nextGraze;
            bestMove = deltaVec;
        }
    }

    // Steer to go to this fabulous place!
    // don't move too fast, we'll need to slow down to graze again!
    steeringForce = bestMove * getMaxForce() * genome->grazeForceCoef;
}

void ExplicitAIAgent::hunt( const double elapsedTime, const double currentTime )
{

    // avoidance force, weighted by distance risk
    // use distance squared => risk decrease quadratically, not that inconsistent
    osg::Vec3d avoidForce(0,0,0);
    double avoidWeights = 0.0;

    // seek only one prey at a time, or we may loose both
    osg::Vec3d preyForce(0,0,0);
    bool hasPrey = false;
    
    osg::Vec3d predatorForce(0,0,0);
    double predatorWeights = 0.0;
    
    osg::Vec3d mateForce(0,0,0);
    double mateWeights = 0.0;
    // risk_weight = 0.5  + 0.5 * (squared_max - squared_dist) / squared_max
    // risk_weight = 0.5 * squared_max / squared_max  + 0.5 * (squared_max - squared_dist) / squared_max
    // risk_weight = 0.5 * squared_max / squared_max  + (0.5 *squared_max - 0.5 *squared_dist) / squared_max
    // risk_weight = (0.5 * squared_max + 0.5 *squared_max - 0.5 *squared_dist) / squared_max
    // risk_weight = (squared_max - 0.5 *squared_dist) / squared_max
    // risk_weight = 1.0 - 0.5 * squared_dist / squared_max
    // risk_weight = 1.0 - squared_dist * coef, with coef = 0.5 / squared_max
    const double srqcoef = 1.0 / (searchRadius*searchRadius);
    
    for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

        CrogaiAgent* neighbor = dynamic_cast<CrogaiAgent*>(it->second);

        // if not an agent, maybe it is an obstacle?
        if (!neighbor) {
            
            // Avoid obstacles & other non-crogai agents
            osg::Vec3d force = steerToAvoid(it->second);

            // take into account weighting only if there is a risk
            if (force.length2()>0) {
                double distanceRisk = 1.0 - it->first * srqcoef;
                avoidForce += force * distanceRisk;
                avoidWeights += distanceRisk;
            }
            
            continue;
        }


        // so, we have either a prey, a predator, one of our mates, or a passer-by
        // hunt preys, flee predators, avoid collision with passer-by, and do something for this mate

        if (isPredator(neighbor)) {
            // should flee, but there is a special case:
            // Prey is also predator: Let's try to catch it if it flees, or flee if it wants to catch us!
            if (isPrey(neighbor)) {
                // Catch it if it flees! In scenarii where we see only in front, this also makes sense
                if (neighbor->forward * forward < 0) {
                    // if already have one prey, don't divide attention
                    if (hasPrey) continue;
                    preyForce = steerForPursuit(neighbor);
                    hasPrey = true;
                    continue;
                }
            }
            // Oops, avoid fight
            double distanceRisk = 1.0 - it->first * srqcoef;
            predatorForce += steerForEvasion(neighbor) * distanceRisk;
            predatorWeights += distanceRisk;
            continue;
        }
        
        if (isPrey(neighbor)) {
            // if already have one prey, don't divide attention
            if (hasPrey) continue;
            preyForce = steerForPursuit(neighbor);
            hasPrey = true;
            continue;
        }

        // it's a mate! what to do?
        if (neighbor->specie == specie) {

            double distanceRisk = 1.0 - it->first * srqcoef;
            
            // first try to avoid collision
            osg::Vec3d force = steerToAvoid(neighbor);
            if (force.length2()>0) {
                avoidForce += force * distanceRisk;
                avoidWeights += distanceRisk;
            }

            // now, is there something we could do? group hunt? individual hunt?
            // for now, just head off in different direction than mate
            mateForce += neighbor->forward * (distanceRisk * genome->huntMateForwardCoef);
            mateForce += steerForEvasion(neighbor) * (distanceRisk * genome->huntMateEvasionCoef);
            mateWeights += 2.0*distanceRisk;
            continue;
        }

        // it's a passer-by. Ignore, unless there is a collision risk
        osg::Vec3d force = steerToAvoid(neighbor);
        if (force.length2()>0) {
            double distanceRisk = 1.0 - it->first * srqcoef;
            avoidForce += force * distanceRisk;
            avoidWeights += distanceRisk;
        }

    }
    
    // Use random wander if nothing match
    if ((avoidWeights + predatorWeights + mateWeights == 0.0) && (!hasPrey)) {
        steeringForce = steerForWander(genome->huntWanderRate) * genome->huntWanderCoef;
        return;
    }    

    // Now, average all this!
    steeringForce.set(0,0,0);
    if (avoidWeights>0.0) steeringForce += avoidForce * (genome->huntAvoidCoef / avoidWeights);
    if (hasPrey) steeringForce += preyForce * genome->huntPreyCoef;
    if (predatorWeights>0.0) steeringForce += predatorForce * (genome->huntPredatorCoef / predatorWeights);
    if (mateWeights>0.0) steeringForce += mateForce * (genome->huntMateCoef / mateWeights);
}


void ExplicitAIAgent::searchQuietPlace( const double elapsedTime, const double currentTime )
{
    
    // avoidance force, weighted by distance risk
    // use distance squared => risk decrease quadratically, not that inconsistent
    osg::Vec3d avoidForce(0,0,0);
    double avoidWeights = 0.0;

    // seek only one prey at a time, or we may loose both
    osg::Vec3d preyForce(0,0,0);
    bool hasPrey = false;
    
    osg::Vec3d predatorForce(0,0,0);
    double predatorWeights = 0.0;
    
    osg::Vec3d otherForce(0,0,0);
    double otherWeights = 0.0;
    
    const double srqcoef = 1.0 / (searchRadius*searchRadius);
    
    for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

        CrogaiAgent* neighbor = dynamic_cast<CrogaiAgent*>(it->second);

        // if not an agent, maybe it is an obstacle?
        if (!neighbor) {
            
            // Avoid obstacles & other non-crogai agents
            osg::Vec3d force = steerToAvoid(it->second);

            // take into account weighting only if there is a risk
            if (force.length2()>0) {
                double distanceRisk = 1.0 - it->first * srqcoef;
                avoidForce += force * distanceRisk;
                avoidWeights += distanceRisk;
            }
            
            continue;
        }


        // so, we have either a prey, a predator, one of our mates, or a passer-by
        // flee predators, avoid collision with all others. Don't spent time too
        // much chasing after preys at this point, except maybe if they could be
        // caught unaware and not too far?

        if (isPredator(neighbor)) {
            // should flee, but there is a special case:
            // Prey is also predator: Let's try to catch it if it flees, or flee if it wants to catch us!
            if (isPrey(neighbor)) {
                // Catch it if it flees! In scenarii where we see only in front, this also makes sense
                if (neighbor->forward * forward < 0) {
                    // if already have one prey, don't divide attention
                    if (hasPrey) continue;
                    if (it->first  < genome->seekQuietPreyCutoff * searchRadius* searchRadius) {
                        preyForce = steerForPursuit(neighbor);
                        hasPrey = true;
                    }
                    continue;
                }
            }
            // Oops, avoid fight
            double distanceRisk = 1.0 - it->first * srqcoef;
            predatorForce += steerForEvasion(neighbor) * distanceRisk;
            predatorWeights += distanceRisk;
            continue;
        }
        
        if (isPrey(neighbor)) {
            // if already have one prey, don't divide attention
            if (hasPrey) continue;
            // respect cutoff above which, don't bother chasing prey
            if (it->first  < genome->seekQuietPreyCutoff * searchRadius* searchRadius) {
                // respect scenarii where agents see only in front. In which case, don't engage
                // in fight at this point
                if (seeOnlyInFront && (neighbor->forward * forward >= 0)) continue;
                preyForce = steerForPursuit(neighbor);
                hasPrey = true;
            }
            continue;
        }

        // Flee all other people
        double distanceRisk = 1.0 - it->first * srqcoef;
        
        // first try to avoid collision
        osg::Vec3d force = steerToAvoid(neighbor);
        if (force.length2()>0) {
            avoidForce += force * distanceRisk;
            avoidWeights += distanceRisk;
        }

        // now, is there something we could do? group hunt? individual hunt?
        // for now, just head off in different direction than mate
        otherForce += steerForEvasion(neighbor) * (distanceRisk * genome->seekQuietOtherEvasionCoef);
        otherWeights += distanceRisk;

    }

    // Use random wander if nothing match
    if ((avoidWeights + predatorWeights + otherWeights == 0.0) && (!hasPrey)) {
        steeringForce = steerForWander(genome->seekQuietWanderRate) * genome->seekQuietWanderCoef;
        return;
    }    

    // Now, average all this!
    steeringForce.set(0,0,0);
    if (avoidWeights>0.0) steeringForce += avoidForce * (genome->seekQuietAvoidCoef / avoidWeights);
    if (hasPrey) steeringForce += preyForce * genome->seekQuietPreyCoef;
    if (predatorWeights>0.0) steeringForce += predatorForce * (genome->seekQuietPredatorCoef / predatorWeights);
    if (otherWeights>0.0) steeringForce += otherForce * (genome->seekQuietOtherCoef / otherWeights);
}

bool ExplicitAIAgent::checkForQuietPlace( const double elapsedTime, const double currentTime )
{
    int nobstacles=0, nunknown=0, nprey=0, npredator=0, nmate=0;
    for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

        CrogaiAgent* neighbor = dynamic_cast<CrogaiAgent*>(it->second);

        // if not an agent, maybe it is an obstacle?
        if (!neighbor) {
            if (dynamic_cast<Obstacle*>(it->second) != 0) ++nobstacles;
            else ++nunknown;
            continue;
        }

        if (isPredator(neighbor)) {
            // Prey is also predator: special case
            if (isPrey(neighbor)) {
                // respect scenarii where agents see only in front. In which case, if the predator
                // can't see us, then all is fine!
                if (seeOnlyInFront && (neighbor->forward * forward < 0)) {
                    ++nprey;
                    continue;
                }
            }
            ++npredator;
            continue;
        }
        
        if (isPrey(neighbor)) {
            ++nprey;
            continue;
        }
        
        if (neighbor->specie == specie) {
            ++nmate;
            continue;
        }
        
        ++nunknown;
    }

    return nobstacles * genome->quietWeightForObstacle + nunknown * genome->quietWeightForUnknown + nprey * genome->quietWeightForPrey + npredator * genome->quietWeightForPredator + nmate * genome->quietWeightForMate < genome->quietThreshold;
}


void ExplicitAIAgent::updateAI( const double elapsedTime, const double currentTime )
{
    CrogaiAgent::updateAI(elapsedTime,currentTime);

    findNeighbors(searchRadius);
    
    // Basic AI:
    //    IF not enough energy THEN hunt or graze
    //    ELSE IF good quiet place THEN spawn
    //    ELSE seek good quiet place

    if ((mass < spawnCriticalMass) || (energy < genome->hungryRatioThreshold * getMaxEnergy())) {

        GrassyTerrain* yummyGrassFloor = dynamic_cast<GrassyTerrain*>(terrain);
        if (!yummyGrassFloor) {
            hunt(elapsedTime, currentTime); // no choice, no grass
            return;
        }

        // really, really hungry, then graze. If ratio is negative, then the agent
        // will proudly hunt to its death and keep its dignity, rather than eat grass.
        if (energy < genome->hungryLastResortRatioThreshold * getMaxEnergy()) {
            graze(elapsedTime, currentTime); // no choice, no grass
            return;
        }

        // So, must decide what strategy to do to get energy
        double numPreys = 0.0;
        double numPredators = 0.0;
        double numTotal = 0.0; // for normalized result
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {
            CrogaiAgent* neighbor = dynamic_cast<CrogaiAgent*>(it->second);
            // if not an agent, maybe it is an obstacle?
            if (!neighbor) continue;
            if (isPrey(neighbor)) ++numPreys;
            if (isPredator(neighbor)) ++numPredators;
            ++numTotal;
        }

        // do just a combination with genome coefficients, to decide
        if (numTotal>0.0) numTotal = 1.0 / numTotal;
        if (genome->hungryGrassWeight * yummyGrassFloor->getNormalizedGrassDensity(getPosition().x(), getPosition().y())
          + genome->hungryPreyWeight * numPreys * numTotal
          + genome->hungryPredatorWeight * numPredators * numTotal
          > genome->hungryWeightedThreshold) {

            hunt(elapsedTime, currentTime);
        }
        else {
            graze(elapsedTime, currentTime);
        }
        return;
    }

    // now, check if this is a good quiet place
    findNeighbors(searchRadius);
    if (checkForQuietPlace(elapsedTime, currentTime)) {

        // yes! Are there spores in the environment?
        SporeCloud::ConcentrationList * concentrations = cloud->getConcentrationList(getPosition().x(), getPosition().y(), getPosition().z());

        bool speciesSporesPresent = false;
        
        // create our own copy with only compatible genomes.
        multimap<double, Genome*> compatibleGenomes;
        double totalCompatibleConcentration = 0.0;
        for (SporeCloud::ConcentrationList* l = concentrations; l!=0; l=l->next) {
            if (genome==l->genome) {
                speciesSporesPresent = true;
                continue; // no self-reproduction
            }
            // cast it to this agent genome (Genome in different naming space!)
            Genome* g = dynamic_cast<Genome*>(l->genome);
            if (g==0) continue; // not compatible
            if (g->specie!=specie) continue; // not compatible too
            speciesSporesPresent = true;
            totalCompatibleConcentration += l->concentration;
            compatibleGenomes.insert(make_pair(l->concentration, g));
        }

        
        // yes, we have spores! spawn!!!
        if (totalCompatibleConcentration>0.0) {
            // first get one genome. use concentrations as random distribution for their genomes
            double chosenC = totalCompatibleConcentration * Utility::Random();
            // split total concentration in intervals of lenghts equal to each concentration
            // then see where the chosen on falls
            // order of multimap not important here
            multimap<double, Genome*>::iterator it = compatibleGenomes.begin();
            for (; it!=compatibleGenomes.end(); ++it) {
                if (chosenC < it->first) break;
                // reomve this interval length
                chosenC -= it->first;
            }
            if (it!=compatibleGenomes.end()) { // equality should never happen at this point, but...
                spawn(it->second);
                return;
            }
        }
        // Otherwise, emit spores if this place would be fine, but lack spores
        else if (!speciesSporesPresent) {
            energy -= sporeEmissionEnergyCost;
            cloud->addSpore(getPosition().x(), getPosition().y(), getPosition().z(), genome, 1.0);
        }
        // too bad... continue search
    }

    // no, search for a good quiet place
    searchQuietPlace(elapsedTime, currentTime);

}

void ExplicitAIAgent::spawn( Genome * spore )
{
    double childMass = (maxMass - minMass) * genome->childMassRatio + minMass;
    if (childMass<minMass) childMass = minMass; // pathologic mutation case
    // return if we don't have enough mass
    if (mass - childMass < minMass) return;

    // give birth!
    ExplicitAIAgent* child = new ExplicitAIAgent(specie, predatorSpecies, preySpecies);
    // take from our own mass
    mass -= childMass;
    // to give birth
    child->mass = childMass;
    
    // give child some amount of our own energy, genome dependant
    child->energy = genome->childEnergyRatio * energy;
    // Pathologic mutation case. Child will die soon, this agent will never reproduce
    // successfully, and has lost the child mass. It won't survive long...
    if (child->energy<0.0) child->energy = 0.0;
    energy -= child->energy;

    // introduce updates from parent
    updateChild(child);

#define UPDATE_COLOR(c,g) \
    if (c>0.5f) c *= 0.6 + 0.4*sin(g*6.2831853*0.1); \
    else c = 0.4*cos(g*6.2831853*0.1);
    
    // child has default specie color values. Update with generation
    UPDATE_COLOR(child->genome->r, child->generation)
    UPDATE_COLOR(child->genome->g, child->generation)
    UPDATE_COLOR(child->genome->b, child->generation)

#define CROSS_MUTATE(gene,range) \
    child->genome->gene = ((Utility::Random() < 0.5) ? genome->gene : spore->gene) \
    + ((2.0*range) * Utility::Random()) - range; 
    
    // do crossover: each time choose one of our gene or the spore gene
    // do mutation: add random values with a sensible range
    CROSS_MUTATE(hungryRatioThreshold, 0.1);
    CROSS_MUTATE(hungryLastResortRatioThreshold, 0.01);
    CROSS_MUTATE(hungryGrassWeight, 0.1);
    CROSS_MUTATE(hungryPreyWeight, 0.1);
    CROSS_MUTATE(hungryPredatorWeight, 0.1);
    CROSS_MUTATE(hungryWeightedThreshold, 0.1);
    CROSS_MUTATE(grazeRate, 0.05);
    CROSS_MUTATE(grazeNextAmountThreshold, 0.1);
    CROSS_MUTATE(grazeForceCoef, 0.1);
    CROSS_MUTATE(huntMateForwardCoef,0.1)
    CROSS_MUTATE(huntMateEvasionCoef,0.1)
    CROSS_MUTATE(huntWanderCoef,0.1)
    CROSS_MUTATE(huntWanderRate,0.05)
    CROSS_MUTATE(huntAvoidCoef,0.5)
    CROSS_MUTATE(huntPreyCoef,0.1)
    CROSS_MUTATE(huntPredatorCoef,0.1)
    CROSS_MUTATE(huntMateCoef,0.1)
    CROSS_MUTATE(seekQuietPreyCutoff,0.1)
    CROSS_MUTATE(seekQuietOtherEvasionCoef,0.1)
    CROSS_MUTATE(seekQuietWanderCoef,0.1)
    CROSS_MUTATE(seekQuietWanderRate,0.05)
    CROSS_MUTATE(seekQuietAvoidCoef,0.5)
    CROSS_MUTATE(seekQuietPreyCoef,0.1)
    CROSS_MUTATE(seekQuietPredatorCoef,0.1)
    CROSS_MUTATE(seekQuietOtherCoef,0.1)
    CROSS_MUTATE(quietWeightForObstacle,0.1)
    CROSS_MUTATE(quietWeightForUnknown,0.1)
    CROSS_MUTATE(quietWeightForPrey,0.1)
    CROSS_MUTATE(quietWeightForPredator,0.1)
    CROSS_MUTATE(quietWeightForMate,0.1)
    CROSS_MUTATE(quietThreshold,0.5)
    CROSS_MUTATE(energyForFightRatio,0.1)
    CROSS_MUTATE(energyConvertionRatio,0.15)
    CROSS_MUTATE(childEnergyRatio,0.1)
    CROSS_MUTATE(childMassRatio,0.1)

    if (logGenomes) {
        vector<Genome*> *v;
        Genome* meanGenome, *varGenome;
        
        v = &genomesMean[specie];
        if (child->generation >= v->size()) v->resize(child->generation+1, 0);
        meanGenome = (*v)[child->generation];
        if (!meanGenome) {
            meanGenome = new Genome();
            (*v)[child->generation] = meanGenome;
        }

        v = &genomesVariance[specie];
        if (child->generation >= v->size()) v->resize(child->generation+1, 0);
        varGenome = (*v)[child->generation];
        if (!varGenome) {
            varGenome = new Genome();
            (*v)[child->generation] = varGenome;
        }
        
        meanGenome->specie += 1; /* use specie as counter for number of agents in this generation */
        varGenome->specie += 1; /* use specie as counter for number of agents in this generation */

#define UPDATE_GENOME_STATS(agent, gene) \
        meanGenome->gene += agent->genome->gene; \
        varGenome->gene += agent->genome->gene * agent->genome->gene;

        UPDATE_GENOME_STATS(child, hungryRatioThreshold);
        UPDATE_GENOME_STATS(child, hungryLastResortRatioThreshold);
        UPDATE_GENOME_STATS(child, hungryGrassWeight);
        UPDATE_GENOME_STATS(child, hungryPreyWeight);
        UPDATE_GENOME_STATS(child, hungryPredatorWeight);
        UPDATE_GENOME_STATS(child, hungryWeightedThreshold);
        UPDATE_GENOME_STATS(child, grazeRate);
        UPDATE_GENOME_STATS(child, grazeNextAmountThreshold);
        UPDATE_GENOME_STATS(child, grazeForceCoef);
        UPDATE_GENOME_STATS(child, huntMateForwardCoef)
        UPDATE_GENOME_STATS(child, huntMateEvasionCoef)
        UPDATE_GENOME_STATS(child, huntWanderCoef)
        UPDATE_GENOME_STATS(child, huntWanderRate)
        UPDATE_GENOME_STATS(child, huntAvoidCoef)
        UPDATE_GENOME_STATS(child, huntPreyCoef)
        UPDATE_GENOME_STATS(child, huntPredatorCoef)
        UPDATE_GENOME_STATS(child, huntMateCoef)
        UPDATE_GENOME_STATS(child, seekQuietPreyCutoff)
        UPDATE_GENOME_STATS(child, seekQuietOtherEvasionCoef)
        UPDATE_GENOME_STATS(child, seekQuietWanderCoef)
        UPDATE_GENOME_STATS(child, seekQuietWanderRate)
        UPDATE_GENOME_STATS(child, seekQuietAvoidCoef)
        UPDATE_GENOME_STATS(child, seekQuietPreyCoef)
        UPDATE_GENOME_STATS(child, seekQuietPredatorCoef)
        UPDATE_GENOME_STATS(child, seekQuietOtherCoef)
        UPDATE_GENOME_STATS(child, quietWeightForObstacle)
        UPDATE_GENOME_STATS(child, quietWeightForUnknown)
        UPDATE_GENOME_STATS(child, quietWeightForPrey)
        UPDATE_GENOME_STATS(child, quietWeightForPredator)
        UPDATE_GENOME_STATS(child, quietWeightForMate)
        UPDATE_GENOME_STATS(child, quietThreshold)
        UPDATE_GENOME_STATS(child, energyForFightRatio)
        UPDATE_GENOME_STATS(child, energyConvertionRatio)
        UPDATE_GENOME_STATS(child, childEnergyRatio)
        UPDATE_GENOME_STATS(child, childMassRatio)
        
    }
    
    // Cache Child own critical mass
    double childChildMass = (maxMass - minMass) * child->genome->childMassRatio + minMass;
    if (childChildMass<minMass) childChildMass = minMass; // pathologic mutation case
    child->spawnCriticalMass = childChildMass + minMass;
    
    
    // child is ready. Make sure it appears for other agents during this
    // simulator step => reproducible results
    // the visual block may be deferred to next graphics frame, though
    if (viewer!=0) {
        child->addAvatarToGroup(root, true);
    }

    child->addToLocalityDatabase();
    
    child->aiEvent = simu->postCyclicAt<Agent>(
        // add jitter so all ai computations not done at the same time
        simu->getTime() + agentsAITimeStep * Utility::Random(),
        agentsAITimeStep, child, &Agent::aiStep);
    
    child->integrationEvent = simu->postCyclicAt<Agent>(
        // add jitter so all integration step computations not done at the same time
        simu->getTime() + agentsIntegrationTimeStep * Utility::Random(),
        agentsIntegrationTimeStep, child, &Agent::integrationStep);
    
}


void ExplicitAIAgent::energyGainAction( double gain )
{
    // First, metabolize part of the energy, if needed.
    if (mass < maxMass) {
        double r = genome->energyConvertionRatio * getEnergyFootprintRatio();
        if (r>0.0) {
            double convertedMass = gain * r;
            if (mass+convertedMass > maxMass) convertedMass = maxMass - mass;
            mass += convertedMass;
            gain -= convertedMass / r;
        }
    }

    // Call parent for update
    CrogaiAgent::energyGainAction(gain);

    // Emit spores as we're very happy now!
    if (energy > sporeEmissionEnergyThreshold) {
        // Emit a spore only if the gain justifies it
        if (gain > sporeEmissionEnergyCost * Utility::Random()) {
            energy -= sporeEmissionEnergyCost;
            cloud->addSpore(getPosition().x(), getPosition().y(), getPosition().z(), genome, 1.0);
        }
    }

}

osg::Node *ExplicitAIAgent::buildAvatar()
{
    // Share the body of these agents across all members of the class, thanks to ref counting
    osg::ShapeDrawable *body = new osg::ShapeDrawable(crogaiAgentShape);

    // get color from genome
    body->setColor(osg::Vec4(genome->r, genome->g, genome->b, 1.0f));

    // OSG used geodes as generic containers. Bridge from osg::Shape to osg::Node
    osg::Geode* geode = new osg::Geode();
    geode->addDrawable(body);
    return geode;
}

bool ExplicitAIAgent::isPrey( CrogaiAgent * agent )
{
    for (vector<int>::iterator it = preySpecies.begin(); it!=preySpecies.end(); ++it) {
        if (agent->specie == *it) return true;
    }
    return false;
}

bool ExplicitAIAgent::isPredator( CrogaiAgent * agent )
{
    for (vector<int>::iterator it = predatorSpecies.begin(); it!=predatorSpecies.end(); ++it) {
        if (agent->specie == *it) return true;
    }
    return false;
}

double ExplicitAIAgent::getEnergyForFight( )
{
    return energy * genome->energyForFightRatio;
}

static int logNumberOfSpecies = 0;

void ExplicitAIAgent::createGenomesLogger( const std::string & id, int numSpecies )
{
    logNumberOfSpecies = numSpecies;
    vector<string> names;
    names.push_back("generations");
    for (int i=1; i<=numSpecies; ++i) {
        ostringstream os;

#define CREATE_LOG_ENTRY(gene) \
        os << #gene << "MeanSpecie" << i; \
        names.push_back(os.str()); \
        os.str(""); \
        os << #gene << "StandardDeviationSpecie" << i; \
        names.push_back(os.str()); \
        os.str(""); 

        CREATE_LOG_ENTRY(hungryRatioThreshold);
        CREATE_LOG_ENTRY(hungryLastResortRatioThreshold);
        CREATE_LOG_ENTRY(hungryGrassWeight);
        CREATE_LOG_ENTRY(hungryPreyWeight);
        CREATE_LOG_ENTRY(hungryPredatorWeight);
        CREATE_LOG_ENTRY(hungryWeightedThreshold);
        CREATE_LOG_ENTRY(grazeRate);
        CREATE_LOG_ENTRY(grazeNextAmountThreshold);
        CREATE_LOG_ENTRY(grazeForceCoef);
        CREATE_LOG_ENTRY(huntMateForwardCoef)
        CREATE_LOG_ENTRY(huntMateEvasionCoef)
        CREATE_LOG_ENTRY(huntWanderCoef)
        CREATE_LOG_ENTRY(huntWanderRate)
        CREATE_LOG_ENTRY(huntAvoidCoef)
        CREATE_LOG_ENTRY(huntPreyCoef)
        CREATE_LOG_ENTRY(huntPredatorCoef)
        CREATE_LOG_ENTRY(huntMateCoef)
        CREATE_LOG_ENTRY(seekQuietPreyCutoff)
        CREATE_LOG_ENTRY(seekQuietOtherEvasionCoef)
        CREATE_LOG_ENTRY(seekQuietWanderCoef)
        CREATE_LOG_ENTRY(seekQuietWanderRate)
        CREATE_LOG_ENTRY(seekQuietAvoidCoef)
        CREATE_LOG_ENTRY(seekQuietPreyCoef)
        CREATE_LOG_ENTRY(seekQuietPredatorCoef)
        CREATE_LOG_ENTRY(seekQuietOtherCoef)
        CREATE_LOG_ENTRY(quietWeightForObstacle)
        CREATE_LOG_ENTRY(quietWeightForUnknown)
        CREATE_LOG_ENTRY(quietWeightForPrey)
        CREATE_LOG_ENTRY(quietWeightForPredator)
        CREATE_LOG_ENTRY(quietWeightForMate)
        CREATE_LOG_ENTRY(quietThreshold)
        CREATE_LOG_ENTRY(energyForFightRatio)
        CREATE_LOG_ENTRY(energyConvertionRatio)
        CREATE_LOG_ENTRY(childEnergyRatio)
        CREATE_LOG_ENTRY(childMassRatio)

        
        // Add number: allows consistency checks with CrogaiAgent stats, plus
        // allows to recover sums from mean and std dev in archive
        os << "NumberOfGenomesSpecie" << i;
        names.push_back(os.str());
    }
    logGenomes = Utility::createLogger(id, names);


    // Add stats for the initial generation
    vector<Genome*> *v;
    Genome* meanGenome, *varGenome;
        
    for (int i=1; i<=numSpecies; ++i) {
        v = &genomesMean[i];
        if (v->size() < 1) v->resize(1, 0);
        meanGenome = (*v)[0];
        if (!meanGenome) {
            meanGenome = new Genome();
            (*v)[0] = meanGenome;
        }

        v = &genomesVariance[i];
        if (v->size() < 1) v->resize(1, 0);
        varGenome = (*v)[0];
        if (!varGenome) {
            varGenome = new Genome();
            (*v)[0] = varGenome;
        }

        for (list<Agent*>::iterator it = Agent::allAgents.begin(); it != Agent::allAgents.end(); ++it) {
            ExplicitAIAgent* agent = dynamic_cast<ExplicitAIAgent*>(*it);
            if (!agent) continue;

            meanGenome->specie += 1;
            varGenome->specie += 1;


            // re-use previous macro
            UPDATE_GENOME_STATS(agent, hungryRatioThreshold);
            UPDATE_GENOME_STATS(agent, hungryLastResortRatioThreshold);
            UPDATE_GENOME_STATS(agent, hungryGrassWeight);
            UPDATE_GENOME_STATS(agent, hungryPreyWeight);
            UPDATE_GENOME_STATS(agent, hungryPredatorWeight);
            UPDATE_GENOME_STATS(agent, hungryWeightedThreshold);
            UPDATE_GENOME_STATS(agent, grazeRate);
            UPDATE_GENOME_STATS(agent, grazeNextAmountThreshold);
            UPDATE_GENOME_STATS(agent, grazeForceCoef);
            UPDATE_GENOME_STATS(agent, huntMateForwardCoef)
            UPDATE_GENOME_STATS(agent, huntMateEvasionCoef)
            UPDATE_GENOME_STATS(agent, huntWanderCoef)
            UPDATE_GENOME_STATS(agent, huntWanderRate)
            UPDATE_GENOME_STATS(agent, huntAvoidCoef)
            UPDATE_GENOME_STATS(agent, huntPreyCoef)
            UPDATE_GENOME_STATS(agent, huntPredatorCoef)
            UPDATE_GENOME_STATS(agent, huntMateCoef)
            UPDATE_GENOME_STATS(agent, seekQuietPreyCutoff)
            UPDATE_GENOME_STATS(agent, seekQuietOtherEvasionCoef)
            UPDATE_GENOME_STATS(agent, seekQuietWanderCoef)
            UPDATE_GENOME_STATS(agent, seekQuietWanderRate)
            UPDATE_GENOME_STATS(agent, seekQuietAvoidCoef)
            UPDATE_GENOME_STATS(agent, seekQuietPreyCoef)
            UPDATE_GENOME_STATS(agent, seekQuietPredatorCoef)
            UPDATE_GENOME_STATS(agent, seekQuietOtherCoef)
            UPDATE_GENOME_STATS(agent, quietWeightForObstacle)
            UPDATE_GENOME_STATS(agent, quietWeightForUnknown)
            UPDATE_GENOME_STATS(agent, quietWeightForPrey)
            UPDATE_GENOME_STATS(agent, quietWeightForPredator)
            UPDATE_GENOME_STATS(agent, quietWeightForMate)
            UPDATE_GENOME_STATS(agent, quietThreshold)
            UPDATE_GENOME_STATS(agent, energyForFightRatio)
            UPDATE_GENOME_STATS(agent, energyConvertionRatio)
            UPDATE_GENOME_STATS(agent, childEnergyRatio)
            UPDATE_GENOME_STATS(agent, childMassRatio)
            
        }
    }

}

void ExplicitAIAgent::recordGenomesStatistics( )
{
    if (!logGenomes) return;

    // first step: get max gen
    int maxGeneration = 0;
    vector<Genome*> *v;

    for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {

        v = &genomesMean[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
        
        v = &genomesVariance[specie];
        if (maxGeneration <= v->size()) maxGeneration = v->size();
    }

    // second step: make each vector match max size
    for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {
        genomesMean[specie].resize(maxGeneration, 0);
        genomesVariance[specie].resize(maxGeneration, 0);
    }
    
    Genome* meanGenome, *varGenome;

    // now we can safely log variables
    for (int generation = 0; generation<maxGeneration; ++generation) {
        int varIdx = 0;
        (*logGenomes)[varIdx++] = generation;
        
        for (int specie = 1; specie <= logNumberOfSpecies; ++specie) {

            meanGenome = (genomesMean[specie])[generation];
            varGenome = (genomesVariance[specie])[generation];
                
            double n = (meanGenome!=0) ? meanGenome->specie : 0;
            // assert(n== ((varGenome!=0) ? varGenome->specie : 0));
            if (n==0.0) n=1.0;
            double ninv = 1.0 / n;
            
            double m,v;

#define RECORD_LOG_ENTRY(gene) \
            m = (meanGenome!=0) ? (meanGenome->gene * ninv) : 0.0; \
            (*logGenomes)[varIdx++] = m; \
            v = (varGenome!=0) ? (varGenome->gene * ninv) : 0.0; \
            (*logGenomes)[varIdx++] = sqrt(v - m * m);

            RECORD_LOG_ENTRY(hungryRatioThreshold);
            RECORD_LOG_ENTRY(hungryLastResortRatioThreshold);
            RECORD_LOG_ENTRY(hungryGrassWeight);
            RECORD_LOG_ENTRY(hungryPreyWeight);
            RECORD_LOG_ENTRY(hungryPredatorWeight);
            RECORD_LOG_ENTRY(hungryWeightedThreshold);
            RECORD_LOG_ENTRY(grazeRate);
            RECORD_LOG_ENTRY(grazeNextAmountThreshold);
            RECORD_LOG_ENTRY(grazeForceCoef);
            RECORD_LOG_ENTRY(huntMateForwardCoef)
            RECORD_LOG_ENTRY(huntMateEvasionCoef)
            RECORD_LOG_ENTRY(huntWanderCoef)
            RECORD_LOG_ENTRY(huntWanderRate)
            RECORD_LOG_ENTRY(huntAvoidCoef)
            RECORD_LOG_ENTRY(huntPreyCoef)
            RECORD_LOG_ENTRY(huntPredatorCoef)
            RECORD_LOG_ENTRY(huntMateCoef)
            RECORD_LOG_ENTRY(seekQuietPreyCutoff)
            RECORD_LOG_ENTRY(seekQuietOtherEvasionCoef)
            RECORD_LOG_ENTRY(seekQuietWanderCoef)
            RECORD_LOG_ENTRY(seekQuietWanderRate)
            RECORD_LOG_ENTRY(seekQuietAvoidCoef)
            RECORD_LOG_ENTRY(seekQuietPreyCoef)
            RECORD_LOG_ENTRY(seekQuietPredatorCoef)
            RECORD_LOG_ENTRY(seekQuietOtherCoef)
            RECORD_LOG_ENTRY(quietWeightForObstacle)
            RECORD_LOG_ENTRY(quietWeightForUnknown)
            RECORD_LOG_ENTRY(quietWeightForPrey)
            RECORD_LOG_ENTRY(quietWeightForPredator)
            RECORD_LOG_ENTRY(quietWeightForMate)
            RECORD_LOG_ENTRY(quietThreshold)
            RECORD_LOG_ENTRY(energyForFightRatio)
            RECORD_LOG_ENTRY(energyConvertionRatio)
            RECORD_LOG_ENTRY(childEnergyRatio)
            RECORD_LOG_ENTRY(childMassRatio)
            
            
            
            // record the number. See the log creater function 
            (*logGenomes)[varIdx++] = n;
        }
        // record an entry
        logGenomes->record();
    }
}

