/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <osg/Geometry>

#include "GrassyTerrain.h"
#include "Environment.h"
using namespace Environment;

GrassyTerrain::GrassyTerrain( double worldSize, double maxHeight, int level, double persistance, double power )
: Terrain(worldSize, maxHeight, level, persistance, power)
{
    // one grass value per data value
    grass = new double[dsize*dsize];

    grassEnd = grass + (dsize*dsize);
    
    setGrassGrowthRate(1.0);
    setMaximumGrassDensity(10.0);
    setGrassDensity(5.0);
    timeStep = 1.0; gcoef = 1.0; // will be used only when relevant anyway, initial values not used
    
    if (!batch) {
        osg::Geometry * geom = dynamic_cast<osg::Geometry *>(getDrawable(0));
        geom->setUseDisplayList(false);
        updateColors();
    }
    cleanVisualState = true;
}


GrassyTerrain::~GrassyTerrain()
{
    
}

double GrassyTerrain::getGrassDensity( const double x, const double y )
{
    double px = (x + halfWorldSize) * invWorldSize_dsize;
    double py = (y + halfWorldSize) * invWorldSize_dsize;

    if (px>dsize) px = dsize;
    if (px<0) px = 0;
    if (py>dsize) py = dsize;
    if (py<0) py = 0;

    // round up to nearest integer and mask the result for level-modulo arithmetic
    const int xidx = static_cast<int>(rint(px)) & dsizeMask;
    const int yidx = static_cast<int>(rint(py)) & dsizeMask;

    return grass[(yidx << level) + xidx];
}

double GrassyTerrain::getNormalizedGrassDensity( const double x, const double y )
{
    return getGrassDensity(x,y) * invMaximumGrassDensityPerGridPoint;
}

double GrassyTerrain::graze( const double x, const double y, const double amount )
{
    double px = (x + halfWorldSize) * invWorldSize_dsize;
    double py = (y + halfWorldSize) * invWorldSize_dsize;

    if (px>dsize) px = dsize;
    if (px<0) px = 0;
    if (py>dsize) py = dsize;
    if (py<0) py = 0;

    // round up to nearest integer and mask the result for level-modulo arithmetic
    const int xidx = static_cast<int>(rint(px)) & dsizeMask;
    const int yidx = static_cast<int>(rint(py)) & dsizeMask;

    double* g = &grass[(yidx << level) + xidx];
    if (*g >= amount) {
        *g -= amount;
        return amount;
    }
    double ret = *g;
    *g = 0.0;
    return ret;
}

void GrassyTerrain::setMaximumGrassDensity( const double density )
{
    maximumGrassDensity = density;
    maximumGrassDensityPerGridPoint = maximumGrassDensity * worldSize * worldSize / (dsize*dsize);
    invMaximumGrassDensityPerGridPoint = 1.0 / maximumGrassDensityPerGridPoint;
}

double GrassyTerrain::getMaximumGrassDensity()
{
    return maximumGrassDensity;
}

void GrassyTerrain::setGrassDensity( const double density )
{
    double grassDensityPerGridPoint = density * worldSize * worldSize / (dsize*dsize);
    
    for (double * g = grass; g<grassEnd; ++g) {
        *g = grassDensityPerGridPoint;
    }
}

void GrassyTerrain::setGrassGrowthRate( const double rate )
{
    grassGrowthRate = rate;
    grassGrowthRatePerGridPoint = grassGrowthRate * worldSize * worldSize / (dsize*dsize);
    gcoef = grassGrowthRatePerGridPoint * timeStep;
}

double GrassyTerrain::getGrassGrowthRate()
{
    return grassGrowthRate;
}

void GrassyTerrain::scheduleForGrowth( double integrationTimeStep )
{
    timeStep = integrationTimeStep;
    gcoef = grassGrowthRatePerGridPoint * integrationTimeStep;
    Environment::simu->postCyclic(timeStep, this, &GrassyTerrain::grow);
}

void GrassyTerrain::grow( )
{
    for (double * g = grass; g<grassEnd; ++g) {
        *g += gcoef;
        if (*g < maximumGrassDensityPerGridPoint) continue;
        *g = maximumGrassDensityPerGridPoint;
    }
    if ((Environment::viewer!=0) && cleanVisualState) {
        Environment::viewer->post(this, &GrassyTerrain::updateColors);
        cleanVisualState = false;
    }
}

static osg::Vec4 colorEmpty(232.0 /255.0, 229.0 /255.0, 238.0 /255.0, 1.0);
//static osg::Vec4 colorFull(146.0 /255.0, 162.0 /255.0, 100.0 /255.0, 1.0);
static osg::Vec4 colorDelta((146.0 - 232.0) /255.0, (162.0 - 229.0) /255.0, (100.0 - 238.0) /255.0, 1.0);

void GrassyTerrain::updateColors( )
{
    
    double * g = grass;
    osg::Vec4Array::iterator it = colorArray->begin();
    
    for (int i=0; i<dsize; ++i) {
        for (int j=0; j<dsize; ++j) {
            *it++ = colorEmpty + colorDelta * (*(g+j) * invMaximumGrassDensityPerGridPoint);
        }
        // cycle last point on line
        *it++ = colorEmpty + colorDelta * (*g * invMaximumGrassDensityPerGridPoint);
        g += dsize;
    }
    // cycle last line
    for (int j=0; j<dsize; ++j) {
        *it++ = colorEmpty + colorDelta * (*(grass+j) * invMaximumGrassDensityPerGridPoint);
    }
    // cycle last point on last line
    *it++ = colorEmpty + colorDelta * (*grass * invMaximumGrassDensityPerGridPoint);

    cleanVisualState = true;
}


