/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef GRASSYTERRAIN_H
#define GRASSYTERRAIN_H

#include "Terrain.h"

/** Adds grass to the terrain.

    Grass is modelled as an energy density per surface.
    It can grow with a given rate, and be eaten.

    @author Nicolas Brodu
*/
class GrassyTerrain : public Terrain
{
public:
    GrassyTerrain(double worldSize, double maxHeight, int level, double persistance=0.5, double power=1.0);

    ~GrassyTerrain();

    /** Gets the current grass density at the given location.
        The grass is considered a discrete grid, like for numerical integration.
        Unlike altitude (in parent Terrain), the grass density is not interpolated.
        Instead, the closest discrete grid cell is used.
        The reason for this is to avoid problems when eating the grass:
        Example: Suppose a request is made for a point C between A and B, with respective distance ratio 2/5 and 3/5.
                 Now suppose we interpolate the values at A=2 and B=7. Then C would have interpolated value of 4.
                 Now an agent request to eat 4 grass units at C. How do we split this request among the points A & B?
                 Suppose we remove some amount from both A and B. Then since neither A and B can be negative, the value
                 at C will be positive too.
                 Thus, after removing 4 units at a place where the agent saw 4 units, grass still remains... till both
                 A and B are consumed entirely, for a grand total of 9 units, thus miserably destroying the initial
                 intent of the interpolation.
        Using only the closest value corrects this problem, and is consistent with numerical integration schemes anyway.
    */
    virtual double getGrassDensity(const double x, const double y);

    /// Same as above, but normalized between 0 and 1
    virtual double getNormalizedGrassDensity(const double x, const double y);

    /// Tries to eat a certain amount of grass at the given location
    /// @returns the real amount eaten, if the desired amount is more than the current grass density
    virtual double graze(const double x, const double y, const double amount);

    /// Maximum grass density = maximum energy per terrain square unit.
    /// Note this is really per square unit, whatever the discretization of the grid:
    /// It uses the worldsize argument in the constructor, not the level one.
    /// Default is a completely arbitrary 10.0
    void setMaximumGrassDensity(const double density);
    double getMaximumGrassDensity();

    /// Give an initial grass density. Default is 5.0
    void setGrassDensity(const double density);

    /// The rate at which grass regenerates, in density per time unit.
    /// This is constant growth, not exponential, and clamped at MaximumGrassDensity.
    /// Default is a completely arbitrary 1.0
    void setGrassGrowthRate(const double rate);
    double getGrassGrowthRate();
    
    virtual void scheduleForGrowth(double integrationTimeStep);

protected:
    double *grass, *grassEnd;
    double maximumGrassDensity, maximumGrassDensityPerGridPoint, invMaximumGrassDensityPerGridPoint;
    double grassGrowthRate, grassGrowthRatePerGridPoint;

    /// Integration step: grass grows till
    virtual void grow();
    double timeStep, gcoef;
    bool cleanVisualState;

    /// update the colors, but only when necessary (no useless polling)
    virtual void updateColors();
};

#endif
