/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef LOCALITY_H
#define LOCALITY_H

#include <list>
#include <map>

#include <osg/Vec3d>

#include "lqXYCyclic.h"

/**
    Type for objects in the locality database
    
    Handles locality parameters and definitions
    
    Note: the routines in this class
    are thread-safe

    @author Nicolas Brodu
*/

class Localized {
public:    

    Localized();

//////////////////////////////////////////////////////////////////////////
/// Attributes
//////////////////////////////////////////////////////////////////////////

    /** The only attribute required for a localized object is its position.
        Too bad for dependencies, but I use osg::Vec3d as a commodity!
        If you plan to re-use this class in another context, let's talk about it
        and you're welcome to cut the dependencies (disclamer: Environment.h is drawn
        by this class implementation)
    */
    virtual const osg::Vec3d& getPosition() const;
    virtual void setPosition(const osg::Vec3d& pos);
    virtual void setPosition(const double& x, const double& y, const double& z);


//////////////////////////////////////////////////////////////////////////
/// Locality Policy functions
//////////////////////////////////////////////////////////////////////////

    /**
        Sets an algorithm for locality queries
        - Brute: brute-checking over all localized objects : O(n) for each object => O(n^2)
        - Bricks: use super-brick DB facility from lq.c: O(k) for each object => ~O(n)
        - NPropagation: Maintain a list of neighbors in each object and propagate
          the list to other objects.
          Main pb: isolated individuals are lost, two groups may cross without
                   seeing each other.
          Advantage: faster than Bricks
        - Mixed: Do a locality check in DB to complement list of neighbors.
          This allows to find back isolated individuals and to make groups see each
          other after few iterations.
          Consistency depends on the proportion of objects in the database.

        Note: Bricks and mixed are available only if createLocalityDatabase was called
              to set up the database! Brute and neighbors are respectively used instead
              so long as there is no DB.

        Note2: Default is Brute force, because it requires no setup from the user.
        Note3: Unless overloaded, mode "User" behaves like NPropagation
     */
    enum LocalityMode { Brute, Bricks, NPropagation, Mixed, User};

    /** The locality mode should be a per-class setting. The Localized class shares the locality mode across
        all its objects, but 2 different subclasses may use different policies. Hence the possibility to overload
        these 2 functions. Subclasses should call getLocalityMode(), to always use the correct
        policy. setDefaultLocalityMode is used by the main program.
     */
    virtual void setLocalityMode(LocalityMode mode);
    virtual LocalityMode getLocalityMode();
    /// static versions to set the values for this class, and default for non-overloaded versions
    static void setDefaultLocalityMode(LocalityMode mode);
    static LocalityMode getDefaultLocalityMode();


    
//////////////////////////////////////////////////////////////////////////
/// Locality query function, whatever the policy
//////////////////////////////////////////////////////////////////////////

    /** Find all neighbors of this object within a sphere of the given radius.

        Agents are returned classified by distance order, in a multimap
        whose pair values are: square of the distance and agent pointer.
        See neighbors field

        Note: This function should not be protected with neighborsAccessLock();
        Indded, it already uses it internally. Besides which, we need to keep
        allObjectAccess mutex around the neighbors, and not the opposite, so this
        is the only choice. Note that it doesn't matter if the updated neighbor
        list is modified before you re-acquire the mutex.
    */
    virtual void findNeighbors(double radius);
    
    /** Keep neighbor list whatever the locality mode for quick access.
        Possibly, the current neighbor list may be quickly used, and updated only
        when necessary (findNeighbors may be quite costly!).
        May also be useful for fast collision detection...
        
        Agents are returned classified by distance order, in a multimap
        whose pair values are: square of the distance and agent pointer.
        See findNeighbors.
        Of course, when using this cached list in an integration step for fast
        collision detection check, then the distance may not be correct.
    */
    std::multimap<const double, Localized*>* neighbors;

    
///////////////////////////////////////////////////////////////
/// Bricks mode: use a Locality Database
///////////////////////////////////////////////////////////////
    
    /** Add this object to the locality database.
        Note: This operation is thread-safe
    */
    virtual void addToLocalityDatabase();
    
    /** Removes this object from the locality database.
        Note: This operation is thread-safe
    */
    virtual void removeFromLocalityDatabase();


    /** See comments in lq.h.
    
        Allocate and initialize an locality database.
        The application needs to call this before using the LQ facility.
        The nine parameters define the properties of the "super-brick":
            (1) origin: coordinates of one corner of the super-brick, its
                minimum x, y and z extent.
            (2) size: the width, height and depth of the super-brick.
            (3) the number of subdivisions (sub-bricks) along each axis.
        This routine also allocates the bin array, and initialize its
        contents.

        @param cyclicXY if true (the default), the X/Y coordinate will
                        be wraped-up around the superbrick, but Z is still infinite.
                        This have the side effect of setting the world environment topology
        
        Note: This operation is thread-safe
    */
    static void createLocalityDatabase(double originx, double originy, double originz,
            double sizex,   double sizey,   double sizez,
            int   divx,    int   divy,    int   divz, bool cyclicXY=true);

    /** Takes care of invalidating all Localized object proxies.
        Note: This operation is thread-safe
    */
    static void deleteLocalityDatabase();
    
    /** See comments in lq.h. Is null by default.
        @see createLocalityDatabase
    */
    static lqDB* localityDatabase;

    /** Locks the locality Database with a mutex to prevent concurrent access.
        Never do a allObjectsAccessLock() inside a database-protected section, dead-lock may occur.
        The other hand is OK, you can do databaseAccessLock() inside a allObjectsAccessLock() section.
    */
    static void databaseAccessLock();
    
    /** Release the lock on the locality Database.
        Never do a allObjectsAccessLock() inside a database-protected section, dead-lock may occur.
        The other hand is OK, you can do databaseAccessLock() inside a allObjectsAccessLock() section.
    */
    static void databaseAccessUnlock();

    /// see lq.h. This is an handler for this object. Public by implementation side effect
    lqClientProxy* lqProxy;



////////////////////////////////////////
/// Internal use & administrative stuff
////////////////////////////////////////

protected:
    osg::Vec3d position;
    static LocalityMode localityMode;

    /// Protected destructor: see destroy
    virtual ~Localized();

public:

    /** Destroy this object for all purposes, make it diseappear from agent lists and locality database, etc.
        The real C++ deletion may be delayed, for example
        The default version just deletes this object now
    */
    virtual void destroy();

    /** A list containing all objects of this class
        Especially useful when not using the database mode.
    */
    static std::list<Localized*> allObjects;

};

#endif
