/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <iostream>

#include "SporeCloud.h"
#include "Environment.h"

using namespace Environment;
using namespace std;

SporeCloud::SporeVolume* SporeCloud::sporeVolumeFactory(void* user) {
    return new SporeVolume;
}

SporeCloud::SporeCloud( double originx, double originy, double originz, double sizex, double sizey, double sizez, int divx, int divy, int divz, double baseDiffusionCoef, double lowestConcentration, SporeVolumeFactory factory, void* user)
{
    this->divx = divx;
    this->divy = divy;
    this->divz = divz;
    this->originx = originx;
    this->originy = originy;
    this->originz = originz;
    this->sizex = sizex;
    this->sizey = sizey;
    this->sizez = sizez;
    this->lowestConcentration = lowestConcentration;

    // Create spore volumes
    vsize = divx * divy * divz;
    volume = new SporeVolume*[vsize];
    volumeEnd = volume + vsize;
    for (SporeVolume** it = volume; it != volumeEnd; ++it) {
        *it = factory(user);
    }
    // init traversal pointers
    SporeVolume* prev = 0;
    for (int z=0; z<divz; ++z) for (int y=0; y<divy; ++y) for (int x=0; x<divx; ++x) {
        SporeVolume* v = cyclicVolume(x,y,z);
        if (v==0) continue; // not allocated by factory
        if (prev!=0) prev->next = v;
        else firstVolume = v;
        prev = v;
        v->xn = cyclicVolume(x-1,y,z);
        v->xp = cyclicVolume(x+1,y,z);
        v->yn = cyclicVolume(x,y-1,z);
        v->yp = cyclicVolume(x,y+1,z);
        if (z==0) v->zn = 0;
        else v->zn = cyclicVolume(x,y,z-1);
        if (z==divz-1) v->zp = 0;
        else v->zp = cyclicVolume(x,y,z+1);
        v->concentrations = 0;
        v->processFlag = false;
    }
    if (prev!=0) prev->next = 0;
    else firstVolume = 0;

    firstProcess=0;

    diffusionCoefficient = baseDiffusionCoef;
    
    // Use flags for first & second time steps: 0 and then a negative value
    timeStep = 0.0;
    // pre-compute x,y,z diffusion coefficients with grid size taken into account.
    // We'll update these with time step once it's known
    diffuseXN = sizex / divx;
    diffuseXN = baseDiffusionCoef / (diffuseXN * diffuseXN);
    diffuseXP = diffuseXN;
    diffuseYN = sizey / divy;
    diffuseYN = baseDiffusionCoef / (diffuseYN * diffuseYN);
    diffuseYP = diffuseYN;
    diffuseZN = sizez / divz;
    diffuseZN = baseDiffusionCoef / (diffuseZN * diffuseZN);
    diffuseZP = diffuseZN;
}


SporeCloud::~SporeCloud()
{
    if (volume==0) return;
    for (SporeVolume** it = volume; it != volumeEnd; ++it) {
        if (*it !=0) delete *it;
        *it=0;
    }
    
    delete [] volume;
    volumeEnd = volume = 0;
}

SporeCloud::Genome::~ Genome( )
{
    // nothing to do
}

SporeCloud::SporeVolume::~ SporeVolume( )
{
    // Unref genomes as list is destroyed
    ConcentrationList* next = concentrations;
    while (next!=0) {
        next = concentrations->next;
        concentrations->genome->unref(); 
        delete concentrations;
    }
    concentrations = 0;
}

// Diffuse a neighbor volume into this volume
// diffusion to this cell is cumulative
void SporeCloud::diffuseVolume(SporeVolume* thisCell, SporeVolume* neighbor, double diffuseCoef) {

    // Naive double loop is O(n logn) with maps, plus we still have
    // to look for unmatched genomes in this cell at the end.
    // O(n) algo possible with sorted lists & a bit more work.
    // And this is not useless, since this routine is a potential bottleneck
    // when the cloud is full, according to profiler.

    ConcentrationList* tit = thisCell->concentrations;
    ConcentrationList* nit = neighbor->concentrations;
    ConcentrationList* prevtit = 0;

    // process so long as no list have ended
    while ((tit!=0) && (nit!=0)) {

    //does not gain much, mov to register needed for mul :while ((reinterpret_cast<long>(tit) * reinterpret_cast<long>(nit)) !=0) {

    // Help compiler this way... seem not to help either
/*    while (true) {
        if (!tit) break;
        if (!nit) break;
*/

        // In fact, the real bottleneck is the pointer de-reference
        // 28934 15.0145 :   6d7f1:   mov    %eax,0xffffffdc(%ebp)
        // 45693 23.7111 :   6d7f9:   mov    %eax,0xffffffd8(%ebp)
        // So... more than 40% total time spent on a comparison...
        // ...it seems that indeed main memory access is slooooowwwww...
        // Try to improve things by storing to registers (but compiler may ignore) & re-order tests
        // use unconditional "continue" jumps instead of else statements (help branching mis-predection)

        register const void* titgenome = tit->genome;
        register const void* nitgenome = nit->genome;

        // neighbor genome pointer is lower & different => this cell won't have it as it is sorted
        // we thus must introduce a new genome in this cell, just before the current one
        if (titgenome > nitgenome) {
            ConcentrationList* newCell = new ConcentrationList;
            newCell->genome = nit->genome->ref(); // ref neighbor genome for us
            newCell->concentration = 0.0;         // initial concentration 0.0
            newCell->nextConcentration = diffuseCoef * nit->concentration;   // 0 += xxx * (... - 0)
            newCell->prev = tit->prev;
            tit->prev = newCell;
            newCell->next = tit;
            // next for only neighbor till it catches this cell
            nit = nit->next;
            
            // insert in list: first element is special case...
            if (newCell->prev==0) {
                thisCell->concentrations = newCell;
                continue;
            }
            // otherwise, thanks double-linked list
            newCell->prev->next = newCell;
            continue;
        }
        
        // this genome pointer is lower & different => neighbor won't have it as it is sorted
        if (titgenome < nitgenome) {
            // nextConcentration += diffuseCoef * (neighbor->concentration - concentration);
            tit->nextConcentration -= diffuseCoef * tit->concentration; // += xxx * (0 - ...)
            // next for only this map till it catches on neighbor
            prevtit = tit; tit = tit->next;
            continue;
        }

        // both genomes are equal => process it
        // nextConcentration += diffuseCoef * (neighbor->concentration - concentration);
        tit->nextConcentration += diffuseCoef * (nit->concentration - tit->concentration);
        // update last element processed in this list
        prevtit = tit;
        // next for both lists
        tit = tit->next;
        nit = nit->next;
    }

    // This cell list is the one which ended => insert remaining neighbors, if any
    if (tit==0) {
        while (nit!=0) {
            ConcentrationList* newCell = new ConcentrationList;
            newCell->genome = nit->genome->ref(); // ref neighbor genome for us
            newCell->concentration = 0.0;         // initial concentration 0.0
            newCell->nextConcentration = diffuseCoef * nit->concentration;   // 0 += xxx * (... - 0)
            newCell->prev = prevtit;
            newCell->next = 0;
            // insert in list: first element is special case...
            if (prevtit==0) thisCell->concentrations = newCell;
            // otherwise, thanks double-linked list
            else prevtit->next = newCell;
            // next for only neighbor till it catches this cell
            nit = nit->next;
            prevtit = newCell; // update end of this list
        }
    }

    // else remaining elements in this list are diffused
    else {
        while (tit!=0) {
            // nextConcentration += diffuseCoef * (neighbor->concentration - concentration);
            tit->nextConcentration -= diffuseCoef * tit->concentration; // += xxx * (0 - ...)
            // till the end
            tit = tit->next;
        }
    }
}

void SporeCloud::intermediateStepDecay() {
    
    ConcentrationList* cit;
    
    // Spore decay, + initialize double buffer
    // for all volumes
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        // For all genomes
        for (cit = thisCell->concentrations; cit!=0; cit = cit->next) {
            // Euler integrate precision will be enough for our use
            // concentration += genome->decayConstant * timeStep * concentration;
            cit->concentration += cit->genome->decayConstant * timeStep * cit->concentration;
            // init next value
            cit->nextConcentration = cit->concentration;
        }
    }
}

void SporeCloud::intermediateStepXn() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->xn!=0) diffuseVolume(thisCell, thisCell->xn, diffuseXN);
    }
}

void SporeCloud::intermediateStepXp() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->xp!=0) diffuseVolume(thisCell, thisCell->xp, diffuseXP);
    }
}

void SporeCloud::intermediateStepYn() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->yn!=0) diffuseVolume(thisCell, thisCell->yn, diffuseYN);
    }
}

void SporeCloud::intermediateStepYp() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->yp!=0) diffuseVolume(thisCell, thisCell->yp, diffuseYP);
    }
}

void SporeCloud::intermediateStepZn() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->zn!=0) diffuseVolume(thisCell, thisCell->zn, diffuseZN);
    }
}

void SporeCloud::intermediateStepZp() {
    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        if (thisCell->zp!=0) diffuseVolume(thisCell, thisCell->zp, diffuseZP);
    }
}

bool SporeCloud::schedule( double integrationTimeStep )
{
    if (integrationTimeStep<0) return false;
    timeStep = integrationTimeStep; // needed for spore decay

    double stableCondition = 0.5*timeStep*(diffuseXN + diffuseYN + diffuseZN + diffuseXP + diffuseYP + diffuseZP);
    if (stableCondition>=0.5) {
        cerr << "Error! Cloud spore diffusion process is out of stability range: " << stableCondition << ">=0.5" << endl;
        return false;
    }

    if (stableCondition>=0.45) cout << "Warning! Cloud spore diffusion process is close to instability: " << stableCondition << " close to 0.5" << endl;
    else cout << "stableCondition=" << stableCondition << endl;

    // update the diffusion coefs with time step
    diffuseXN *= timeStep;
    diffuseYN *= timeStep;
    diffuseZN *= timeStep;
    diffuseXP *= timeStep;
    diffuseYP *= timeStep;
    diffuseZP *= timeStep;

    double now = simu->getTime();
    
    /// Now schedule intermediate steps to distribute computations in time
    /// In fact: final step + post process + next decay still no match for an intermediate step according
    /// to oprofile => schedule all these in only one delta compared to intermediate steps.
    double delta = timeStep / 7.0;
    simu->postCyclicAt(now + 0.666 * delta, timeStep, this, &SporeCloud::intermediateStepDecay);
    simu->postCyclicAt(now + 1.333 * delta, timeStep, this, &SporeCloud::intermediateStepXn);
    simu->postCyclicAt(now + 2.333 * delta, timeStep, this, &SporeCloud::intermediateStepXp);
    simu->postCyclicAt(now + 3.333 * delta, timeStep, this, &SporeCloud::intermediateStepYn);
    simu->postCyclicAt(now + 4.333 * delta, timeStep, this, &SporeCloud::intermediateStepYp);
    simu->postCyclicAt(now + 5.333 * delta, timeStep, this, &SporeCloud::intermediateStepZn);
    simu->postCyclicAt(now + 6.333 * delta, timeStep, this, &SporeCloud::intermediateStepZp);
    // step() at now+timeStep will complete the cycle!
    simu->postCyclicAt(now + timeStep, timeStep, this, &SporeCloud::step);
    simu->postCyclicAt(now + 7.333 * delta, timeStep, this, &SporeCloud::intermediateStepPostProcess);

    /// Now computations are distributed in time, let the fun begins!
    return true;
}

void SporeCloud::step( )
{
    SporeVolume** it;
    SporeVolume *thisCell;
    ConcentrationList* cit, *tmpcit;

/// Original all-in-one code below. Now split into several parts
/// to distribute the computations in time thanks to the simulator
/// Step does the final update

/// TODO: restore all-in-one mode for batch mode, where peaks don't matter
///       => could this improve perfs by de-allocating lists only once per update?
/* 
    // Spore decay, + initialize double buffer
    // for all volumes
    for (it = volume; it != volumeEnd; ++it) {
        thisCell = *it;
        if (thisCell == 0) continue; // skip cells removed by subclasses, or not allocated by factory
        // For all genomes
        for (cit = thisCell->concentrations; cit!=0; cit = cit->next) {
            // Euler integrate precision will be enough for our use
            // concentration += genome->decayConstant * timeStep * concentration;
            cit->concentration += cit->genome->decayConstant * timeStep * cit->concentration;
            // init next value
            cit->nextConcentration = cit->concentration;
        }
    }

    // Compute finite difference
    // V_n+1_xyz = V_n_xyz + Sum_i=xyz C * (V_n_i-1 -2*V_n_i + V_n_i+1)  with i=x,y,z and the other 2 constant
    for (it = volume; it != volumeEnd; ++it) {
        thisCell = *it;
        if (thisCell == 0) continue; // skip cells removed by subclasses, or not allocated by factory
        if (thisCell->xn!=0) diffuseVolume(thisCell, thisCell->xn, diffuseXN);
        if (thisCell->xp!=0) diffuseVolume(thisCell, thisCell->xp, diffuseXP);
        if (thisCell->yn!=0) diffuseVolume(thisCell, thisCell->yn, diffuseYN);
        if (thisCell->yp!=0) diffuseVolume(thisCell, thisCell->yp, diffuseYP);
        if (thisCell->zn!=0) diffuseVolume(thisCell, thisCell->zn, diffuseZN);
        if (thisCell->zp!=0) diffuseVolume(thisCell, thisCell->zp, diffuseZP);
        // TODO: include gravity coef: like this?
        //          nextConcentration += gcoef * ZP->concentration
        //          if (Z>0) nextConcentration -= gcoef * concentration
    }
*/


    double concentration;
    
    // update concentration, and remove dead spores
    // remove dead spores now, after diffusion to avoid useless removal/insertion
    for (SporeVolume *thisCell = firstProcess; thisCell != 0; thisCell = thisCell->nextProcess) {
        // For all genomes
        for (cit = thisCell->concentrations; cit!=0; ) {
            // use next concentration
            concentration = cit->nextConcentration;
            // erase element if concentration is too low
            if (concentration <= lowestConcentration) {
                cit->genome->unref(); // discard this genome
                tmpcit = cit->next;
            
                // take care of last element
                if (tmpcit!=0) tmpcit->prev = cit->prev;
                
                // take care of first element
                if (cit->prev==0) thisCell->concentrations = tmpcit;
                else cit->prev->next = tmpcit;
                
                delete cit;
                cit = tmpcit;
            }
            else {
                // or update the concentration
                cit->concentration = concentration;
                //cout << "concentration=" << concentration << endl;
                cit = cit->next;
            }
        }
        // all process flags are false before this loop.
        // Update the process flags for next loop: this cell and its neighbors
        // as neighbors may be duplicated, it's not easy to re-build the next
        // process list directly => use of flag ensures O(n) algo.
        if (thisCell->concentrations!=0) {
            thisCell->processFlag = true;
            if (thisCell->xn!=0) thisCell->xn->processFlag = true;
            if (thisCell->xp!=0) thisCell->xp->processFlag = true;
            if (thisCell->yn!=0) thisCell->yn->processFlag = true;
            if (thisCell->yp!=0) thisCell->yp->processFlag = true;
            if (thisCell->zn!=0) thisCell->zn->processFlag = true;
            if (thisCell->zp!=0) thisCell->zp->processFlag = true;
        }
    }

}

void SporeCloud::intermediateStepPostProcess( )
{
    // update process list. Unfortunately, run through all cells. Good new is
    // the flag check is fast now. Reset the flag for next sequence
    SporeVolume* prev = 0;
    for (SporeVolume *thisCell = firstVolume; thisCell != 0; thisCell = thisCell->next) {
        if (thisCell->processFlag) {
            // re-link cells for which it's not useless to diffuse!
            if (prev!=0) prev->nextProcess = thisCell;
            else firstProcess = thisCell;
            prev = thisCell;
            // reset flag for next time
            thisCell->processFlag = false;
        }
        
    }
    if (prev!=0) prev->nextProcess = 0;
    else firstProcess = 0;
}


void SporeCloud::addSpore( double x, double y, double z, Genome * genome, double concentration)
{
    int ix, iy, iz;

    // do not cycle but saturate in z
    if (z < originz) z = originz;
    if (z >= originz + sizez) z = originz + sizez - 1e-7; // use epsilon to make sure it falls in

    x -= originx;
    x = fmod(x, sizex);
    if (x<0) x+= sizex;
    ix = (int)(x * divx / sizex);
    
    y -= originy;
    y = fmod(y, sizey);
    if (y<0) y+= sizey;
    iy = (int)(y * divy / sizey);

    iz = (int) ((z - originz) * divz / sizez);

    SporeVolume* thisCell = volume[iz * (divx * divy) + iy * divx + ix];
    if (thisCell == 0) return; // Cells removed by subclass, should be an error as no agent is supposed to be here!

    
    bool found = false;
    // search for an existing entry for this genome
    ConcentrationList* it;
    ConcentrationList* previt = 0;
    for (it = thisCell->concentrations; it != 0; it = it->next) {
        if (it->genome == genome) {
            found = true;
            break;
        }
        else if (it->genome > genome) {
            break;
        }
        previt = it;
    }

    // found!!! update its concentration
    if (found) {
        it->concentration += concentration;
    
    // so, add a new genome, at the right place
    } else {
    
        ConcentrationList* newCell = new ConcentrationList;
        newCell->genome = genome->ref(); // ref genome for us
        newCell->concentration = concentration;    // initial concentration
        newCell->nextConcentration = concentration;   // should not matter
        newCell->prev = previt; // whether null or not
        newCell->next = it;     // whether null or not
        // restore double chain, take care of end of list
        if (it!=0) it->prev = newCell;
        // insert in list: first element is special case...
        if (previt==0) thisCell->concentrations = newCell;
        // otherwise, thanks double-linked list
        else previt->next = newCell;
    }

    // schedule for processing in next cloud integration
    thisCell->processFlag = true;
}

SporeCloud::ConcentrationList * SporeCloud::getConcentrationList( double x, double y, double z )
{
    int ix, iy, iz;

    // do not cycle but saturate in z
    if (z < originz) z = originz;
    if (z >= originz + sizez) z = originz + sizez - 1e-7; // use epsilon to make sure it falls in

    x -= originx;
    x = fmod(x, sizex);
    if (x<0) x+= sizex;
    ix = (int)(x * divx / sizex);
    
    y -= originy;
    y = fmod(y, sizey);
    if (y<0) y+= sizey;
    iy = (int)(y * divy / sizey);

    iz = (int) ((z - originz) * divz / sizez);

    SporeVolume* thisCell = volume[iz * (divx * divy) + iy * divx + ix];
    if (thisCell == 0) return 0; // Cells removed by subclass, should be an error as no agent is supposed to be here!

    return thisCell->concentrations;
    
}


/*
void SporeCloud::lock(SporeVolume* volume) const
{
    // To be overloaded
}

void SporeCloud::unlock(SporeVolume* volume) const
{
    // To be overloaded
}
*/
