/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef TERRAINAGENT_H
#define TERRAINAGENT_H

#include "Agent.h"
#include "TerrainInterface.h"

/** Adds terrain detection to the Agent class.

Default attitude correction policy is now keep up, with up defined
by the terrain normal at the agent location. This allows the agents to
keep their horizontal right on the floor, whatever the floor slope.

Note: The static class terrain field should be set up externally.
      By default, the agents consider Z=0 as the floor, and Z up the normal.

@author Nicolas Brodu

*/
class TerrainAgent : public Agent {
public:

//////////////////
/// Public API 
//////////////////

    TerrainAgent();

    static TerrainInterface* terrain;
    
    inline double getAltitude() {return altitude;}
    inline const osg::Vec3d& getFloorNormal() {return floorNormal;}

protected:
    virtual ~TerrainAgent();


///////////////////////////
/// API for subclasses
///////////////////////////

    /** Perform integration update:
        The position is constrained to be above the terrain, and attitude is updated accordingly.
        Two new internal variables are set for subclasse, which can be checked together with other state
        variables (position, attitude, etc...)
        - altitude: This variable is set to the position Z delta before calling the stubs
            If there was a terrain collision, it is strictly negative. Otherwise, it represents the
            altitude of this agent above the floor. At exact floor level, there is nothing done by
            default (consider no collision). The newPosition.z() is set to the floor level in case
            of collision.
        - floorNormal: the floor normal at the agent location.
        
        Subclasses should call this function at the beginning of their AI update routine
    */
    virtual void updateIntegration(const double elapsedTime, const double currentTime);

    /** Perform AI update:
        The position is constrained to be above the terrain, and attitude is updated accordingly.
        Two new internal variables are set for subclasse, which can be checked together with other state
        variables (position, attitude, etc...)
        - altitude: This variable is set to the position Z delta before calling the stubs
            If there was a terrain collision, it is strictly negative. Otherwise, it represents the
            altitude of this agent above the floor. At exact floor level, there is nothing done by
            default (consider no collision). The position field is not updated, this is done during 
            integration update.
        - floorNormal: the floor normal at the agent location.

        Subclasses should call this function at the beginning of their AI update routine
    */
    virtual void updateAI(const double elapsedTime, const double currentTime);

    double altitude;
    osg::Vec3d floorNormal;
};

#endif
