/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osg/ShapeDrawable>
#include <osgProducer/Viewer>
#include <osgGA/GUIEventHandler>

#include <iostream>
#include <sstream>

//#include "OpenThreadSimulator.h"
#include <time.h>

#include "Environment.h"
#include "Obstacle.h"
#include "TerrainAgent.h"
#include "Terrain.h"
#include "Utility.h"

using namespace std;
using namespace Environment;
using namespace Utility;

Terrain* theFloor = 0;


struct CollisionTerrainAgent : public TerrainAgent {

    static const double maxSearchRadius;
    static const double minSearchRadius;

    double searchRadius, minThreshold, maxThreshold;
    static map<int,long> collisionNumbers;
    static map<int,long> collisionAvoidedNumbers;
    bool collisionRisk, collisionHappened;
    double collisionDistance;
    Localized* collisionTarget;
    
    CollisionTerrainAgent() {
        searchRadius = (maxSearchRadius + minSearchRadius)* 0.5; // half to begin with
        minThreshold = 2;
        maxThreshold = 5;
        collisionRisk = collisionHappened = false;
        collisionDistance = 0;
        collisionTarget = 0;
    }

    // Experimental adaptative search radius
    // TODO: make this optional
    void updateSearchRadius(int n) {

        if (n>=maxThreshold) {
            if (searchRadius>minSearchRadius) {
                searchRadius--; // reduce radius
            }
            // TODO: find a rule to update min/max
        }

        if (n<=minThreshold) {
            if (searchRadius<maxSearchRadius) {
                searchRadius++; // increase radius
            }
            // TODO: find a rule to update min/max
        }
    }
    
    void updateAI( const double elapsedTime, const double currentTime) {
        TerrainAgent::updateAI(elapsedTime,currentTime);

        if (collisionRisk && (!collisionHappened)) {
            int dist = (int)(collisionDistance / radius);
            if (dist>6) dist=6;
            if (dist<1) dist=1;
            ++collisionAvoidedNumbers[dist];
            ostringstream os;
            os << "Target range:Avoided  ";
            for (map<int,long>::iterator it = collisionAvoidedNumbers.begin(); it!=collisionAvoidedNumbers.end(); ++it) {
                if (it==collisionAvoidedNumbers.begin()) os << "<=";
                else if (it->first==6) os << ", >=";
                else os << ", ";
                os << it->first << ":" << it->second;
            }
            Utility::setStatus(os.str(), Utility::TOP_LEFT);
        }
        collisionHappened = false;
        
        // avoidance force. Avoid only the closest object
        // Since neighbors are returned sorted, this is easy
        osg::Vec3d avoidForce(0,0,0);

        // Forces for Boid-like behavior
        osg::Vec3d localCenter(0,0,0);
        osg::Vec3d forwardForce(0,0,0);

        // number of visible neighbor agents
        int n = 0;
        
        findNeighbors(searchRadius);
        
        updateSearchRadius(neighbors->size());
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

                if (wrappedDiffVector(it->second->getPosition(), getPosition()) * forward < 0)
                    continue; // see only what's in front of us

                // collision risk? no need to look for further AI
                avoidForce = steerToAvoid(it->second);
                if (avoidForce.length2()!=0) {
                    collisionDistance = it->first;
                    collisionTarget = it->second;
                    break;
                }
                
                CollisionTerrainAgent* neighbor = dynamic_cast<CollisionTerrainAgent*>(it->second);

                // if not an agent, nothing else to do
                if (!neighbor) continue;
                
                // One of our mates, no risk of collision. Boidizing...
                steeringForce += steerForFlee(neighbor->getPosition()) / (it->first+0.001);
                forwardForce += neighbor->forward;
                localCenter += neighbor->getPosition();
                ++n;
        }

        // Collision risk? Priority over anything else
        if (avoidForce.length2()!=0) {

            // Don't try to avoid by digging a hole, it doesn't work!
            // When not flying, project avoidance direction on the horizontal plane
            if (altitude<radius) {
                steeringForce = avoidForce - floorNormal * (avoidForce*floorNormal);
                steeringForce *= sqrt(getMaxForceSquared() / steeringForce.length2());
            }
            else steeringForce = avoidForce;

            collisionRisk = true;
            return;
        }

        collisionRisk = false;

        // Don't move to the origin when no other agent is present
        // In this case, move randomly in a consistent way
        if (n==0) steeringForce += steerForWander(0.4);
        
        // Otherwise, agent behavior when in company of friends
        // - steer to avoid neighbors
        // - steer toward the local center
        // - and also to keep the local group direction
        else if (n>0) {
            steeringForce /= (double)n; // average of avoidance vectors
            steeringForce += steerForSeek(localCenter/(double)n) * 2; // local center
            steeringForce += forwardForce /(double)n  -  forward; // local group direction
        }
        
        // If flying, try to avoid floor, up to max height were we're "safe"
        if (altitude>radius) {
            if ( altitude < radius*5.0) 
                steeringForce += floorNormal * getMaxForce();
        }
        
        // Otherwise, have a 5% chance to take off at random, supposing gravity doesn't bring us back
        else {
            if (Random() < 0.05) steeringForce += floorNormal * getMaxForce();
        }

    }

    // Handle collision detection
    void updateIntegration(const double elapsedTime, const double currentTime)
    {
        TerrainAgent::updateIntegration(elapsedTime, currentTime);
        
        // Ad-hoc hack to make collisions effective & more efficient
        const double radius22 = 2.0 * radius * radius;
        
        osg::Vec3d collisionForward(0,0,0);
        int collision = 0;
        Localized* collidedAgent = 0;
        
        // collision detection. Use current neighbor list
        for (multimap<const double,Localized*>::iterator itn = neighbors->begin(); itn != neighbors->end(); ++itn) {
            if (itn->second == this) continue; // should not happen, but...
            
            double len = wrappedDist2(itn->second->getPosition(), newPosition);
            if (len <= radius22) {

                // Collision: bounce back from obstacle!
                collisionForward = wrappedDiffVector(newPosition, itn->second->getPosition());

                // In case of collision with another agent, add previous directions for consistency
                Agent* agent = dynamic_cast<Agent*>(itn->second);
                if (agent!=0) collisionForward += forward + agent->forward;

                collidedAgent = itn->second;
                if (collisionTarget==collidedAgent) collision = 1;
                else collision = 2;
                
                // One collision is enough to handle
                break;
            }
        }
        // Experimental adaptative search radius.
        if (collision) updateSearchRadius(neighbors->size());

        // no collision: done!
        if (!collision) return;

        // change direction
        if (collisionForward.normalize()==0)
            collisionForward = -newForward; // bounce back if exactly at the same place of colliding target

        // Attitude quaternion must be updated since we change direction
        attitudeCorrection->updateAttitude(newAttitude, newAttitude, collisionForward, newForward);

        // Now update our field        
        newForward = collisionForward;

        // Hack to slow down on collision
        newSpeed *= 0.8;

        // avoid duplicates by setting flag for collided target => if flag already set, return
        if (collisionHappened) return;

        // different target, but this agent matches the other agent's target
        // => take the other into account, not this one
        if (collision==2) {
            CollisionTerrainAgent* cta = dynamic_cast<CollisionTerrainAgent*>(collidedAgent);
            if ((cta!=0) && (cta->collisionTarget == this)) {
                collisionHappened = true;
                return;
            }
        }
        
        // display counter
        if (collision==1) {
            int dist = (int)(collisionDistance / radius);
            if (dist>6) dist=6;
            if (dist<1) dist=1;
            ++collisionNumbers[dist];
        } else collisionNumbers[100]++;
        ostringstream os;
        os << "Collisions ";
        for (map<int,long>::iterator it = collisionNumbers.begin(); it!=collisionNumbers.end(); ++it) {
            //if (it==collisionNumbers.begin()) os << "<=";
            //else if (it->first==6) os << ", >=";
            //else os << ", ";
            if (it!=collisionNumbers.begin()) os << ", ";
            if (it->first==100) os << "wrong target";
            else os << it->first;
            os << ":" << it->second;
        }
        Utility::setStatus(os.str());
        
        collisionHappened = true;
        CollisionTerrainAgent* cta = dynamic_cast<CollisionTerrainAgent*>(collidedAgent);
        if (cta!=0) cta->collisionHappened = true;
        
/*
        ++collisionNumber;

        ostringstream os;
        os << "Collisions: " << collisionNumber;
        Utility::setStatus(os.str(), Utility::TOP_RIGHT);*/
    
    }

    /// For fun, just make a different avatar for each agent
    osg::Node * buildAvatar()
    {
        if (!body) {
            body = new osg::Box(osg::Vec3(0,0,0),1,2,0.5);
            leftEye = new osg::ShapeDrawable(new osg::Sphere(osg::Vec3(-0.5,1,0.25),0.1));
            rightEye = new osg::ShapeDrawable(new osg::Sphere(osg::Vec3(0.5,1,0.25),0.1));
            osg::Box* queueBox = new osg::Box(osg::Vec3(0,-1.1,0.25),0.1,0.5,0.1);
            queueBox->setRotation(osg::Quat(-M_PI*0.25, osg::Vec3d(1,0,0)));
            queue = new osg::ShapeDrawable(queueBox);
        }
        
        // create a container that makes the cube drawable
        osg::ShapeDrawable *drawableBody = new osg::ShapeDrawable(body);

        // pick a color at random
        double coef[3];
        coef[0] = 0.6 * Random() + 0.4;
        coef[1] = 0.6 * Random() + 0.4;
        coef[2] = 1.0 - coef[1] - coef[0];
        if (coef[2]<0) coef[2] = 0;
        int i0 = IRandom(0,3);
        int i1; do {i1 = IRandom(0,3);} while (i1==i0);
        int i2 = 3 - i1 - i0; //if (i0+i1==1) i2=2; else if (i0+i1==2) i2 = 1; else i2 = 0;
        drawableBody->setColor(osg::Vec4(
            coef[i0],
            coef[i1],
            coef[i2],
            1
        ));
    
        // OSG used geodes as generic containers. Bridge from osg::Shape to osg::Node
        osg::Geode* geode = new osg::Geode();
        geode->addDrawable(drawableBody);
/*  // takes quite a lot of perf! => todo: build only one geometry
        geode->addDrawable(leftEye);
        geode->addDrawable(rightEye);
        geode->addDrawable(queue);
*/        return geode;
    }
    
    static osg::Shape *body;
    static osg::ShapeDrawable *leftEye, *rightEye, *queue;
    

};

const double CollisionTerrainAgent::maxSearchRadius = 40.0;
const double CollisionTerrainAgent::minSearchRadius = 20.0;
map<int,long> CollisionTerrainAgent::collisionNumbers;
map<int,long> CollisionTerrainAgent::collisionAvoidedNumbers;

osg::Shape *CollisionTerrainAgent::body = 0;
osg::ShapeDrawable *CollisionTerrainAgent::leftEye = 0;
osg::ShapeDrawable *CollisionTerrainAgent::rightEye = 0;
osg::ShapeDrawable *CollisionTerrainAgent::queue = 0;

osg::Group *metaroot = 0;

int main(int argc, char *argv[])
{
    osg::Referenced::setThreadSafeReferenceCounting(true);
    RandomInit();

    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();

    root = new osg::Group();
    
    //root->addChild(createFloor(200));

//    theFloor = new Terrain(800, 100, 7, 0.3, 3.0);
//    theFloor = new Terrain(400, 50, 7, 0.3, 3.0);
    theFloor = new Terrain(500, 100, 7, 0.30, 3.5);
    TerrainAgent::terrain = theFloor;
    root->addChild(theFloor);
    

    //theFloor->debugNormals();

    metaroot = new osg::Group();
    metaroot->addChild(root);

    createViewer("Collision & Terrain");
    
    createKeyboardHandler(root, metaroot);

    osg::Group * wholeScene = new osg::Group();
    wholeScene->addChild(metaroot);
    createKeyboardHandler(root, metaroot);
    
    viewer->setSceneData( wholeScene );
    viewer->realize();
    
    //Localized::setDefaultLocalityMode(Localized::Mixed);
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        100,100,50);

    AgentCore::setDefaultMaxForce(30);
    AgentCore::setDefaultMaxSpeed(8);
    gravity=20;

    agentsIntegrationTimeStep = 0.04; // 25 fps
    agentsAITimeStep = 0.1;

// Obstacle code
    const int nobstacles = 100;
    for (int i=0; i<nobstacles; ++i) {
        double x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        
        // Position Obstacle at floor surface
        Obstacle *obs = new Obstacle(x, y, theFloor->getFloorZ(x,y));
        obs->addAvatarToGroup(root);

        // Obstacles update their neighbor lists to help agents in Mixed mode
        if (Localized::getDefaultLocalityMode()==Localized::Mixed) {
            simu->postCyclicAt(
                // add jitter so all ai computations not done at the same time
                agentsAITimeStep * Random(),
                agentsAITimeStep, obs, &Obstacle::updateNeighbors, 30.0);
        }
        
    }

    
// Agent code
    const int nagents = 500;
    for (int i=0; i<nagents; ++i) {
        CollisionTerrainAgent *agent = new CollisionTerrainAgent();
        double x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));
        agent->addAvatarToGroup(root);

        simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            agentsAITimeStep * Random(),
            agentsAITimeStep, agent, &Agent::aiStep);
        
        simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            agentsIntegrationTimeStep * Random(),
            agentsIntegrationTimeStep, agent, &Agent::integrationStep);

        // add roughly one fourth of the agents to the locality database
        // the others keep their own neighborhood
        if ((Localized::getDefaultLocalityMode()==Localized::Bricks)
        || ((Localized::getDefaultLocalityMode()==Localized::Mixed) && (Random() < 0.25)))
            agent->addToLocalityDatabase();
        
    }

    
    // Provide an initial list of neighbors to all agents in mixed mode
    if (Localized::getDefaultLocalityMode()==Localized::Mixed) {
        for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
            (*it)->findNeighbors(30.0);
        }
        // Second order search
        for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
            (*it)->findNeighbors(30.0);
        }
    }
   
    mainLoop();
    
/*

    // Uncomment for stats. Could also use a logger

    int n=0;
    double meanRadius = 0.0;
    double meanMinThreshold = 0.0;
    double meanMaxThreshold = 0.0;
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        CollisionTerrainAgent* neighbor = dynamic_cast<CollisionTerrainAgent*>(*it);
        if (!neighbor) continue;
        meanRadius += neighbor->searchRadius;
        meanMinThreshold += neighbor->minThreshold;
        meanMaxThreshold += neighbor->maxThreshold;
        ++n;
    }
    cout << "Average search radius=" << meanRadius/(double)n << ", average min threshold=" << meanMinThreshold/n << ", average max threshold=" << meanMaxThreshold/n << endl;

*/

    return 0;
}



