/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osgDB/ReadFile>
#include <osg/Material>
#include <osg/Geode>

#include <iostream>
#include <sstream>

#include "Environment.h"
#include "BillboardObstacle.h"
#include "GrassyTerrain.h"
#include "ExplicitAIAgent.h"
#include "Utility.h"

using namespace std;
using namespace Environment;
using namespace Utility;

static Terrain* theFloor = 0;

static osg::Group* metaroot = 0;

static double initialTime;

static std::map<int, int> maxPopulations;
static Logger *extinctionsLogger = 0;

void populationUpdateCallback(int delta, int specie) {
    if (delta>0) {
        int p = CrogaiAgent::populations[specie];
        if (p > maxPopulations[specie]) maxPopulations[specie] = p;
        if (p>=3000) {
            double timeSpent = simu->getSystemTime() - initialTime;
            ostringstream os;
            os << "Explosion of specie " << specie <<". max pop reached: " << maxPopulations[specie];
            setStatus(os.str());
            // stop simu loop
            simu->stop();
        }
        return;
    }
    // so, delta negative. Check extinctions
    if (CrogaiAgent::populations[specie]!=0) return;
    
    (*extinctionsLogger)[0] = simu->getTime();
    (*extinctionsLogger)[1] = specie;
    (*extinctionsLogger)[2] = maxPopulations[specie];
    extinctionsLogger->record();

    double timeSpent = simu->getSystemTime() - initialTime;
    ostringstream os;
    os << "Extinction of specie " << specie <<". max pop reached: " << maxPopulations[specie];
    setStatus(os.str());
    // stop simu loop when all extinctions have occured
    if (Agent::allAgents.size()==0) simu->stop();
}

int main(int argc, char *argv[])
{
    if (argc>1) batch = true;
    
    osg::Referenced::setThreadSafeReferenceCounting(true);

    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();
    if (batch) simu->setTimeFactor(100000.0);
    else simu->setTimeFactor(1.0);

    
    // Use explicit seed instead of system time
    // TODO: use command-line argument
    //RandomInit(47);
    
    // debug code
    //statusToConsole = true;
    //simu->postAt(30.0, &exit, 0);

    if (batch) {
        statusTimeStamped = true;
        CrogaiAgent::displayPopulationCounter = false;
    }
    else toggleSimulationTime();
    CrogaiAgent::displayScores = false;
    CrogaiAgent::displayPopulationNumbers = false;
    
/// Set global parameters
    

    ExplicitAIAgent::maxMass = 2.0;
    ExplicitAIAgent::minMass = 0.5;
    AgentCore::setDefaultInitialMass(1.5); // for the first generation
    
    AgentCore::setDefaultEnergyFootprintRatio(0.001);
    // Energy spent by parent to create a child. Child will have initial energy.
    // this may possibly be more than initial energy. The system is dissipative, but
    // at this point, the child is created with a mass => this is where the extra energy comes from,
    // when it gets eaten. 
    //ExplicitAIAgent::spawningEnergyCost = 580.0;
    AgentCore::setDefaultMaxEnergy(5000.0);
    // Don't give max to begin with... let them seek it!
    double initialEnergy = 800.0;
    ExplicitAIAgent::sporeEmissionEnergyThreshold = 1000.0;
    ExplicitAIAgent::sporeEmissionEnergyCost = 20.0;
    AgentCore::setDefaultSelfMaintainEnergyCost(1.0); // in energy per simulator time unit
    ExplicitAIAgent::sporeHalfLife = 120.0;

    AgentCore::setDefaultMaxForce(20.0);
    AgentCore::setDefaultMaxSpeed(10.0);
    ExplicitAIAgent::maxGrazingSpeed = 1.0;
    gravity=16.0;
    
    CrogaiAgent::seeOnlyInFront = true;

    agentsIntegrationTimeStep = 0.04; // 25 fps
    agentsAITimeStep = 0.1;
    cloudIntegrationTimeStep = 0.3;

    const int nobstacles = 100;
    const int nagents = 1500;
    const int nspecies = 50;

    const double diffusionCoef = 40.0;
    const double diffusionCutoff = 0.001;

    GrassyTerrain* grass = new GrassyTerrain(500, 100, 7, 0.35, 3.5);
    TerrainAgent::terrain = theFloor = grass;
    grass->setMaximumGrassDensity(50.0);
    grass->setGrassGrowthRate(0.5);
    grass->setGrassDensity(50.0);
    grass->scheduleForGrowth(agentsAITimeStep);

    if (!batch) {
        createViewer("Random ecosystem");
        root = new osg::Group();
        root->addChild(theFloor);
        metaroot = new osg::Group();
        metaroot->addChild(root);
        osg::Group * wholeScene = new osg::Group();
        wholeScene->addChild(metaroot);
        createKeyboardHandler(root, metaroot);
        
        viewer->setSceneData( wholeScene );
        viewer->realize();
    }

    // create the spore diffusion cloud
    cloud = new DrawableSporeCloud(theFloor, 32, 32, 10, diffusionCoef, diffusionCutoff);
    // stable condition error already written on cerr in this case
    if (!cloud->schedule(cloudIntegrationTimeStep)) exit(1);

    
    if (!batch) {
        cloud->setDefaultAttributes("smoke.rgb", false, false);
        osg::Geode* geode = new osg::Geode();
        geode->addDrawable(cloud);
        root->addChild(geode);
    }
    
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        50,50,20);

    // Create obstacles
    osg::Texture2D *texture0 = new osg::Texture2D;
    if (!batch) texture0->setImage(osgDB::readImageFile("tree.rgba"));
    osg::Texture2D *texture1 = new osg::Texture2D;
    if (!batch) texture1->setImage(osgDB::readImageFile("buisson.tga"));
    for (int i=0; i<nobstacles; ++i) {
        double x,y,z;
        
        // Position Obstacle at floor surface with bias for low altitude
        do {
            x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
            y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
            z = theFloor->getFloorZ(x,y);
        } while (z * Random() > theFloor->getMaxHeight() * 0.25);

        Obstacle *obs;
        switch (IRandom(0,5)) {
            case 0:
            case 1:
            case 2:
                obs = new BillboardObstacle(x, y, z, texture0,
                    2.0 + 2.0*Random(),
                    4.0 + 2.0*Random()
                );
                break;
            case 3:
            case 4:
                obs = new BillboardObstacle(x, y, z, texture1,
                    2.0 + Random(),
                    2.0 + Random()
                );
                break;
            case 5:
                obs = new Obstacle(x, y, z);
                break;
        };
        if (!batch) obs->addAvatarToGroup(root);
    }

    
// Agent code

    // create predation matrix
    // each specie predates on a random number of species randomly chosen
    map<int, vector<int> > preySpecies, predatorSpecies;
    for (int specie=1; specie<=nspecies; ++specie) {
        int n = IRandom(1,nspecies);
        preySpecies[specie].resize(n);
        for (int j=0; j<n; ++j) {
            int p;
            do {p = IRandom(1,nspecies);} while (p==specie);
            (preySpecies[specie])[j] = p;
            predatorSpecies[p].push_back(specie);
        }
    }

    // create agents
    for (int i=0; i<nagents; ++i) {
        int specie = (i%nspecies) + 1;
        ExplicitAIAgent *agent = new ExplicitAIAgent(specie, predatorSpecies[specie], preySpecies[specie]);
        double x = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Random() - theFloor->getHalfWorldSize();
        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));
        // give it initial energy only instead of max!
        agent->energy = initialEnergy;
        // let it move
        agent->speed = AgentCore::getDefaultMaxSpeed() * Random();
        
        if (!batch) agent->addAvatarToGroup(root);

        agent->aiEvent = simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            agentsAITimeStep * Random(),
            agentsAITimeStep, agent, &Agent::aiStep);

        agent->integrationEvent = simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            agentsIntegrationTimeStep * Random(),
            agentsIntegrationTimeStep, agent, &Agent::integrationStep);
        
        agent->addToLocalityDatabase();
    }
    
    CrogaiAgent::createPopulationsLogger("populations",nspecies);

    // These take too much memory to maintain, use only variables that can be updated live!
    //CrogaiAgent::createScoresLogger("scores",nspecies);
    //CrogaiAgent::createStatisticsLogger("statistics",nspecies);
    //ExplicitAIAgent::createGenomesLogger("genomes",nspecies);

    CrogaiAgent::addPopulationUpdateCallback(&populationUpdateCallback);
    initialTime = simu->getSystemTime();
    // initialize populations counter
    for (int i=1; i<=nspecies; ++i) maxPopulations[i] = CrogaiAgent::populations[i];
    vector<string> names; names.push_back("time"); names.push_back("specie"); names.push_back("maxPopulation");
    extinctionsLogger = Utility::createLogger("extinctions",names);

    mainLoop();
    
    // Tell all agents to update their stats now
    for (list<Agent*>::iterator it = Agent::allAgents.begin(); it != Agent::allAgents.end(); ++it) {
        CrogaiAgent* crogai = dynamic_cast<CrogaiAgent*>(*it);
        if (crogai!=0) crogai->updateStatistics();
    }
    CrogaiAgent::recordStatistics();
    ExplicitAIAgent::recordGenomesStatistics();
    // avoid useless updates on agents final destruction
    CrogaiAgent::logStatistics = 0;
    
    return 0;
}



