/*
// ----------------------------------------------------------------------------
//
//
// OpenSteer -- Steering Behaviors for Autonomous Characters
//
// Copyright (c) 2002-2003, Sony Computer Entertainment America
// Original author: Craig Reynolds <craig_reynolds@playstation.sony.com>
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// ----------------------------------------------------------------------------
*/
/* ------------------------------------------------------------------ */
/*                                                                    */
/* Locality Query facility                                            */
/*                                                                    */
/* (by Craig Reynolds, see lq.h file for documentation)               */
/*                                                                    */
/*  5-17-99: created                                                  */
/*  5-20-99: found elusive "allocate 0 bins" bug                      */
/*  5-28-99: lqMapOverAllObjectsInLocality: clipped, incremental      */
/*  6- 7-99: clean up, split off annotation stuff into debuglq.c      */
/*  6- 8-99: tried screening by sum of coords ("first mean"?) but     */
/*           it was slightly slower, moved unused code to debuglq     */
/* 10-19-99: Change lqClientObject, lqObject from: "struct x {};" to  */
/*           "typedef struct x {} x;" for EE compiler.                */
/* 12- 2-00: Make lqObject "private" using lqInternalDB.              */
/* 12- 5-00: Rename lqObject to lqDB, lqClientObject to lqClientProxy */
/* 12- 6-00: Change lqCallBackFunction from arglist of (void*) to:    */
/*           (void* clientObject, float distanceSquared, void*        */
/*           clientQueryState).  Add void* clientQueryState arg to    */
/*           lqMapOverAllObjectsInLocality and its helper functions   */
/*           lqMapOverAllObjectsInLocalityClipped and                 */
/*           lqMapOverAllOutsideObjects. Change macro                 */
/*           lqTraverseBinClientObjectList to invoke callback         */
/*           function with three arguments, add "state" to its        */
/*           arglist.  Remove extern lqDistanceSquared.               */
/* 12- 7-00: Rename lqInitClientObject to lqInitClientProxy, make     */
/*           "func" be an argument to lqTraverseBinClientObjectList,  */
/*           add comments.                                            */
/* 12- 8-00: Add lqFindNearestNeighborWithinRadius and related        */
/*           definitions: lqFindNearestHelper lqFindNearestState      */
/*           Add lqMapOverAllObjects and lqRemoveAllObjects (plus:    */
/*           lqMapOverAllObjectsInBin and lqRemoveAllObjectsInBin)    */
/*                                                                    */
/* ------------------------------------------------------------------ */

/* 30-11-04: by Nicolas Brodu: Made the bins cyclic in X and Y.
             This is adapted for proximity computation on a surface "tore",
             where the height in Z is considered infinite.
   27-01-05: by Nicolas Brodu: converted the whole thing to use double
             instead of float. Rational is double are better supported
             on current CPU.
*/

#include <stdlib.h>
#include "lqXYCyclic.h"
#include <stdio.h>
#include <math.h>

/*
It's a pity, but we really need internal lq.c types
Rather than duplicating the definitions, le'ts include the file
Of course, now an external file should include and link
with the XYCyclic files only, as they include the standard files
*/
#include "lq.c"

/* ------------------------------------------------------------------ */
/* Find the bin ID for a location in space.  The location is given in
   terms of its XYZ coordinates.  The bin ID is a pointer to
   the bin contents list.  */

lqBin* lqXYCyclicBinForLocation (lqInternalDB* lq, double x, double y, double z)
{
    int i, ix, iy, iz;

    /* if point outside super-brick, return the "other" bin */
    if (z < lq->originz)              return lq->other;
    if (z >= lq->originz + lq->sizez) return lq->other;

    x -= lq->originx;
    x = fmod(x, lq->sizex);
    if (x<0) x+= lq->sizex;
    ix = (int)(x * lq->divx / lq->sizex);
    
    y -= lq->originy;
    y = fmod(y, lq->sizey);
    if (y<0) y+= lq->sizey;
    iy = (int)(y * lq->divy / lq->sizey);

    /* if point inside super-brick, compute the bin coordinates */
    iz = (int) (((z - lq->originz) / lq->sizez) * lq->divz);

    /* convert to linear bin number */
    i = lqBinCoordsToBinIndex (lq, ix, iy, iz);

    /* return pointer to that bin */
    return &(lq->binArray[i]);
}


/* ------------------------------------------------------------------ */
/* Call for each client object every time its location changes.  For
   example, in an animation application, this would be called each
   frame for every moving object.  */


void lqXYCyclicUpdateForNewLocation  (lqInternalDB* lq,
			      lqClientProxy* object,
			      double x, double y, double z)
{
    /* find bin for new location */
    lqBin* newBin = lqXYCyclicBinForLocation (lq, x, y, z);

    /* store location in client object, for future reference */
    object->x = x;
    object->y = y;
    object->z = z;

    /* has object moved into a new bin? */
    if (newBin != object->bin)
    {
	lqRemoveFromBin (lq, object);
 	lqAddToBin (lq, object, newBin);
    }
}


/* ------------------------------------------------------------------ */
/* Given a bin's list of client proxies, traverse the list and invoke
   the given lqCallBackFunction on each object that falls within the
   search radius.  */

#define lqXYCyclicTraverseBinClientObjectList(co, radiusSquared, func, state) \
    while (co != NULL)                                                  \
    {                                                                   \
    /* compute distance (squared) from this client   */                 \
    /* object to given locality sphere's centerpoint */                 \
    /* NB043011: remainder is our friend to find wrap-up distance! */   \
    dx = remainder(x - co->x, lq->sizex);                  \
    dy = remainder(y - co->y, lq->sizey);                  \
    dz = z - co->z;                                               \
    distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);          \
                                                                        \
    /* apply function if client object within sphere */                 \
    if (distanceSquared < radiusSquared)                                \
       (*func) (co->object, distanceSquared, state);                    \
                                                                        \
    /* consider next client object in bin list */                       \
    co = co->next;                                                      \
    }


/* ------------------------------------------------------------------ */
/* This subroutine of lqXYCyclicMapOverAllObjectsInLocality efficiently
   traverses of subset of bins specified by max and min bin
   coordinates. */
/* NB043011: This clipped routine is always called with correct min/max bin ranges */

void lqXYCyclicMapOverAllObjectsInLocalityClipped (lqInternalDB* lq,
					   double x, double y, double z,
					   double radius,
					   lqCallBackFunction func,
					   void* clientQueryState,
					   int minBinX,
					   int minBinY, 
					   int minBinZ,
					   int maxBinX,
					   int maxBinY,
					   int maxBinZ)
{
    int i, j, k;
    int iindex, jindex, kindex;
    int slab = lq->divy * lq->divz;
    int row = lq->divz;
    int istart = minBinX * slab;
    int jstart = minBinY * row;
    int kstart = minBinZ;
    lqClientProxy* co;
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    lqBin* bin;
    double radiusSquared = radius * radius;
    /* Sphere test: consider the bounding sphere of each brick. It has radius
       precomputed at database creation time. Consider Dc the distance from the
       locality query point to the center of the sphere.
       Then if Dc - brickBoundingRadius > neighborhood query radius, the whole
       brick is outside the neighborhood query sphere.
       Thus, test for Dc^2 > sphereTest with sphereTest a constant defined by
       sphereTest = (neighborhood query radius + brickBoundingRadius)^2
    */
    double sphereTest = radius + lq->brickBoundingRadius;
    sphereTest *= sphereTest;

    /* loop for x bins across diameter of cube */
    iindex = istart;
    for (i = minBinX; i <= maxBinX; i++)
    {
        /* loop for y bins across diameter of cube */
        jindex = jstart;
        for (j = minBinY; j <= maxBinY; j++)
        {
            /* loop for z bins across diameter of cube */
            kindex = kstart;
            for (k = minBinZ; k <= maxBinZ; k++)
            {
                bin = lq->bins[iindex + jindex + kindex];
                if (bin==0) {
                    kindex += 1;
                    continue;
                }
                
                /* Eliminate bricks clearly outside sphere */
                dx = remainder(x - bin->x, lq->sizex);
                dy = remainder(y - bin->y, lq->sizey);
                dz = bin->z - z;
                distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);
                
                /* If the whole brick is outside sphere, its elements can't be in! */
                if (distanceSquared > sphereTest) {
                    kindex += 1;
                    continue;
                }
                
                /* get current bin's client object list */
                co = bin->objects;

                /* traverse current bin's client object list */
                lqXYCyclicTraverseBinClientObjectList (co,
                    radiusSquared,
                    func,
                    clientQueryState);


                kindex += 1;
            }
            jindex += row;
        }
        iindex += slab;
    }
}


/* ------------------------------------------------------------------ */
/* If the query region (sphere) extends outside of the "super-brick"
   we need to check for objects in the catch-all "other" bin which
   holds any object which are not inside the regular sub-bricks  */


void lqXYCyclicMapOverAllOutsideObjects (lqInternalDB* lq,
				 double x, double y, double z,
				 double radius,
				 lqCallBackFunction func,
				 void* clientQueryState)
{
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    lqClientProxy* co = lq->other->objects;
    double radiusSquared = radius * radius;

    /* traverse the "other" bin's client object list */
    lqXYCyclicTraverseBinClientObjectList (co,
				   radiusSquared,
				   func,
				   clientQueryState);


}



/* ------------------------------------------------------------------ */
/* If there is a gain by mapping over the non-empty list, in terms of
   number of cells, then use it.
  */

   
void lqXYCyclicMapOverNonEmptyList(lqInternalDB* lq, 
                 double x, double y, double z,
                 double radius,
                 lqCallBackFunction func,
                 void* clientQueryState)
{
    double dx;
    double dy;
    double dz;
    double distanceSquared;
    lqClientProxy* co;
    lqBin* bin;
    double radiusSquared = radius * radius;
    /* Sphere test: consider the bounding sphere of each brick. It has radius
       precomputed at database creation time. Consider Dc the distance from the
       locality query point to the center of the sphere.
       Then if Dc - brickBoundingRadius > neighborhood query radius, the whole
       brick is outside the neighborhood query sphere.
       Thus, test for Dc^2 > sphereTest with sphereTest a constant defined by
       sphereTest = (neighborhood query radius + brickBoundingRadius)^2
    */
    double sphereTest = radius + lq->brickBoundingRadius;
    sphereTest *= sphereTest;


    for (bin = lq->nonEmptyList; bin!=NULL; bin = bin->nonEmptyNext) {

/*
TODO:
avoid dereferencing bin->x
To do this, deduce bin position from index
TODO 2: To to this, pre-compute shifts when divy and divz are powers of 2, to avoid divisions.

//#define lqBinCoordsToBinIndex(lq, ix, iy, iz) \
//    ((ix * (lq)->divy * (lq)->divz) + (iy * (lq)->divz) + iz)

*/
/*        {
        long offset = (long)(bin - lq->binArray);
        long dyz = ((lq)->divy * (lq)->divz);
        long ix = offset / dyz;
        long rem = offset - ix * dyz;
        long iy = rem / (lq)->divz;
        long iz = rem - (iy * (lq)->divz);

        double xxx = lq->originx + (ix+0.5) * lq->sizex / (double)lq->divx - bin->x;
        double yyy = lq->originy + (iy+0.5) * lq->sizey / (double)lq->divy - bin->y;
        double zzz = lq->originz + (iz+0.5) * lq->sizez / (double)lq->divz - bin->z;
        }
*/
        
        /* Eliminate bricks clearly outside sphere
           The idea is to avoid de-referencing co = bin->objects
        */
        dx = remainder(x - bin->x, lq->sizex);
        dy = remainder(y - bin->y, lq->sizey);
        dz = z - bin->z;
        distanceSquared = (dx * dx) + (dy * dy) + (dz * dz);
        
        /* If the whole brick is outside sphere, its elements can't be in! */
        if (distanceSquared > sphereTest) continue;
        
        /* get current bin's client object list */
        co = bin->objects;
        
        /* traverse current bin's client object list */
        lqXYCyclicTraverseBinClientObjectList (co,
                radiusSquared,
                func,
                clientQueryState);

    }

}


/* ------------------------------------------------------------------ */
/* Apply an application-specific function to all objects in a certain
   locality.  The locality is specified as a sphere with a given
   center and radius.  All objects whose location (key-point) is
   within this sphere are identified and the function is applied to
   them.  The application-supplied function takes three arguments:

     (1) a void* pointer to an lqClientProxy's "object".
     (2) the square of the distance from the center of the search
         locality sphere (x,y,z) to object's key-point.
     (3) a void* pointer to the caller-supplied "client query state"
         object -- typically NULL, but can be used to store state
         between calls to the lqCallBackFunction.

   This routine uses the LQ database to quickly reject any objects in
   bins which do not overlap with the sphere of interest.  Incremental
   calculation of index values is used to efficiently traverse the
   bins of interest. */


void lqXYCyclicMapOverAllObjectsInLocality (lqInternalDB* lq,
				    double x, double y, double z,
				    double radius,
				    lqCallBackFunction func,
				    void* clientQueryState)
{
    int partlyOut = 0;
    int minBinX, minBinY, minBinZ, maxBinX, maxBinY, maxBinZ, tmp;

    /* is the sphere completely outside the "super brick"? */
    if ((((z + radius) < lq->originz) || ((z - radius) >= lq->originz + lq->sizez))) {
        lqXYCyclicMapOverAllOutsideObjects (lq, x, y, z, radius, func,
            clientQueryState);
        return;
    }
    
    /* compute min and max bin coordinates for each dimension */
    /* NB043011: use floor to compute negative bounds in X & Y */
    minBinX = (int) floor((((x - radius) - lq->originx) / lq->sizex) * lq->divx);
    minBinY = (int) floor((((y - radius) - lq->originy) / lq->sizey) * lq->divy);
    minBinZ = (int) ((((z - radius) - lq->originz) / lq->sizez) * lq->divz);
    maxBinX = (int) floor((((x + radius) - lq->originx) / lq->sizex) * lq->divx);
    maxBinY = (int) floor((((y + radius) - lq->originy) / lq->sizey) * lq->divy);
    maxBinZ = (int) ((((z + radius) - lq->originz) / lq->sizez) * lq->divz);
    
    /* clip bin coordinates */
    /* NB043011: don't clip negative X & Y, but round-up */
    if (minBinZ < 0)         {partlyOut = 1; minBinZ = 0;}
    if (maxBinZ >= lq->divz) {partlyOut = 1; maxBinZ = lq->divz - 1;}

    /* map function over outside objects if necessary (if clipped) */
    if (partlyOut) 
    lqXYCyclicMapOverAllOutsideObjects (lq, x, y, z, radius, func,
                                        clientQueryState);

    /* NB043011: wrap-up, but no more than maximum number of divisions! */
    tmp = maxBinX - minBinX;
    if (tmp >= lq->divx) {
        minBinX = 0;
        maxBinX = lq->divx - 1;
    } else {
        if (minBinX < 0) {
            minBinX = (minBinX % lq->divx) + lq->divx;
            maxBinX = minBinX + tmp;
        } else if (minBinX >= lq->divx) {
            minBinX = (minBinX % lq->divx);
            maxBinX = minBinX + tmp;
        }
        if (maxBinX >= lq->divx) {
            maxBinX %= lq->divx;
        }
    }
    
    /* NB043011: round-up, but no more than maximum number of divisions! */
    tmp = maxBinY - minBinY;
    if (tmp >= lq->divy) {
        minBinY = 0;
        maxBinY = lq->divy - 1;
    } else {
        if (minBinY < 0) {
            minBinY = (minBinY % lq->divy) + lq->divy;
            maxBinY = minBinY + tmp;
        } else if (minBinY >= lq->divy) {
            minBinY = (minBinY % lq->divy);
            maxBinY = minBinY + tmp;
        }
        if (maxBinY >= lq->divy) {
            maxBinY %= lq->divy;
        }
    }
    
    /* NB043011: now wrap-up detected by max<min
       Call the clipped function in all domains
    */

    /* standard case for X */
    if (minBinX < maxBinX) {
        /* standard case for Y */
        if (minBinY < maxBinY) {

            /* Apply non-empty list traversal if there is a gain in number of cells */
            if (lq->nonEmptyListSize < (maxBinX - minBinX + 1) * (maxBinY - minBinY + 1) * (maxBinZ - minBinZ + 1)) {
                lqXYCyclicMapOverNonEmptyList(lq, x, y, z, radius, func, clientQueryState);
                return;
            }
            
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, minBinY, minBinZ,
                            maxBinX, maxBinY, maxBinZ);
        }
        /* Split domain for Y */
        else {

            /* Apply non-empty list traversal if there is a gain in number of cells */
            if (lq->nonEmptyListSize < (maxBinX - minBinX) * (maxBinY /*- 0*/ + 1 + lq->divy - minBinY) * (maxBinZ - minBinZ)) {
                lqXYCyclicMapOverNonEmptyList(lq, x, y, z, radius, func, clientQueryState);
                return;
            }
            
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, 0, minBinZ,
                            maxBinX, maxBinY, maxBinZ);
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, minBinY, minBinZ,
                            maxBinX, lq->divy - 1, maxBinZ);
        }        
    }
    /* Split domain for X */
    else {
        /* standard case for Y */
        if (minBinY < maxBinY) {
            
            /* Apply non-empty list traversal if there is a gain in number of cells */
            if (lq->nonEmptyListSize < (maxBinX /*- 0*/ + 1 + lq->divx - minBinX) * (maxBinY - minBinY) * (maxBinZ - minBinZ)) {
                lqXYCyclicMapOverNonEmptyList(lq, x, y, z, radius, func, clientQueryState);
                return;
            }
            
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            0, minBinY, minBinZ,
                            maxBinX, maxBinY, maxBinZ);
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, minBinY, minBinZ,
                            lq->divx - 1, maxBinY, maxBinZ);
        }
        /* Split domain for Y */
        else {

            /* Apply non-empty list traversal if there is a gain in number of cells */
            if (lq->nonEmptyListSize < (maxBinX /*- 0*/ + 1 + lq->divx - minBinX) * (maxBinY /*- 0*/ + 1 + lq->divy - minBinY) * (maxBinZ - minBinZ)) {
                lqXYCyclicMapOverNonEmptyList(lq, x, y, z, radius, func, clientQueryState);
                return;
            }
            
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            0, 0, minBinZ,
                            maxBinX, maxBinY, maxBinZ);
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, 0, minBinZ,
                            lq->divx - 1, maxBinY, maxBinZ);
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            0, minBinY, minBinZ,
                            maxBinX, lq->divy - 1, maxBinZ);
            lqXYCyclicMapOverAllObjectsInLocalityClipped (lq,
                            x, y, z,
                            radius,
                            func,
                            clientQueryState,
                            minBinX, minBinY, minBinZ,
                            lq->divx - 1, lq->divy - 1, maxBinZ);
        }        
    }

}


/* ------------------------------------------------------------------ */
/* Search the database to find the object whose key-point is nearest
   to a given location yet within a given radius.  That is, it finds
   the object (if any) within a given search sphere which is nearest
   to the sphere's center.  The ignoreObject argument can be used to
   exclude an object from consideration (or it can be NULL).  This is
   useful when looking for the nearest neighbor of an object in the
   database, since otherwise it would be its own nearest neighbor.
   The function returns a void* pointer to the nearest object, or
   NULL if none is found.  */


void* lqXYCyclicFindNearestNeighborWithinRadius (lqInternalDB* lq,
					 double x, double y, double z,
					 double radius,
					 void* ignoreObject)
{
    /* initialize search state */
    lqFindNearestState lqFNS;
    lqFNS.nearestObject = NULL;
    lqFNS.ignoreObject = ignoreObject;
    lqFNS.minDistanceSquared = FLT_MAX;

    /* map search helper function over all objects within radius */
    lqXYCyclicMapOverAllObjectsInLocality (lq,
				   x, y, z,
				   radius,
				   lqFindNearestHelper,
				   &lqFNS);

    /* return nearest object found, if any */
    return lqFNS.nearestObject;
}

