/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osg/ShapeDrawable>
#include <osgProducer/Viewer>
#include <osgGA/GUIEventHandler>

#include <iostream>
#include <fstream>

//#include <OpenThreadSimulator.h>
#include <time.h>

#include <map>

#include "Terrain.h"
#include "Environment.h"
#include "TranslateTransform.h"
#include "KeepUpAttitudeCorrection.h"
#include "Obstacle.h"
#include "TwoLayerPerceptron.h"
#include "NeuronalAgent.h"

using namespace std;
using namespace Environment;

double istep = 0.04; // 25 f/s
double aistep = 0.07;

static const double initialEnergy = 500.0;
static const double reproductionEnergy = 1000.0;
static const int maxReproductionNeighbors = 5;
static const int maxPopulationSize = 300;
static const double maxForce = 20;
static const double worldGravity = 17;
static const double maxSpeed = 10;
static const double energyFootprintRatio = 0.001;


Terrain* theFloor = 0;

osg::Group* root = 0, *metaroot = 0;
static int blueCount=0; 
static int redCount=0;

struct ExplicitAIAgent : public CrogaiAgent {

    ExplicitAIAgent(int specie) : CrogaiAgent(specie) {}

    void updateAI( const double elapsedTime, const double currentTime) {
        CrogaiAgent::updateAI(elapsedTime,currentTime);
        
        osg::Vec3d localCenter(0,0,0);
        osg::Vec3d forwardForce(0,0,0);
        osg::Vec3d obsForce(0,0,0); // obstacle avoidance
        osg::Vec3d seekFleeForce(0,0,0); // obstacle avoidance
        
        // avoidance force. Avoid only the closest object
        // Since neighbors are returned sorted, this is easy
        osg::Vec3d avoidForce(0,0,0);
        
        int nSeekFlee = 0; // number of neighbor preys + predators
        int n = 0; // number of neighbor agents
        
        findNeighbors(searchRadius);
        neighborsAccessLock();
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

                if (wrappedDiffVector(it->second->getPosition(), getPosition()) * forward < 0)
                    continue; // see only what's in front of us

                Agent* neighbor = dynamic_cast<Agent*>(it->second);

                // if not an agent, maybe it is an obstacle?
                if (!neighbor) {
                    Obstacle* obs = dynamic_cast<Obstacle*>(it->second);
                    // Avoid obstacles, but only the closest one for which there is a collision risk
                    if (obs!=0) {
                        avoidForce = steerToAvoid(obs->getNode(), searchRadius);
                        break;
                    }
                    continue;
                }


                ExplicitAIAgent* sa = dynamic_cast<ExplicitAIAgent*>(neighbor);
                // Escape from unknown things from outer space!!!
                if (sa==0) {
                    steeringForce += steerForEvasion(neighbor);
                }
                // One of our mates. Boidizing...
                else if (sa->specie == specie) {
                    // Check for risk of collision
                    avoidForce = steerToAvoid(neighbor, searchRadius);
                    
                    // Collision risk => no need to look for further AI
                    if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0))
                        break;
    
                    steeringForce += steerForFlee(neighbor->getPosition()) / (it->first+0.001);
                    forwardForce += neighbor->forward;
                    localCenter += neighbor->getPosition();
                    ++n;
                }
                // Hey, another guy! Let's try to catch it unaware...
                else {
                    // hehe, we are behind, it can't see us
                    if (sa->forward * forward < 0) {
                        seekFleeForce += steerForPursuit(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                    // Oops, avoid fight
                    else {
                        seekFleeForce += steerForEvasion(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                }
        }
        neighborsAccessUnlock();

        // Collision risk? Priority over anything else
        if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0)) {

            // Don't try to avoid by digging a hole, it doesn't work!
            // When not flying, project avoidance direction on the horizontal plane
            if (altitude<radius) {
                steeringForce = avoidForce - floorNormal * (avoidForce*floorNormal);
                steeringForce *= sqrt(getMaxForceSquared() / steeringForce.length2());
            }
            else steeringForce = avoidForce;

            return;
        }

        // don't move to the origin when no other agent present
        // In this case, move at random, but only if not already doing something else
        if ((n==0) && (nSeekFlee==0)) steeringForce += steerForWander(0.4);

        // Otherwise, agent behavior when in company of friends
        // - steer to avoid neighbors
        // - steer toward the local center
        // - and also to keep the local group direction
        else if (n>0) {
            steeringForce /= (double)n; // average of avoidance vectors
            steeringForce += steerForSeek(localCenter/(double)n) * 2; // local center
            steeringForce += forwardForce /(double)n  -  forward; // local group direction
        }

        // prey/predator avoidance?
        if (nSeekFlee>0) {
            steeringForce += seekFleeForce * 10; // give it more weight than boidizing
        }
        
        // If flying, try to avoid floor, up to max height were we're "safe"
        if (altitude>radius) {
            if ( altitude < radius*5.0) 
                steeringForce += floorNormal * getMaxForce();
        }
        
        // Otherwise, have a 1% chance to take off at random, supposing gravity doesn't bring us back
        else {
            if (Utility::Random() < 0.01) steeringForce += floorNormal * getMaxForce();
        }
        

    }


    virtual void winnerAction(CrogaiAgent* looser) {
        CrogaiAgent::winnerAction(looser);
        cout << "Team " << specie << " winner, energy=" << energy << ". Scores: ";
        for (map<int, int>::iterator it = scores.begin(); it != scores.end(); ++it) {
            if (it!=scores.begin()) cout << ", ";
            cout << it->first << ":" << it->second;
        }
        cout << endl;
    }

/*
    virtual void looserAction( CrogaiAgent * winner ) {
        CrogaiAgent::looserAction(winner);

        cout << "Scores: ";
        for (int i=0; i<scores.size(); ++i) {
            if (i!=0) cout << ", ";
            cout << (i+1) << ":" << scores[i+1];
        }
        cout << endl;
    }
*/
    
};

osg::Group *normalGroup = 0;

class KeyboardEventHandler : public osgGA::GUIEventHandler {
public:

    void translateRoot(int tx, int ty) {
        double s = theFloor->getWorldSize();
        TranslateTransform* tt=0;
        tt = new TranslateTransform(tx*s, ty*s, 0);
        tt->addChild(root);
        metaroot->addChild(tt);
    }

    virtual bool handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter&) {
        switch(ea.getEventType()) {

            case osgGA::GUIEventAdapter::KEYDOWN:
                switch(ea.getKey()) {
                    case 'g':
                        gravity += 0.1;
                        cout << "gravity is now " << gravity << endl;
                        break;
                    case 'G':
                        gravity -= 0.1;
                        if (gravity<0) gravity=0;
                        cout << "gravity is now " << gravity << endl;
                        break;

                    case 'n':
                    case 'N':
                        if (normalGroup==0) {
                            normalGroup = theFloor->drawNormals();
                            root->addChild(normalGroup);
                            cout << "Floor normals on" << endl;
                        }
                        else {
                            root->removeChild(normalGroup);
                            normalGroup = 0;
                            cout << "Floor normals off" << endl;
                        }
                        break;

                    case 'C':
                    case 'c':
                        // keep only the first child, root
                        if (metaroot->getNumChildren() > 9) {
                            metaroot->removeChild(1, metaroot->getNumChildren() -1);
                            cout << "Cyclic terrain visualization off" << endl;
                            break;
                        }
                        if (metaroot->getNumChildren() >1 ) {
                            translateRoot(-2,-2);
                            translateRoot(-2,-1);
                            translateRoot(-2,0);
                            translateRoot(-2,1);
                            translateRoot(-2,2);
                            translateRoot(2,-2);
                            translateRoot(2,-1);
                            translateRoot(2,0);
                            translateRoot(2,1);
                            translateRoot(2,2);
                            translateRoot(-1,-2);
                            translateRoot(0,-2);
                            translateRoot(1,-2);
                            translateRoot(-1,2);
                            translateRoot(0,2);
                            translateRoot(1,2);
                            cout << "Cyclic terrain visualization: level 2" << endl;
                            break;
                        }
                        translateRoot(1,0);
                        translateRoot(-1,0);
                        translateRoot(0,1);
                        translateRoot(0,-1);
                        translateRoot(1,1);
                        translateRoot(-1,1);
                        translateRoot(1,-1);
                        translateRoot(-1,-1);

                        cout << "Cyclic terrain visualization: level 1" << endl;
                        break;

                    
                    default:
                        return false;
                }
                return true;
/*
            case osgGA::GUIEventAdapter::KEYUP:
                return true;
*/
            default:
                return false;
        }
    }

    virtual void accept(osgGA::GUIEventHandlerVisitor& v) {
            v.visit(*this);
    }

};

double tanhDerivativeAsF(double t) {
    return 1.0 - t*t;
}

int main(int argc, char *argv[])
{
    osg::Referenced::setThreadSafeReferenceCounting(true);
    
    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();

    if (!batch) {
        createViewer("Crowd - GA - AI, training");
        // create new keyboard event handler
        KeyboardEventHandler* keh = new KeyboardEventHandler();
        viewer->getEventHandlerList().push_front(keh);
    }

    root = new osg::Group();
    
    //root->addChild(createFloor(200));

//    theFloor = new Terrain(800, 100, 7, 0.3, 3.0);
//    theFloor = new Terrain(400, 50, 7, 0.3, 3.0);
    theFloor = new Terrain(500, 100, 7, 0.35, 3.5);
    root->addChild(theFloor);
    TerrainAgent::terrain = theFloor;

    //theFloor->debugNormals();

    if (!batch) {
        metaroot = new osg::Group();
        metaroot->addChild(root);
    
        viewer->setSceneData( metaroot );
        viewer->realize();
    }
        
    //Localized::setDefaultLocalityMode(Localized::Mixed);
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        100,100,50);

    Environment::gravity = worldGravity;
    AgentCore::setDefaultMaxForce(maxForce);
    AgentCore::setDefaultMaxSpeed(maxSpeed);
    AgentCore::setDefaultEnergyFootprintRatio(energyFootprintRatio);
    AgentCore::setDefaultInitialEnergy(initialEnergy);
    //simu->setTimeFactor(10.0);

// Obstacle code
    const int nobstacles = 75;
    for (int i=0; i<nobstacles; ++i) {
        double x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        
        // Position Obstacle at floor surface
        Obstacle *obs = new Obstacle(x, y, theFloor->getFloorZ(x,y));
        obs->addAvatarToGroup(root);

        // Obstacles update their neighbor lists to help agents
        simu->postCyclicAt(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),
            aistep, obs, &Obstacle::updateNeighbors, 30.0);
    }

    // Create a neuronal agent!!!
    NeuronalAgent *na = new NeuronalAgent(50, 1);//, NeuronalAgent::WinnerTakeAll, NeuronalAgent::Average);
    na->monitorError = true;
/*    na->aloneTLP.transfer = &tanh;
    na->aloneTLP.transferDerivativeAsF = &tanhDerivativeAsF;
    na->neighborTLP.transfer = &tanh;
    na->neighborTLP.transferDerivativeAsF = &tanhDerivativeAsF;
*/
    
    ifstream tlpfile;
    tlpfile.open("tlp.txt");
    if (tlpfile.good()) {
        tlpfile >> na->aloneTLP;
        tlpfile >> na->neighborTLP;
    }
    tlpfile.close();
na->neighborsAccessLock();
na->neighborsAccessUnlock();

// Agent code
    const int nagents = maxPopulationSize;
    for (int i=0; i<nagents; ++i) {
        ExplicitAIAgent *agent = new ExplicitAIAgent((i<nagents/2) ? 1 : 2);
        double x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();

        if (agent->specie==1) na->addTeacher(agent);
        
        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));

        agent->addAvatarToGroup(root);
        
        agent->aiEvent = simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),
            aistep, agent, &Agent::aiStep);
        
        agent->integrationEvent = simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            istep * Utility::Random(),
            istep, agent, &Agent::integrationStep);

        // add roughly one fourth of the agents to the locality database
        // the others keep their own neighborhood
        //if (Utility::Random() < 0.25)
            agent->addToLocalityDatabase();
        
    }
/*    
    // Provide an initial list of neighbors to all agents
    Localized::allObjectsAccessLock();
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }
    // Second order search
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }
    Localized::allObjectsAccessUnlock();
*/
    simu->start(); // in its own thread

    while( !viewer->done() )
    {
        viewer->sync();
        viewer->update();
        viewer->frame();
    }
    // wait for all OSG threads to sync
    viewer->sync();

    // flush all pending simu events (cyclic...)
    simu->flush();
    
    // wait for simu to stop
    simu->stop(true);

    ofstream tlpfilesav;
    tlpfilesav.open("tlp.txt");
    if (tlpfilesav.good()) {
        tlpfilesav << na->aloneTLP;
        tlpfilesav << na->neighborTLP;
    }
    tlpfilesav.close();

    exit(0); // bug in osg/openthreads at object deletion. Use exit instead of return
    return 0;
}



